#ifndef geneobject_include_declared
#define geneobject_include_declared

#include "exceptions.h"
#include "dualSequenceComputation.h" 
#include "SpeciesMap.h"
#include "PWM.h"

typedef map<SequenceObject *, vector<dualSequenceComputation*> *> SequenceMap; 
typedef map<SequenceObject *, vector<dualSequenceComputation*> *>::iterator SeqIterator;

class GeneObject{

 protected:
  // common variables
  SequenceMap *_pSequences;
  int _nbrSeq; 
  vector<int> _strand;
  SeqIterator seqIter; 
  DoubleMap *_pProximityMap; 
  InstanceMap * _pSpeciesInstances;
  InstanceIterator instIter;
  int * _pW; 
 
 public:

  // constructor
  GeneObject(const bool &bMinus);
  // destructor
  virtual ~GeneObject();

  // input 
  void AddSequence(SequenceObject *pSeq); 
  void StoreProximity(const string &id, const double &q, const bool &species); 
  void ApplyGeneRate(const double &r);
  void BuildGeneSpeciesMap(SpeciesMap * _pSpeciesMap);
  void ApplySpeciesWeight(const string &id, const double &w);
  void LinkMotifWidth(int * width){_pW = width; return;};
  virtual void InitializeInstanceTools(){return;};
  virtual void InitializeComputationTools(SpeciesMap * spMap){return;};

  // motif detection
  void InitSpeciesInstanceMap();
  // motif manipulation
  InstanceMap * BuildShiftedSpeciesMap(const int &shift, string Species);
  // motif sampling
  void SamplingStep(SeqIterator Iter, const string &step);

  // inspectors
  int NbrSequences(){return _nbrSeq;}; 
  SequenceMap * GetSequences(){return _pSequences;};
  InstanceMap * GetSpeciesInstances(){return _pSpeciesInstances;};
  vector<dualSequenceComputation*> * FindSequence(const string &species);
  SequenceObject * FindSeqObject(const string &species);
  DoubleMap * GetProxMap(){return _pProximityMap;};

  // virtuals - NOrtho
  virtual void InitNodeInstanceMap(SpeciesMap * spMap){return;};
  virtual void ComputePriorCounts(SpeciesMap * spMap, const double &slowrate, const int &priorscale){return;};
  virtual void StoreEvolPWM(string * id_end, const bool &bSpecies_end, string * daughterid, double*** pRedCounts, vector<PWM*> * pPWMs, vector<double>* pPSPs){return;};
  virtual double ComputeReducedInstBackground(){return 0;};
  virtual void UpdateScoresStep(SeqIterator Iter, PWM * pMotifModel){return;};
  virtual void SetSpeciesInstances(InstanceMap * map){return;}; 
  virtual InstanceMap * GetNodeInstances(){return NULL;};
  // virtuals - phylo
  virtual void InitializeComputationTools(SpeciesMap * spMap, double const &y, double const &z,const bool &bPhylo_2){return;};
  virtual void UpdateNtWeights(){return;};
  virtual void UpdateScoresStep(SeqIterator Iter, double** redgen){return;};
  virtual double** GetNtWeights(){return NULL;};

};

#endif
