#ifndef samplerrun_include_declared
#define samplerrun_include_declared

#include "phyloFastaIO.h"
#include "GeneObject.h"
#include "GFFWriter.h"
#include "PWMIO.h"

typedef map<string, GeneObject*> GeneMap;
typedef map<string, GeneObject*>::iterator GeneMapIterator;

class SamplerRun{

 protected:
  // common variables
  GeneMap * _pMap;
  GeneMapIterator iter;
  int _nbrGenes;
  int _nbrSequences; 
  SpeciesMap *_pSpeciesMap; 
  // other
  bool _tree; // if >(newicktree) has been processed
  bool _bMinusStrand;
  // centralization
  int *_pWidth; 
  // motif
  double * _pDatasetSnf;
  double * _pDatasetPseudo; 

 public:
  // constructor
  SamplerRun();
  // destructor
  virtual ~SamplerRun();

  // input 
  void LoadSequences(string *pFasta, const strand_modes &s);
  virtual GeneObject * AddGene(string * gene){return NULL;};
  void CheckDatasetDimension();
  void LoadBackgroundModels(string *pBg);
  string LoadCorrelations(string *pCorr); 
  string StoreTreeBranchLenghts(string *input);
  bool RemoveSequences(string *pNewick);
  bool UpdateRoot(string *pNewick);
  void StoreGeneMutationRates(string *input);
  void StoreCoregulationWeights(string *input);
  void SetNbrInstPrior(const int &max, string *prior, string Default);
  void InitializeLocalVariables1(const int &w);
  void LoadPSP(string *pPSPfile, bool bPSP_s);

  // motif detection
  void ResetMasks();
  virtual void InitInstanceMap(){return;};
  virtual void DetectionStep(const string &step, const int &nbrIter){return;};
  void UpdateMasks(const double &maxMask, const int &overlap);

  // motif manipulation
  void ShiftFullInstanceMap(const int &max);
  PWM * BuildMotifFromSpeciesInstanceMap(const int &s, string Species);

  // motif result
  int * EvaluateMotif(const double &treshold, const string &Species);
  void PrintMotifScores(GFFWriter * gff, string Motif,PWM * motif, const string &Species);
  void PrintMotifInstances(GFFWriter * gff, const string &Species);
  void PrintToOutput(PWM * pmotif, GFFWriter * pText, PWMIO * pMtrx);
  void PrintToSpecOutput(string box_r_m, GFFWriter * pText_spec, PWMIO * pMtrx_spec);
};


#endif
