#ifndef dualsequencecomputation_include_declared
#define dualsequencecomputation_include_declared

#include "MaskVector.h"
#include "GFFWriter.h"
#include "PWM.h"
#include "utilities.h"
#include "exceptions.h"


class dualSequenceComputation
{
  protected:
    strand_modes _strand;
    SequenceObject *_pSeqObj;// link
    Sequence * _pSequence; // link
    int _seqlength;
    int * _pwlength; // link 
    double _LL;
    double * _pProx; // link
    
    // segment vectors
    ScoreVector *_pSegmentBgScore; // log
    double _LogRedInstBg;// log
    double _dLogP0;
    MaskVector *_pMask;     
    ScoreVector *_pPspScore;
    bool PSP_s;
    ScoreVector *_pMotifScore; // log
    ScoreVector *_pExpWx; // not-log

    // #inst/seq
    int _maxNbrInst;
    Distribution * _pFixedDistr; // link
    Distribution * _pComputedDistr;
    vector<Distribution *> *_pPriorDistrs; // link
    ScoreVector * _pPrior2NormFactor;

  public:
    // constructor
    dualSequenceComputation(const strand_modes s, SequenceObject *pSeqObj, 
                            int *w, double &prox, vector<Distribution*> * dists);
    // destructor
    virtual ~dualSequenceComputation(); /// moet virtual hierbij ?????

    // initializations
    void ComputeBgScores(BackgroundModel *bgM);
    void LoadPspScores(ScoreVector * psp, bool psp_s);

    // inspectors
    int SeqLength(){return _seqlength;};
    int MotifLength(){return *_pwlength;};
    double GetLLscore(){return _LL;};
    double * GetProximity(){return _pProx;};
    MaskVector * GetMaskVector(){return _pMask;};
    double GetPSPAt(int i){if (_pPspScore == NULL) {return 1;} else {return (*_pPspScore)[i];}};
    double GetBgScoreAt(int i){ if (_pSegmentBgScore == NULL || i >= int(_pSegmentBgScore->size())) {return 0;} 
                                else {return (*_pSegmentBgScore)[i];}};

    // motif detection 
    Instance * SampleUniformInstance();
    Distribution * UpdateNbrInstDistribution();
    void StoreRedInstBg(double redBgScore){_LogRedInstBg = redBgScore;}; // log
    InstanceList * SampleInstances(int n);
    InstanceList * SelectBestInstances(int n);
    void LogLikelihoodScore(InstanceList * pList);

    // virtuals
    virtual void UpdateSegmentScores(PWM * pEvolPWM){return;}; // northo
    virtual void ComputeBasicMutProb(double** reduced){return;}; // phylo
    virtual void UpdateMotifScores(double ** redseq){return;}; // phylo
    virtual void UpdatepExpWx(double fixedLogDenum){return;}; // phylo


};

#endif
