#ifndef phylosequencecomputation_include_declared
#define phylosequencecomputation_include_declared

#include "dualSequenceComputation.h"

class phyloSequenceComputation : public dualSequenceComputation
{
  private:
     
    // computation tools for sampling
    double*** _mutationProb; // log
    bool _phylo_2;

  public:
    // constructor
    phyloSequenceComputation(const strand_modes s, SequenceObject *pSeqObj, 
                              int *w, double &prox, vector<Distribution*> * dists, const bool &phylo2);
    // destructor
    ~phyloSequenceComputation();

    // motif detection scoring
    void ComputeBasicMutProb(double** reduced); // log 
    void UpdateMotifScores(double ** redseq); // redseq = extern2
    void UpdatepExpWx(double fixedLogDenum);

};

#endif




















/*

class phyloSequenceComputation{
  private:
    SequenceObject *_pSeqObj;// link
    int _seqlength;
    int * _pwlength; // link
    bool _bMinusStrand; 
    double * _pProximity;// link
    double _LL;
    DataBank * _pDatabankkk; // BETA
    DoubleMap *_pProximityMappp; // BETA
    
  
    // strand segment vectors
    ScoreVector *_pSegmentBgScore; // log
    ScoreVector *_pRevSegmentBgScore; 
    MaskVector *_pMask; //  
    MaskVector *_pRevMask; 
 
    // computation tools for sampling
    double*** _mutationProb; // log
    ScoreCompList *_pCompPt1; // log
    ScoreCompList *_pCompPt2; // log
    ScoreCompIterator compPt;
    StringCompList *_pCompStrPt1; // BETA
    StringCompList *_pCompStrPt2; // BETA
    StringCompIterator StrPt; // BETA
    double _denumerator;
    int _nbrRetain; 
    vector<IndexList *> * _pScoreIndexAll;
    vector<IndexList *> * _pRevScoreIndexAll;
    IndexList::iterator indexPt;
    vector<ScoreList *> *_pScoreVectorAll;// no log
    vector<ScoreList *> *_pRevScoreVectorAll;// no log
    ScoreList::iterator scorePt;

    // #inst/seq
    int _maxNbrInst;
    vector<Distribution *> *_pPriorDistrs; // link
    Distribution * _pFixedDistr;// link
    Distribution * _pComputedDistr;
    Distribution * _pRevComputedDistr;

    // local functions for sampling 
    Instance * _SampleUniformInstance(MaskVector * mask, const strand_modes &s);
    Instance * _SampleOneInstance(ScoreList * pScores, IndexList* pIndex, const strand_modes &s);
    void _SampleInstances(InstanceList * list, const bool &bSample,
                     Distribution * pDist, vector<ScoreList*> * pScores, 
                     vector<IndexList*> * pIndices, const strand_modes &s);
    int * _SampleFromScoreVector(const bool &bSample, const int &N, 
                                 ScoreList * pScores, IndexList* pIndex);

    // local functions for score computations
    bool _ComputePt1(vector<ScoreList *> * pScores,vector<IndexList *> *pIndices,
                     Sequence* seq,ScoreVector* bg, MaskVector* mask, 
                     Matrix redmut);
    bool _ComputeBetaPt1(vector<ScoreList *> * pScores, vector<IndexList *> *pIndices,
                         Sequence* seq, ScoreVector* bg, MaskVector* mask, 
                         StringVector*** redseqstr); // BETA
    bool _ComputePt2(vector<ScoreList *> * pScores, vector<IndexList *> *pIndices, 
                     const int &n, Sequence *seq, ScoreVector *bg, MaskVector *mask);
    bool _ComputeBetaPt2(vector<ScoreList *> * pScores, vector<IndexList *> *pIndices,
                    const int &n, Sequence *seq, ScoreVector *bg, MaskVector *mask); // BETA
    void _AddSegmentToComp(Matrix comp, Sequence *seq, const int &start, const double &bgnorm);
    void _AddStringToComp(StringVector***comp, Sequence *seq, const int &start); // BETA
    void _BuildNbrInstDistr(const strand_modes &s, ScoreVector* part1, const int &N);
    void _ComputeLLscore(InstanceList * List, Distribution * pNbr, double ** RedMutProb);
    void _ComputeBetaLLscore(InstanceList * List,
                             Distribution * pNbr, StringVector ***pRedSeqStr); // BETA


  public:
    // constructor
    phyloSequenceComputation(SequenceObject *pSeqObj, int *w, DoubleMap * prox,
            vector<Distribution*> * dists, DataBank * databank); // integrated BETA
    // destructor
    ~phyloSequenceComputation(); // integrated BETA
    void DeleteCompPt(ScoreCompList * pCompPt);
    void DeleteCompStrPt(StringCompList * pCompPt); // BETA
    void DeleteIndices(vector<IndexList *> * pIndices);
    void DeleteScores(vector<ScoreList *> * pScores);
    void CleanupComputations(); // integrated BETA

    // initializations
    void ComputeBgScores(BackgroundModel *bgM);
  
    // motif detection
    void ResetMasks();
    InstanceList * SampleUniformInstance();
    InstanceList * SampleInstances(const bool &bMultiple, const bool &bSample);
    // score computations
    bool BuildScoreDistribution(double** redmutprob, const bool &bMultiple, 
                                const double & prct, const double&denum);
    bool BuildBetaScoreDistribution(double ** redgenW, StringVector*** pRedSeqStr, 
                         const bool &bMultiple, const double& denum); // BETA
    void BuildMultipleCaseFixed(vector<ScoreList *> * pScores, vector<IndexList* > * pIndices, 
                 Sequence * seq, ScoreVector * bg, MaskVector * mask);
    void BuildBetaMultipleCaseFixed(vector<ScoreList *> * pScores, vector<IndexList* > * pIndices, 
         Sequence * seq, ScoreVector * bg, MaskVector * mask); // BETA
    void BuildMultipleCasePrior(vector<ScoreList *> * pScores, vector<IndexList* > * pIndices, 
                            Sequence * seq, ScoreVector * bg, MaskVector * mask, 
                            const strand_modes &s);
    void BuildBetaMultipleCasePrior(vector<ScoreList *> * pScores,vector<IndexList* > * pIndices,  
                            Sequence * seq, ScoreVector * bg, MaskVector * mask, 
                      const strand_modes &s, StringVector*** redseqstr); // BETA
    // other
    void ComputeBasicMutProb(Matrix reduced);
    void ComputeLLscore(InstanceList * pList, Matrix redmutprob, 
                     StringVector ***predseqstr); // integrated BETA

    // inspectors
    SequenceObject * ParentSequence() const {return _pSeqObj;};
    int GetMaxNbrInst(){return _maxNbrInst;};
    double GetLLscore(){return _LL;};
    Distribution * GetCompDistr(const strand_modes &s);

};

#endif

*/

