#ifndef phylutilities_include_defined
#define phylutilities_include_defined

#include "exceptions.h"
#include "Distribution.h"
#include "Instance.h"
#include "BackgroundModel.h"
typedef list<Instance*> InstanceList; 
typedef list<Instance*>::iterator InstanceIterator; 

/*
 * Here the PHYLINCLUSIVE namespace is defined where all 
 * auxillary function are declared that could not be
 * assigned to a specific class for phylogenetic motifdetection
 */

namespace PHYLINCLUSIVE {

  // read input from a prior information on #inst/seq (SpeciesMap-object)
  double ReadValue(string &value);
  bool ReadPriorIndex(string &line);
  ScoreVector * ExtractPriorParameters(const int &max, string input);
  vector<Distribution*> * CreateNbrInstDistribution(const int &max, ScoreVector * prior);
  ScoreVector * CreateDefaultProbabilities(const int &m, const double &one, const double &kappa);
  ScoreVector * CreateBinomialProbabilities(const double &one, const int &m);
  void ReadProbabilities(ScoreVector * prob, string &line, const int &m);

  // specific sequence scoring functions (based on utilities.h) 
  void SegmentLogBackgroundScore(ScoreVector *pbx, SequenceObject *pSeq, 
      const strand_modes &s, BackgroundModel *pBgM, const int &W); //(phyloSequenceComputation-object)
  int Site2Index(SequenceObject * pSeq, const strand_modes &s, const int &start, const int &order);

  // general mathematical functions 
  int fac(int n);
  double SumVector(vector<double> *pVector, int start, int length);
  double SumVector(vector<double> * pVector);
  double ComputeLogBetaScore(double** hw, int *W);
  double LogGamma(const double &x);
  double LogBeta(double* pWeightj);

  // added for NOrthoSampler
  bool gluInvertMatrix(const double m[16], double invOut[16]);

  // transform
  InstanceList * OrderInstances(InstanceList * list);

}

#endif
