#include "FastaIO.h"

#include <iostream>


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
		 
  Description:  constructor used for FuzzyClustering of MS-3.1.5 files
                the lines start with #id: instead of >
  Author_Date:  MC_2009/12/07
  Date:         MC_2011/07/03 allow phylofasta file as input in MS.
******************************************************************************/
FastaIO::FastaIO(string * fileName)
{
  // open file for reading
  _ifs.open(fileName->c_str(), ios::in);

  if (!_ifs)
  {
    // file is not open for reading
    cerr << "--Error-- FastaIO::FastaIO() unable to open file: "
      << *fileName << " for reading!" << endl;
    _isOpen = false;
    _hasNext = false;
    /* 
     * the program will not stop here
     * the user should check with IsOpen() in main program
     */
  }
  else
  {
    _isOpen = true;

    // read lines untill first line starting with a '>' is found
    // but skip lines with double >> (refers to gene-group)
    do
    {
      getline(_ifs, _pLine, '\n');
      // cerr << "DEBUG: " << _pLine << endl;
    }
    //while (_pLine[0] != '>' && !_ifs.eof()); 
    //while (!( _pLine[0] == '>' || 
    while (!( (_pLine[0] == '>' && _pLine[1] != '>') || 
               (_pLine[0] == '#' && _pLine[1] == 'i' && _pLine[2] == 'd' 
                 && _pLine[3] == ':' && _pLine[4] == ' ') // 2009/12/07
            ) && !_ifs.eof()); 

    if (!_ifs.eof())
    {
      _hasNext = true;
      // cerr << "DEBUG: _hasNext = true " << endl;
    }
    else
    {
      _hasNext = false;
      // cerr << "DEBUG: _hasNext = false " << endl;
    }
  }
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
FastaIO::~FastaIO()
{
  // close file
  if (_ifs.good())
    _ifs.close();

  return;
}


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
void
FastaIO::Close()
{
  _ifs.close();
  _isOpen = false;
  return;
}


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:  load a sequence from FASTA file
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  Date:         MC_2011/07/03 allow phylofasta file as input in MS.
                // mark this function is not used in 5.0.5, 
                only changed here to keep code the same with 3.1.5  
******************************************************************************/
SequenceObject *
FastaIO::NextSequence()
{

  if (_isOpen && _hasNext)
  {

    // create empty string to store sequence
    string *pSeqStr = new string("");

    string *pDescStr = new string(_pLine);
    
    // cerr << "DEBUG: " << *pDescStr;

    while (!_ifs.eof())
    {
      // read next line
      getline(_ifs, _pLine, '\n');

      // move to next line if line is empt and skip >>gene-group lines 
      if (_pLine.empty() || (_pLine[0] == '>' && _pLine[1] == '>'))
        continue;

      if (_pLine[0] == '>')
      {
        break;
      }
      else
      {
        string::size_type pos = 0;
        string tokens = " \t\n\r\f";
        while ((pos = _pLine.find_first_of(tokens)) != string::npos)
        {
          _pLine.erase(pos, 1);
        }
        // append line to seqStr;
        pSeqStr->append(_pLine);

      }
    }

    if (!_ifs.eof())
    {
      _hasNext = true;
    }
    else
    {
      _hasNext = false;
    }

    // create new sequence object
    // cerr << *seqStr << endl;

    SequenceObject *pSeqObj = new SequenceObject(pSeqStr);
    pSeqObj->SetDescription(pDescStr);

    delete pSeqStr;
    delete pDescStr;

    return pSeqObj;

  }
  else
  {
    // sequence was not open anymore 
    return NULL;
  }
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool
FastaIO::IsOpen()
{
  return _isOpen;
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool
FastaIO::HasNext()
{
  return _hasNext;
}

/******************************************************************************
  Description:  ReadLine : read _pLine (assumes this line is not empty, 
                line may but does not necessarily start with >)
				and set cursor on next valid line : starting with > or #id: 
  Author_Date:  MC_2009/12/07
******************************************************************************/
string
FastaIO::ReadLine()
{
  //cerr << "debug - FastaIO::ReadLine - begin" << endl;
  string thisline = _pLine;
	
  _hasNext = false;
  // proceed untill next valid input line
  while (!_ifs.eof())
  {
    getline(_ifs, _pLine, '\n');
    if (_pLine.empty())
      continue;
    if ( _pLine[0] == '>' || 
        (_pLine[0] == '#' && _pLine[1] == 'i' && _pLine[2] == 'd' 
                 && _pLine[3] == ':' && _pLine[4] == ' ')
       ) _hasNext = true;
    break; // other line not starting with > or "id: 
  }
  if (_ifs.eof()) _isOpen = false;
  //cerr << "debug - FastaIO::ReadLine - end" << endl;
  return thisline;
}
/******************************************************************************
  Description:  CheckLineStart()
  Author_Date:  MC_2011/04/11
******************************************************************************/
bool 
FastaIO::CheckLineStart(string Start)
{
  if (_pLine[0] == Start[0]) 
  {
    cerr << "Line " << _pLine << " starts with # and will be skipped." << endl;
    // set cursor ready on next line
    ReadLine(); 
	return 1;
  } 
  return 0;
}
