#include "GeneObject.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
GeneObject::GeneObject(const bool &bMinus)
{
  //cerr << "Debug-GeneObject::GeneObject -begin" << endl;
  _pSequences = new SequenceMap; 
  _nbrSeq = 0; 
  _strand.push_back(0); if(bMinus) _strand.push_back(1); 
  _pProximityMap = new DoubleMap;
  _pSpeciesInstances = new InstanceMap();
  _pW = NULL;  
  //err << "Debug-GeneObject::GeneObject- end" << endl;
}
/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
GeneObject::~GeneObject()
{
  //cerr << "Debug-GeneObject::~GeneObject- begin" << endl;
  // _pSequences and its Objects
  if (_pSequences != NULL)
  {
    seqIter = _pSequences->begin(); 
    for (; seqIter != _pSequences->end(); seqIter++)
    { 
      // vector and dualSequenceComputation(s)
      if ((*seqIter).second != NULL)
      { for(int i = 0; i < (int)_strand.size(); i++)
          delete (*(*seqIter).second)[i]; // comp-object
        delete (*seqIter).second; // vector
      }
      // SequenceObject
      if ((*seqIter).first != NULL)
        delete (*seqIter).first; 
    }
    delete _pSequences; 
  }
  _pSequences = NULL; 
  // proximitymap
  if (_pProximityMap != NULL)
    delete _pProximityMap;
  _pProximityMap = NULL;  

  // _pInstances 
  if(_pSpeciesInstances != NULL)
    delete _pSpeciesInstances; // delete the map
  _pSpeciesInstances = NULL;
  // _pW
  _pW = NULL; // do not delete - see destructor SamplerRun

  //cerr << "Debug-GeneObject::~GeneObject- end" << endl;
}
/******************************************************************************
  Description:  AddSequence
  Author_Date:  MC_2009/01/25
******************************************************************************/
void 
GeneObject::AddSequence(SequenceObject *pSeq)
{
  //cerr << "Debug-GeneObject::AddSequence- begin" << endl;
    string id = *(pSeq->GetID());
    // check if double input
    for (seqIter = _pSequences->begin(); seqIter != _pSequences->end(); seqIter++)
    {
      if ( ! (id).compare(*(*seqIter).first->GetID()) ) 
      { id.append("-");
        id.append(*(pSeq->GetGenID()));
        id.append(" double sequence input in fastafile.");
        delete pSeq; pSeq = NULL;
        throw ErrorInputfiles(id); // catch in mainPHMS
      }
    }

  // add the sequence, dualSequenceComputation to be created later
  (*_pSequences)[pSeq] = NULL;
  // keep count
  _nbrSeq++; 

  //cerr << "Debug-GeneObject::AddSequence- end" << endl;
  return; 
}
/******************************************************************************
  Description:  StoreProximity(id, value) : store species|node and q in 
                proximityMap 
  the proximity is always related to the parent of the stored id
  (so the root is not present here)
  Author_Date:  MC_2012/11/22
******************************************************************************/
void
GeneObject::StoreProximity(const string &id, const double &q, const bool &species)
{
  //cerr << "Debug-GeneObject::StoreProximity -begin" << endl;
  // proceed if there is a sequence for this species in the dataset
  if ( (species && FindSeqObject(id) != NULL) || !species)
  {
    // catch to StoreStarBranchLenghts if double input
    if (_pProximityMap->find(id) != _pProximityMap->end())
      throw InvalidInput(id);
    // store proximity
    (*_pProximityMap)[id] = q;
  }
  //cerr << "Debug-GeneObject::StoreProximity -end" << endl;
  return;  
} 

/******************************************************************************
  Description:  ApplyGeneRate : power each proximity with rate
  Author_Date:  MC_2012/11/26
******************************************************************************/
void
GeneObject::ApplyGeneRate(const double &r)
{
  //cerr << "Debug-GeneObject::ApplyGeneRate -begin" << endl;
  map<string, double>::iterator proxIter = _pProximityMap->begin();
  for(; proxIter != _pProximityMap->end(); proxIter++)
    (*proxIter).second = pow((*proxIter).second, r);
  //cerr << "Debug-GeneObject::ApplyGeneRate -end" << endl;
  return;
}
/******************************************************************************
  Description:  ApplySpeciesWeight : power species-proximity with 1/weight       
  Author_Date:  MC_2012/11/26
******************************************************************************/
void
GeneObject::ApplySpeciesWeight(const string &id, const double &w)
{
  //cerr << "Debug-GeneObject::ApplySpeciesWeight -begin" << endl;
  if ( _pProximityMap->find(id) != _pProximityMap->end())
    // apply weight if the species(~sequence) is present in _pProximityMap
    (*_pProximityMap)[id] = pow((*_pProximityMap)[id], 1/w);
  //cerr << "Debug-GeneObject::ApplySpeciesWeight -end" << endl;
  return;
}

/******************************************************************************
  Description:  InitSpeciesInstanceMap :  
    for species : sample 1 instance from uniform segment-score-distribution
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
GeneObject::InitSpeciesInstanceMap()
{
  Instance * pInst = NULL;
  //cerr << "Debug-GeneObject::InitSpeciesInstanceMap() -begin" << endl;

  // Species:: iterate through each dualSequenceComputation-object 
  // and store sampled instances (+/-) in _pSpeciesInstances
  seqIter =_pSequences->begin();
  for (; seqIter != _pSequences->end(); seqIter++)
  { //  this means the sequence exists in this geneobject
    if ((*seqIter).second != NULL) 
    {
      for(int i = 0; i < (int)_strand.size(); i++)
      {
        // sample random one Instance from uniform distribution
        pInst = (*(*seqIter).second)[i]->SampleUniformInstance(); 
        // add instance to _pInstances
        if (pInst != NULL){ _pSpeciesInstances->AddInstance(pInst);}
      }
    }
  }
  //cerr << "Debug-GeneObject::InitSpeciesInstanceMap() -end" << endl;
  return;
}

/******************************************************************************
  Description:  BuildShiftedMap(int shift,string Species) : shift + check mask
                // do not replace with _pInstances !
	  'Species' may be all of a particular one.
  Author_Date:  MC_2014/01/22
******************************************************************************/
InstanceMap * 
GeneObject::BuildShiftedSpeciesMap(const int &shift, string Species)
{
  //cerr << "Debug-GeneObject::BuildShiftedSpeciesMap -begin" << endl;
  // new map to be constructed
  InstanceMap * pMap = new InstanceMap();

  SequenceObject *pObj = NULL; 
  int s(0);  
  strand_modes str;
  InstanceIterator ii;
  string * pInstSpeciesID = NULL;
  for (ii = _pSpeciesInstances->begin(); ii != _pSpeciesInstances->end(); ii++)
  {
    pObj = (*ii)->ParentSequence();
    pInstSpeciesID = pObj->GetID();
    if (Species == "all" || Species == *pInstSpeciesID)
    {
      s = (*ii)->Start();
      str = (*ii)->Strand();
      // add shifted instance to new pMap (if not out of range and not masked)
      if ( s + shift + *_pW <= pObj->Length() && s + shift >= 0 &&
         (*(*_pSequences)[pObj])[(int)(str == minus_strand)]->GetMaskVector()->
                                                 GetValueAt(s + shift) != 0 )
      {
        pMap->AddInstance(new Instance(pObj, str, s + shift, *_pW));
      }
    }
  }
  pObj = NULL;
  //cerr << "Debug-GeneObject::BuildShiftedSpeciesMap -end" << endl;
  return pMap;
}
/******************************************************************************
  Description:  SamplingStep: sample instances in a given sequence (Iter)
    // use local psp_s boolean in case of psp impact for sampling requested
  Author_Date:  MC_2014/01/16
******************************************************************************/
void 
GeneObject::SamplingStep(SeqIterator Iter, const string &step)
{
 //cerr << "Debug-GeneObject::SamplingStep -begin" << endl;
  dualSequenceComputation * pComp = NULL;
  InstanceList * pInstsL(NULL);
  int N;
  // one or both strands
  for(int i = 0; i < (int)_strand.size(); i++)
  {
    pComp = (*(*Iter).second)[i];
    // sample the new instances
    if ( step !="conv") // init or core
    { if (step == "init") 
      {N = 1;}
      else // core
      { N = pComp->UpdateNbrInstDistribution()->TakeSample();
//  cerr << "N=" << N << endl;
      }
      // sample the instances
      pInstsL = pComp->SampleInstances(N);
      _pSpeciesInstances->AddInstances(pInstsL);
    }
    else // conv
    { N = pComp->UpdateNbrInstDistribution()->SelectMax();
      pInstsL = pComp->SelectBestInstances(N);
      _pSpeciesInstances->AddInstances(pInstsL);
      // also compute the LL-score
      pComp->LogLikelihoodScore(pInstsL); 
    }
    // cleanup 
    delete pInstsL; pInstsL = NULL; // list, not the instances
  }
  //cerr << "Debug-GeneObject::SamplingStep -end" << endl;
  return;
}

/******************************************************************************
  Description:  FindSeqObject(string species)
  Author_Date:  MC_2014/02/15
******************************************************************************/
SequenceObject *
GeneObject::FindSeqObject(const string &species)
{
  //cerr << "Debug-GeneObject::FindSeqObject -begin" << endl;
  SequenceObject * pObj = NULL;
  for(seqIter=_pSequences->begin(); seqIter!=_pSequences->end(); seqIter++)
  {
    if ( ! (species).compare(*((*seqIter).first->GetID())))
    {
      pObj = (*seqIter).first;
      break;
    }
  }
  //cerr << "Debug-GeneObject::FindSeqObject -end" << endl;
  return pObj;
}
/******************************************************************************
  Description:  FindSequence(const string &species)
  Author_Date:  MC_2014/02/15
******************************************************************************/
vector<dualSequenceComputation*> * 
GeneObject::FindSequence(const string &species)
{
  //cerr << "Debug-GeneObject::FindSequence -begin" << endl;
  vector<dualSequenceComputation*> * pSeqComps = NULL;

  for(seqIter=_pSequences->begin(); seqIter!=_pSequences->end(); seqIter++)
  {
    if ( ! (species).compare(*((*seqIter).first->GetID())))
    {
      pSeqComps = (*seqIter).second;
      break;
    }
  }
  //cerr << "Debug-GeneObject::FindSequence -end" << endl;
  return pSeqComps;
}

