#include "NOrthoGeneObject.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
NOrthoGeneObject::NOrthoGeneObject(const bool &bMinus):GeneObject(bMinus)
{
  //cerr << "Debug-NOrthoGeneObject::NOrthoGeneObject -begin" << endl;
  _pPriorEvolMatrices = new MatrixMap;
  _pNodeInstances = new InstanceMap();
  _priorScale = 1;
  //cerr << "Debug-NOrthoGeneObject::NOrthoGeneObject- end" << endl;
}
/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
NOrthoGeneObject::~NOrthoGeneObject()
{
  //cerr << "Debug-NOrthoGeneObject::~NOrthoGeneObject- begin" << endl;
  // _pPriorEvolMatrices and its matrices
  if (_pPriorEvolMatrices != NULL)
  {
    MatrixIter Iter = _pPriorEvolMatrices->begin(); 
    for (; Iter != _pPriorEvolMatrices->end(); Iter++)
    { 
      // double**
      if ((*Iter).second != NULL)
      { for(int i = 0; i < 4; i++)
          delete[] (*Iter).second[i]; // 
        delete[] (*Iter).second; 
      }
      (*Iter).second = NULL;
    }
    delete _pPriorEvolMatrices; 
  }
  _pPriorEvolMatrices = NULL;
	
  // _pInstances 
  if(_pNodeInstances != NULL)
    delete _pNodeInstances; // delete the map
  _pNodeInstances = NULL;

  //cerr << "Debug-NOrthoGeneObject::~NOrthoGeneObject- end" << endl;
}

/******************************************************************************
  Description:  InitializeComputationTools 
                create northoSequenceComputation objects
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoGeneObject::InitializeComputationTools(SpeciesMap * spMap)
{
  //cerr << "Debug-NOrthoGeneObject::::InitializeComputationTools -begin" << endl;
  int M(0);
  map<string, speciesgeg>* pMap = spMap->GetMap();

  // do not create a local copy of SpeciesMap
  // do not merge proximities for cases in which no sequence is present
  // it complexifies the iteratieve updating steps (~reduced matrix)
  // it may disturb convergence (shorter proximities so less conservation so less guidance)
  // we should assume that (certainly in the future) a lack of sequences is not abundant
	// BUT : remove 'not-informational' nodes (see at end of this section)
	// i.e. that appear to have no downtree-species-sequences,
	// -> the id of such nodes is removed from _pProximityMap 
	// these branches will no longer be incorporated in update/sampling steps
	
  // First do species-sequences:: create northoSeqComp objects 
  // only do this in case -M > 0 and SeqL > motif width
  // if not, then leave noSeqComp object to NULL 
// so the objects are not removed as such when sequences are not suitable for motif detection, 
// but by setting NULL, they will not be assessed throughout motif detection
// ! check for NULL on noSeqComp must be implemented further on !!!!!!!!!!!!!!!
  vector<dualSequenceComputation *> *seqComp = NULL;
  northoSequenceComputation * seqCompObj = NULL;
  string *species = NULL;
  for(seqIter = _pSequences->begin(); seqIter!=_pSequences->end(); seqIter++)
  {
    species = (*seqIter).first->GetID();
    M = 0;
    // check if -M is not zero for this species
    if ((*pMap)[*species].pNbrInstDistr != NULL)
    {
      if ((*(*pMap)[*species].pNbrInstDistr)[0] != NULL) 
        M = (int)(*(*pMap)[*species].pNbrInstDistr)[0]->Size()-1;
      else M = (int) (*pMap)[*species].pNbrInstDistr->size()-1;
    }
    // and the fasta sequences is long enough
    if( (*seqIter).first->Length() <= *_pW )
    { // too short sequence, falls back to M = 0
      M = 0;
    }
    if (M == 0)
    { // do not create computation object
      continue;
    }

    // create phyloSequenceComputation-vector
    seqComp = new vector<dualSequenceComputation *>;
    // link seqobj, width, proximity, priorDistr
    seqCompObj = new northoSequenceComputation( plus_strand, 
                       (*seqIter).first, _pW, (*_pProximityMap)[*species], 
                       (*pMap)[*species].pNbrInstDistr );
    // compute segment/sequence bg scores
    seqCompObj->ComputeBgScores((*pMap)[*species].pBgM);
    // add 
    seqComp->push_back(seqCompObj);
    if ((int)_strand.size() == 2)
    { seqCompObj = new northoSequenceComputation(minus_strand, 
                         (*seqIter).first, _pW, (*_pProximityMap)[*species], 
                         (*pMap)[*species].pNbrInstDistr );
      seqCompObj->ComputeBgScores((*pMap)[*species].pBgM);
      seqComp->push_back(seqCompObj);//(*seqIter).second[1] = seqComp[1];
    }
    (*seqIter).second = seqComp;
    // reset 
    seqComp = NULL;
    seqCompObj = NULL;
    species = NULL;
  }
	
  // all species-sequences have been processed now
  // Now remove 'non-informational' nodes if so present for this GeneObject
    // so in future update/sampling steps, always check for presence of an id 
    // (species or node)-id in _pProximityMap, if it is not there, 
    // it is not to be regarded as existing !
  string nodeid; StringVector * pAll(NULL);
  bool bInformational;
  // work tree-upwards
  for (int node = (spMap->NbrNodes())-1 ; node >=0; node--)
  { 
    bInformational = false;
    // convert node-number to node-string
    nodeid = "N"; ostringstream ostr; ostr << node; nodeid.append(ostr.str());
    // check if this node has 'active' species-or node-daughters 
    // get the relations
    pAll = (*(spMap->GetMap()->find(nodeid))).second.pRelations;
    // process all daughters of this node
    for (int i = 0; i < int(pAll->size()); i++)
    { 
      // skip the first id (parent of this parentid)
      if (nodeid != "N0" && i == 0) {continue;}        
      // store "informational" node-daughters (possible next parent candidates)
      if ( _pProximityMap->find((*pAll)[i]) != _pProximityMap->end())
      { bInformational = true; break;}
    }
    if (!bInformational)
    { // remove this node from _pProximityMap and _pPriorEvolMatrices
      _pProximityMap->erase(nodeid);
cerr << "--Warning::NOrthoGeneObject::InitializeComputationTools: " << nodeid << " removed." << endl;
    }
    pAll = NULL;
  }

  //cerr << "Debug-NOrthoGeneObject::InitializeComputationTools -end" << endl;
  return;  
}

/******************************************************************************
  Description:  ComputePriorCounts
                compute prior pseudo evolution matrix for each 'active'branch,
                to be stored in _pPriorEvolMatrices
                at the same time create pSNF for each node (estimated values)
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoGeneObject::ComputePriorCounts(SpeciesMap * spMap, const double &slowrate, const int &priorscale)
{
  //cerr << "Debug-NOrthoGeneObject::ComputePriorCounts -begin" << endl;
	
  _priorScale = priorscale; // to be used in BuildEvolutionProbs
	
  map<string,speciesgeg> * pMainMap = spMap->GetMap();  
  map<string,speciesgeg>::iterator mapIter;
  vector<string> * pDaughters(NULL);
  string nodeid("NULL"), daughterid("NULL");
  double BgProx(-1), MotifProx(-1); double * pEstSNF(NULL), *snf(NULL); 
  double * pPartSnf = new double[4];
  double sum;

  // start with the highest node number (work tree upwards because the node-SNFs 
  // are sequentially computed at the same time as the PriorEvolMatrices)
  for (int node = (spMap->NbrNodes())-1; node >=0; node--)
  {
    // convert node-number to node-string
    nodeid = "N";
    ostringstream ostr; ostr << node; nodeid.append(ostr.str());

    // skip inactivated branches
    if ((nodeid != "N0") && _pProximityMap->find(nodeid) == _pProximityMap->end())
    { continue;}
 
    // localize the node-id in the speciesMap
    mapIter = pMainMap->find(nodeid);

    // compute the prior evolution matrix for each daughter of this node
    //// iterate through pRelations of this node
    pDaughters = (*mapIter).second.pRelations; 
    pEstSNF = (*mapIter).second.pSNF; 
    int d = 1; if (nodeid == "N0") { d = 0;} // skip parentid except for root
    for (; d < (int)pDaughters->size(); d++)
    {

      daughterid = (*pDaughters)[d];
      // skip if not present in _pProximityMap, then daughter is a species 
      // i.e. sequence not present or node was inactivated
      if (_pProximityMap->find(daughterid) == _pProximityMap->end())
      { continue;}  
      // get the SNF of the daughter (could be species or node)
      snf = (*pMainMap)[daughterid].pSNF;
      if (snf == NULL)
      { cerr << "Error-debug, SNF could not be NULL here..." << endl;exit(-1);}
      // look up the background-proximity for this branch
      //BgProx = (*(_pProximityMap->find(daughterid))).second;
      BgProx = (*_pProximityMap)[daughterid];
      if (BgProx == -1) 
      { cerr << "Error-debug, prox could not be -1 here..." << endl; exit(-1);}
      // compute the BsM for this branch
      double BgSubstM[16];
      for (int i1 = 0; i1<4; i1++)
      { 
        for (int i2 = 0; i2 <4; i2++)
        { BgSubstM[i1*4+i2] = (1-BgProx) * snf[i1];
          if ( i1 == i2 ) BgSubstM[i1*4+i2] += BgProx;
        }
      }

      // Update the intermediate computation of the estimated node-SNF
      // with the backinverted-SNF-contribution of this daughter
      //// use inverse of the BackgroundSubstitutionMatrix (phylutilities)
      double invBgSubstM[16];
      for (int i1 = 0; i1<16; i1++){invBgSubstM[i1] = 0;}
      if (! PHYLINCLUSIVE::gluInvertMatrix(BgSubstM, invBgSubstM))
      { cerr << "Error-debug, non-singular matrix is inversible (gluInvertMatrix)." << endl;exit(-1);}
      for (int i1 = 0; i1<4; i1++)
      { 
        pPartSnf[i1] = 0; 
        for (int i2 = 0; i2 <4; i2++) 
        { pPartSnf[i1] += snf[i2] * invBgSubstM[i1*4+i2];}
        pEstSNF[i1] += pPartSnf[i1]; // toch + ipv *
      }
      // Now compute the prior evolution at slower motif evolution rate.
      MotifProx = pow(BgProx, slowrate); // 0.5 = 50% of background
//cerr << "MsM for " << nodeid << "/" << daughterid << "::" << endl;
      double ** matrix = new double*[4];
      for (int i1 = 0; i1<4; i1++)
      { matrix[i1] = new double[4];
        for (int i2 = 0; i2 <4; i2++)
        { matrix[i1][i2] = (1-MotifProx) * snf[i1];
          if ( i1 == i2 ) matrix[i1][i2] += MotifProx;
//cerr << matrix[i1][i2]<< "\t";
        }
//cerr << endl;
      }
      // add to map
      (*_pPriorEvolMatrices)[daughterid] = matrix;
      // reset
      BgProx = -1; snf = NULL;
    } // all daughters processed

    // now normalize the node estimated SNF 
    sum = 0;
//cerr << "SNF for " << nodeid << "::(";
    for (int i1 = 0; i1<4; i1++)
    { 
      if (pEstSNF[i1] < 0)
      { cerr << "Error-debug, wrong SNF computation (negative value)." << endl;exit(-1);}
      sum += pEstSNF[i1];
    }
    for (int i1 = 0; i1<4; i1++){ pEstSNF[i1] /= sum; 
//cerr << " " << pEstSNF[i1];
    }
//cerr << ")" << endl;
    pEstSNF = NULL; // reset
  }
  //cleanup
  delete[] pPartSnf; pPartSnf = NULL;
  //cerr << "Debug-NOrthoGeneObject::ComputePriorCounts -end" << endl;
  return;  
}


/******************************************************************************
  Description:  InitNodeInstanceMap :  
    for 'active' node : sample 1 instance from uniform PWM
    Not-informational branches (having no species-sequences downwards)
    were inactivated (removed from _pProximityMap) in InitializeComputationTools
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoGeneObject::InitNodeInstanceMap(SpeciesMap * spMap)
{
  Instance * pInst = NULL;
  //cerr << "Debug-NOrthoGeneObject::InitNodeInstanceMap() -begin" << endl;

  // Nodes:: iterate through each node-id bottom-upwards
  // sample a random segment from a randomly chosen nearest-species-daughter
  // nodes are assumed to be only one strand (no +/- differentiation)
  // and store sampled Instance in _pNodeInstances
  string nodeid; string daughterid; string parentid;
  StringVector * pSpeciesDaughters(NULL), * pNodes(NULL), *pAll(NULL);
  MaskVector * pUniform = NULL;
  Distribution * pDist = NULL;
  // work tree upwards
  for (int node = (spMap->NbrNodes())-1 ; node >=0; node--)
  { 
    // convert node-number to node-string
    nodeid = "N"; ostringstream ostr; ostr << node; nodeid.append(ostr.str());
	  
    if ( (nodeid != "N0") && _pProximityMap->find(nodeid) == _pProximityMap->end())
    { continue;} // skip inactivated nodes
    // get the nearest species-daughters
    pNodes = new StringVector;
    pNodes->push_back(nodeid);
    pSpeciesDaughters = NULL;
    while (pSpeciesDaughters == NULL)
    { 
      // explore node-branches that may have species-daughters
      // in a first iteration, this is only the subject node-id
      // in next iterations, node-daughters become next candidate parents 
      while ( pNodes != NULL)
      {
        parentid = (*pNodes)[0];
        pNodes->erase(find (pNodes->begin(), pNodes->end(), parentid)); // not needed anymore as a candidate 
        pSpeciesDaughters = new StringVector;
        // get the relations
        pAll = (*(spMap->GetMap()->find(parentid))).second.pRelations;
        // process all daughters of this parentid
        for (int i = 0; i < int(pAll->size()); i++)
        { 
          // skip the first id (parent of this parentid) except for root
          if (nodeid != "N0" && i == 0) {continue;}
          // and skip inactivated elements
          if (_pProximityMap->find((*pAll)[i]) == _pProximityMap->end())
          { continue;}
          // store active node-daughters (possible next parent candidates)
          if ( (*(spMap->GetMap()->find((*pAll)[i]))).second.pBgM == NULL)
          { pNodes->push_back((*pAll)[i]);}
          else // store species-daughter,
          { pSpeciesDaughters->push_back((*pAll)[i]);}
        }
        if (int(pSpeciesDaughters->size()) > 0)
        { // a valid set of species-daughters has been found
          if (pNodes != NULL) {delete pNodes;} pNodes = NULL;
        } 
        else
        { // further explore the tree downwards
          if (pSpeciesDaughters != NULL){delete pSpeciesDaughters;} pSpeciesDaughters = NULL;
          // unless there are no more next candidate node-parents available
          if ((int)pNodes->size() == 0)
          { // end of tree is reached 
			  cerr << "--Debug1::NOrthoGeneObject::InitNodeInstanceMap:: error in processing tree... Exit." << endl; 
			  exit(-1);
          }
        }
      } 
    }
    // skip this node-branch if not informational
    if (pSpeciesDaughters == NULL || int(pSpeciesDaughters->size()) == 0) 
    { 
 cerr << "--Debug2::NOrthoGeneObject::InitNodeInstanceMap:: error in processing tree... Exit." << endl; 
 exit(-1);
    }
    if (pNodes != NULL) {delete pNodes;} pNodes = NULL;
    //// sample a species-daughter
    pUniform = new MaskVector(int(pSpeciesDaughters->size())); // all one
    pDist = new Distribution(pUniform->GetMask());
    daughterid = (*pSpeciesDaughters)[pDist->TakeSample()];
    // cleanup
    delete pSpeciesDaughters; pSpeciesDaughters = NULL;
    delete pUniform; pUniform = NULL;
    delete pDist; pDist = NULL;
    // sample a random sequence-segment from this species-daughter
    vector<dualSequenceComputation*>* pTest = FindSequence(daughterid);
    pInst = ((*pTest)[0])->SampleUniformInstance(); // take plus_strand
    if (pInst != NULL)
    { 
      // reset some instance-information (unlink SeqObj, add node-id)
      // add instance to pMap
      _pNodeInstances->AddInstance(pInst->UnsetSequenceLink(nodeid));
    }
  }
  //cerr << "Debug-NorthoGeneObject::InitNodeInstanceMap() -end" << endl;
  return;
}
/******************************************************************************
  Description:  StoreEvolPWM : select the appropiate (hard/non-hard) 
                reduced-evol-probs (i.e. based on the site description of the 
                known-branch-end and add prior-pseudo-evol-counts
                // make a PWM and store it
                // extract the psp-site (of the known branch end) and store it
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoGeneObject::StoreEvolPWM(string * id_end, 
		const bool &bSpecies_end, string * daughtid, double*** pRedCounts, 
						vector<PWM*> * pPWMs, vector<double>* pPSPs)
{
//cerr << "--debug--NOrthoGeneObject::StoreEvolPWM--begin" << endl;
//cerr << "id_end=" << *id_end << ",bSp_end=" << bSpecies_end << ",daughter=" << *daughtid << endl;
  InstanceMap * instMap = NULL;
  InstanceIterator i;
  Sequence * pSite = NULL;
  PWM * pEvolPWM(NULL);
  double ** pPriorCounts(NULL);
  double pspsite = 1; // default
  
  // known end is a species or a node
  if (bSpecies_end) { instMap = _pSpeciesInstances; }
  else { instMap = _pNodeInstances;}

  // select correct pseudo_prior_evolution
  pPriorCounts = (*_pPriorEvolMatrices)[*daughtid];

  // for each of the known sites at branch end 
  // by previous checks, there is at least one site at the call of this function
  // i.e. if more than one site, treat it as different branches
  // i.e. store PWM and psp separately
  for(i = instMap->begin(); i != instMap->end(); i++)
  {
    if ((*((*i)->PrintSeqName())) != *id_end) 
    {continue;}
    // get the instance description of the known branch-end
    pSite = (*i)->Site(); 
    pEvolPWM = BuildEvolutionProbs(pRedCounts, pPriorCounts, 
                                    pSite, (id_end != daughtid));
    // and store it in pPWMs
    pPWMs->push_back(pEvolPWM);
    // also extract/store the psp of this site
    pspsite = 1; // default
    if (bSpecies_end)
    {
      if ((*i)->Strand() == plus_strand)
        pspsite = (*((*(_pSequences->find(
				(*i)->ParentSequence()))).second))[0]->
				GetPSPAt((*i)->Start());
      else 
        pspsite = (*((*(_pSequences->find(
				(*i)->ParentSequence()))).second))[1]->
				GetPSPAt((*i)->Start());
	}
	else // set zero for nodes (to be replaced by "minpsp" further on)
    { pspsite = 0;} 
    pPSPs->push_back(pspsite);
    // reset 
    pSite = NULL;
  }
  //cerr << "--debug--NOrthoGeneObject::StoreEvolPWM--end" << endl;
  return;
}
/******************************************************************************
  Description:  BuildEvolutionProbs
                // select appropiate counts from pReducedCounts 
                and apply pseudo_prior from pPriorCounts
  Author_Date:  MC_2014/02/15
******************************************************************************/
PWM *
NOrthoGeneObject::BuildEvolutionProbs(double*** pReducedCounts, 
				double ** pPriorCounts, Sequence * pSite, bool bknownParent)
{
  //cerr << "--debug::NOrthoGeneObject::BuildEvolutionProbs-begin" << endl;
	
  Matrix pMatrix = new double*[*_pW];
  for(int w = 0; w < *_pW; w++)
  {
    pMatrix[w] = new double[4];
    for(int n= 0; n< 4; n++)
      pMatrix[w][n] = 0;
  }
  // extract the contribution of the hard and non-hard pair counts
  // and apply the hard/non-hard scaling factor
  double Nhard, Nnonhard, N; 
  // non hard counts [0][1] = counts of conserved/non-conserved non-hard pairs
  double * NonHard = new double[2]; 
  for(int w = 0; w < *_pW; w++)
  { //cerr << "w=" << w << "   ";
    Nhard = 0; 
    N = 0; // should more or less be the number of genes
    NonHard[0] = 0; NonHard[1] = 0;
    for(int n1 = 0; n1 < 4; n1++)
    { 
      for(int n2=0; n2< 4; n2++)
      { 
        // get the hard counts (that have mutated to/from the same nt as in known end
        if ( (*pSite)[w] == n1)
        { if (bknownParent) Nhard += pReducedCounts[w][n1][n2];
          else Nhard += pReducedCounts[w][n2][n1];
        }
        else // get the non-hard counts
        { if (n1 == n2) // conserved
            NonHard[0] += pReducedCounts[w][n1][n2];
          else // non-conserved
          {
            if (bknownParent)
            { NonHard[1] += pReducedCounts[w][n1][n2];}
            else 
            { NonHard[1] += pReducedCounts[w][n2][n1];}
          }
        }
      }
    }
    N = Nhard + NonHard[0]+NonHard[1]; // ! N may be zero if nothing in reduced genes
//cerr << "N=Nhard+NonHard[0]+NonHard[1] = " << N <<"/"<< Nhard <<"/"<<NonHard[0]<<"/"<<NonHard[1] << endl;
    // compute a scaling factor for non-hard counts
    if (N != 0 && N != Nhard) 
    { Nnonhard = sqrt(NonHard[0]+NonHard[1]); 
      Nnonhard /= sqrt(N);
      Nnonhard /= (NonHard[0]+3*NonHard[1]);
    }
    else Nnonhard = 0;
//  cerr << "Mnonhard=" << Nnonhard << endl;
    // add the scaled reduced count into pMatrix
    if (N == 0) N = 1; // debug 2013/06/18
    for(int n1 = 0; n1 < 4; n1++)
    { 
      if (Nhard != 0)
      {
        // hard counts
        if (bknownParent) 
          pMatrix[w][n1] += (pReducedCounts[w][(*pSite)[w]][n1])/sqrt(Nhard);
        else  
          pMatrix[w][n1] += (pReducedCounts[w][n1][(*pSite)[w]])/sqrt(Nhard);
      }
      if (Nnonhard != 0)
      {
        // non-hard counts
        if (n1 == (*pSite)[w]) // conserved contribution
          pMatrix[w][n1] += NonHard[0]*Nnonhard;
        else // non-conserved contribution
          pMatrix[w][n1] += NonHard[1]*Nnonhard;
      }
      // add the contribution of the pseudo-prior
      if (bknownParent) 
        pMatrix[w][n1] += _priorScale * (pPriorCounts[(*pSite)[w]][n1])/sqrt(N);
      else  
        pMatrix[w][n1] += _priorScale * (pPriorCounts[n1][(*pSite)[w]])/sqrt(N);
    }
  }
/*
	cerr << "matrix:" << endl;
  for(int w = 0; w < *_pW; w++)
  {
    for(int n= 0; n< 4; n++)
      cerr << pMatrix[w][n] <<"\t";
	cerr << endl;
  }
*/
  PWM * pReturnPWM = new PWM(*_pW, pMatrix);
  // cleanup
  for(int w = 0; w < *_pW; w++)
  { delete []pMatrix[w];}
  delete[]pMatrix; pMatrix = NULL;
  //cerr << "Debug-NOrthoGeneObject::::StoreEvolprobs -end" << endl;
  return pReturnPWM;
}

/******************************************************************************
  Description:  UpdateScoresStep : update motif scores for all segments 
	  for a given sequence (seqIter)
  Author_Date:  MC_2014/01/16
******************************************************************************/
void 
NOrthoGeneObject::UpdateScoresStep(SeqIterator Iter, PWM * pMotifModel)
{
  //cerr << "Debug-NOrthoGeneObject::UpdateScoresStep -begin" << endl;

  dualSequenceComputation * pComp = NULL;
  // one or both strands
  for(int i = 0; i < (int)_strand.size(); i++)
  {
    pComp = (*(*Iter).second)[i];
    pComp->UpdateSegmentScores(pMotifModel);
  }
  //cerr << "Debug-NOrthoGeneObject::UpdateScoresStep -end" << endl;
  return;
}



