#include "NOrthoSamplerRun.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
NOrthoSamplerRun::NOrthoSamplerRun():SamplerRun()
{
  //cerr << "--debug::NOrthoSamplerRun-constructor - begin" << endl;
  //cerr << "--debug::NOrthoSamplerRun-constructor - end" << endl;
}

/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2015/02/15
******************************************************************************/
NOrthoSamplerRun::~NOrthoSamplerRun()
{
  //cerr << "destructor NOrthoSamplerRun - begin" << endl;
  //cerr << "destructor NOrthoSamplerRun - end" << endl;
}
/******************************************************************************
  Description:  AddGene : add gene if not yet available, return corresponding
                GeneObject
  Author_Date:  MC_2015/02/15
******************************************************************************/
GeneObject *
NOrthoSamplerRun::AddGene(string * gene)
{
  //cerr << "--debug::NOrthoSamplerRun-AddGene - begin" << endl;
  GeneObject * Second = NULL;
  // check if gene is already present
  if (_pMap->find(*gene) == _pMap->end())
  {
    Second = new NOrthoGeneObject(_bMinusStrand);
    (*_pMap)[*gene] = Second;
    _nbrGenes++;
  }
  else // there will follow sequences for this gene
    Second = (*_pMap)[*gene];
  //cerr << "--debug::NOrthoSamplerRun-AddGene - end" << endl;
  return Second; // is never NULL;
}

/******************************************************************************
  Description:  InitializeLocalVariables2()
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoSamplerRun::InitializeLocalVariables2(const double &slowrate, const int &priorscale)
{
  //cerr << "debug::NOrthoSamplerRun::InitializeLocalVariables2 -begin" << endl;
 
  // 3. initialize local variables in each geneobject
  // *************************************************
  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
  { 
    // northoSequenceComputation objects
    // includes updating of the segment bg-scores
//cerr << "#InitializeComputationTools for " << (*iter).first << endl;
    (*iter).second->InitializeComputationTools(_pSpeciesMap);
    // initialize InstanceMap
    (*iter).second->InitializeInstanceTools();
    // _pPriorEvolMatrices (only for 'valid' branches)
//cerr << "#MsM and node-SNF for " << (*iter).first << endl;
    (*iter).second->ComputePriorCounts(_pSpeciesMap, slowrate, priorscale);
  }
  //cerr << "debug::NOrthoSamplerRun::InitializeLocalVariables2 -end" << endl;
  return;
}
/******************************************************************************
  Description:  InitInstanceMap :  for each geneobject
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
NOrthoSamplerRun::InitInstanceMap()
{ 
  //cerr << "--debug::NOrthoSamplerRun::InitInstanceMap -begin" << endl;

  for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
  { 
//cerr << "gene-" << (*iter).first << endl;
    // repeat for each geneObject 
    (*iter).second->InitSpeciesInstanceMap();
    (*iter).second->InitNodeInstanceMap(_pSpeciesMap);
  }

  /*cerr << "InitIntsanceMap Intermediate print" << endl;
	  GFFWriter * pScreen = new GFFWriter;
	  PrintMotifInstances(pScreen);
		  PrintNodeInstances(pScreen);
  delete pScreen; pScreen = NULL;
*/
  //cerr << "--debug::NOrthoSamplerRun::InitInstanceMap -end" << endl;
  return; 
}
/******************************************************************************
  Description:  DetectionStep : for nbrIter iterations, 
                optimize x instances in each sequence and each node, 
                using phylogenetic information and random sampling
                repeat for each gene
  Author_Date:  MC_2014/02/15 
******************************************************************************/
void
NOrthoSamplerRun::DetectionStep(const string &step, const int &nbrIter)
{
//cerr << "--debug::NOrthoSamplerRun::DetectionStep - begin" << endl;
  string * id(NULL), *id2(NULL); bool bSpecies; bool bParent;
  PWM * pTempMotif(NULL);
  double*** pReducedCounts = NULL;
  vector<string> * pFamily(NULL); 
  vector<PWM*> * pBranchPWMs = NULL;
  vector<double> * pBranchPSPs = NULL;
  DoubleMap * pGeneProxMap = NULL;
  int maxShift = 3;
  int shiftTime = 15;
  InstanceMap *pInstMap = NULL;
  InstanceIterator iii;
  bool HasInstance;
	
//cerr << "step=" << step << " ";
	
  // repeat motif detection for # iterations
  for(int n = 0; n < nbrIter; n++)
  { 
//cerr << "i " << n+1 << " "; // to follow up progress when running
/*
	  cerr << "Intermediate print" << endl;
	  GFFWriter * pScreen = new GFFWriter;
	  PrintNodeInstances(pScreen);
	  PrintMotifInstances(pScreen);
	  delete pScreen; pScreen = NULL;
*/
    // periodically shift instanceMap if IC score is better
    if (step == "core" && maxShift > 0 && (n % shiftTime) == 0 ) 
      ShiftFullInstanceMap(maxShift);

    // repeat for each gene
    for (iter =_pMap->begin(); iter != _pMap->end(); iter++)
    {  
//cerr << "processing gene : " << (*iter).first << endl;
      pGeneProxMap = (*iter).second->GetProxMap();
      // scan through tree top-down
		// mark : no psp impact for "nodes" (only for sequences)
      for (int node = 0; node < int(_pSpeciesMap->NbrNodes()); node++)
      { 
        // create the node id 
        id = new string("N");
        ostringstream ostr; ostr << node; id->append(ostr.str());
  //cerr << "processing node : " << *id << endl;

        // skip nodeid if not existing for this gene
        if ( *id != "N0" && pGeneProxMap->find(*id) == pGeneProxMap->end())
        {continue;}

        // first remove the existing node-instances
        (*iter).second->GetNodeInstances()->RemoveInstanceFromSequence(id);

        // for root: build the reduced instances PWM
        if (*id == "N0")
        { pTempMotif = BuildReducedRootMotif((*iter).first);}
        else // start with a uniform PWM
        { pTempMotif = new PWM(*_pWidth);} 
//cerr << "start with " << pTempMotif->ConsensusScore() << "/" << *(pTempMotif->GetConsensus()) << endl;

        // update pTempMotif with evolution in all direct branches
        pFamily = (*(_pSpeciesMap->GetMap()->find(*id))).second.pRelations;
        pBranchPWMs = new vector<PWM*>;
        pBranchPSPs = new vector<double>; 

        for( int d = 0; d < int(pFamily->size()); d++)
        { 
          bParent = 0; bSpecies = 0;
          if (d == 0 && *id != "N0") { bParent = 1;}

          // get the id2 of the known branch-end (parent or daughter)
          id2 = new string((*pFamily)[d]);
//cerr << "processing branch-end : " << *id2 << endl;

          if ( !bParent)
          { // 2014/01/19: skip if this branch does not exist in this gene
            // i.e. one of the daughters may not exist
            if ( pGeneProxMap->find(*id2) == pGeneProxMap->end())
            { continue;}

            // verify if the daughter is a species or a node
            if ((*(_pSpeciesMap->GetMap()->find(*id2))).second.pBgM != NULL)
            { bSpecies = 1; HasInstance = false;
              // 2014/01/19: 
              // skip if a daughter-species does not have instances at this stage
              // (a node daughter will always have one instance)
              pInstMap = (*iter).second->GetSpeciesInstances();
              for(iii = pInstMap->begin(); iii != pInstMap->end(); iii++)
              {
                if ((*((*iii)->PrintSeqName())) == *id2 ) 
                { HasInstance = true; break;}
              }
              pInstMap = NULL; // reset
              if (! HasInstance)
              { continue;} // skip further processing of this branch
            }
          }

          // count the nucleotide pairs in the reduced genes
          if (bParent) // known end id2 is a parent-node
          { //cerr << "debug 10 " << endl;
            pReducedCounts = CountReducedEvolution((*iter).first, id2, id, 0);
            // build/store evolutionPWM based on the given known branch-end
            // (extract correct hard & non-hard counts and apply prior-pseudo-counts)
            // also store the daughter-site-psp (to weigh branch-contributions)
            (*iter).second->StoreEvolPWM(id2, 0, id ,pReducedCounts, pBranchPWMs, pBranchPSPs);
          }
          else // known end id2 is a daughter(node or species)
          { //cerr << "debug 11 " << endl;
            pReducedCounts = CountReducedEvolution((*iter).first, id, id2, bSpecies);
            // build/store evolutionPWM based on the given known branch-end
            // (extract correct hard/non hard counts and apply prior-pseudo-counts)
            // also store the daughter-site-psp (to weigh branch-contributions)
           (*iter).second->StoreEvolPWM(id2, bSpecies, id2, pReducedCounts, pBranchPWMs, pBranchPSPs);
          }

          // cleanup
          delete id2; id2 = NULL;
          for(int w = 0; w < *_pWidth; w++)
          { for(int n1 = 0; n1 < 4; n1++)
              delete[]pReducedCounts[w][n1];
            delete[]pReducedCounts[w];
          } delete[]pReducedCounts; pReducedCounts = NULL;
        }
        pFamily = NULL;

        // now add all the branch-Evol-PWMs to the pTempMotif
        pTempMotif = pTempMotif->ApplyEvolPWMs(pBranchPWMs, pBranchPSPs);
//cerr << "sample from " << pTempMotif->ConsensusScore() << "/" << *(pTempMotif->GetConsensus()) << endl;

        // sample/select one new node instance
        Sequence * NewSite = NULL;
        if (step != "conv") NewSite = pTempMotif->SampleOneInstance();
        else NewSite = pTempMotif->SelectBestInstance();
        Instance * pInstance = new Instance(NewSite, id);
//cerr << "newsite = " << *pInstance->PrintSite() << endl;
        ((*iter).second->GetNodeInstances())->AddInstance(pInstance);
		  
        // and cleanup 
        delete id; id = NULL; 
        delete pTempMotif; pTempMotif = NULL;
        pInstance = NULL;
        for (int i = 0; i < (int)pBranchPWMs->size(); i++)
        {delete (*pBranchPWMs)[i]; (*pBranchPWMs)[i] = NULL;} // delete PWMs
        delete pBranchPWMs; pBranchPWMs = NULL;
        delete pBranchPSPs; pBranchPSPs = NULL;
      }

      // finish off with the sequences(species) at the bottom of the tree
      SequenceMap *pSeqs = (*iter).second->GetSequences();
      SeqIterator seqIter;
      for (seqIter = pSeqs->begin(); seqIter != pSeqs->end(); seqIter++)
      {
        // skip non-existing sequences for this gene
        if ( (*seqIter).second == NULL) // NorthoSeqComp does not exist
        { continue;} 

        id2 = (*seqIter).first->GetID(); // species-id sequence
//cerr << "processing species : " << *id2 << endl;

        if ( pGeneProxMap->find(*id2) == pGeneProxMap->end())
        { // can only be not-found if sequence does not exist 
          // but then should have been skipped by NorthoSeqComp==NULL
          cerr << "--Error::NOSamplerRun::DetectionStep, something wrong in variable setup--- debug! Exit." << endl;
          exit(-1);
        }

        // first remove the existing species-instance(s)
        (*iter).second->GetSpeciesInstances()->RemoveInstanceFromSequence(id2);
        // start with uniform PWM
        pTempMotif = new PWM(*_pWidth);
        // get the parent node id 
        id = new string((*((*(_pSpeciesMap->GetMap()->find(*id2))).second.pRelations))[0]); 
//cerr << "start with " << pTempMotif->ConsensusScore() << "/" << *(pTempMotif->GetConsensus()) << endl;

        pBranchPWMs = new vector<PWM*>;
        pBranchPSPs = new vector<double>; 
        // count in the reduced evolution pairs
        pReducedCounts = CountReducedEvolution((*iter).first, id, id2, 1);
        // build/store evolutionPWM based on the given known branch-end
        // (extract correct hard/non hard counts and apply prior-pseudo-counts)
        // also store the daughter-site-psp (to weigh branch-contributions)
        (*iter).second->StoreEvolPWM(id, 0, id2 ,pReducedCounts, pBranchPWMs, pBranchPSPs);

		  // now add the branch-Evol-PWM to the pTempMotif
        pTempMotif = pTempMotif->ApplyEvolPWMs(pBranchPWMs, pBranchPSPs);
//cerr << "sample with " << pTempMotif->ConsensusScore() << "/" << *(pTempMotif->GetConsensus()) << endl;

        // cleanup
        delete id; id = NULL;
        for(int w = 0; w < *_pWidth; w++)
        { for(int n1 = 0; n1 < 4; n1++)
            delete[]pReducedCounts[w][n1];
          delete[]pReducedCounts[w];
        } delete[] pReducedCounts; pReducedCounts = NULL;
        for (int i = 0; i < (int)pBranchPWMs->size(); i++)
        {delete (*pBranchPWMs)[i]; (*pBranchPWMs)[i] = NULL;} // delete PWMs
        delete pBranchPWMs; pBranchPWMs = NULL;
        delete pBranchPSPs; pBranchPSPs = NULL;

        // 2014/02/15 (splitup in two functions so reuse in PHMS is possible)
        // use this PWM to score the segments in this sequence
        (*iter).second->UpdateScoresStep(seqIter, pTempMotif);
        // and now select the new instances in this sequence
        (*iter).second->SamplingStep(seqIter, step);
        delete pTempMotif; pTempMotif = NULL;
      }
      // check on nbr of instances
      if ((*iter).second->GetSpeciesInstances()->NumberOfInstances() < 2)
      { throw AbortSampling("not enough instances to proceed procedure");
      }
    }
  }
//cerr << "--debug::NOrthoSamplerRun::DetectionStep - end" << endl;
    //cerr << endl;
  return;
}

/******************************************************************************
  Description:  BuildReducedRootMotif(iter)
                PWM based on all (not-iter-gene/root) instances counts 
                and additional pseudocount
  Author_Date:  MC_2012/12/09
******************************************************************************/
PWM*
NOrthoSamplerRun::BuildReducedRootMotif(string gene)
{
  //cerr << "--debug::NOrthoSamplerRun::BuildReducedRootMotif-begin" << endl;
  GeneMapIterator Iter; // debug
  InstanceIterator ii;

  // new variable to sum counts
  int ** pCounts = new int*[(*_pWidth)]; 
  for (int i = 0; i < (*_pWidth); i++)
  { pCounts[i] = new int[4]; 
    for (int n = 0; n < 4; n++) pCounts[i][n] = 0; 
  } 
  // local variable
  InstanceMap *pInsts = NULL;
  Sequence *pSegment = NULL;
	
  // get the root SNF
  double * pEstSNF = NULL;
  pEstSNF = (*(_pSpeciesMap->GetMap()->find("N0"))).second.pSNF;
  if (pEstSNF == NULL)
  { cerr << "--Error--NOrthoSamplerRun::BuildReducedRootMotif: SNF not found---debug... Exit." << endl;
    exit(-1);
  }

  for (Iter=_pMap->begin();Iter!=_pMap->end(); Iter++)
  { 
    // the root exists for every gene so no need for id-check in each GeneObject
    // get the Gene-nodes-InstanceMap
    pInsts = (*Iter).second->GetNodeInstances();
    // and add the counts of the root-instances (i.e. with id == N0).
    for( ii = pInsts->begin(); ii != pInsts->end(); ii++)
    { 
      if ((*((*ii)->PrintSeqName())) == "N0")
      {
        pSegment = (*ii)->Site();
        if (pSegment != NULL)
        { for (int w = 0; w < *_pWidth; w++)
          { if ((*pSegment)[w] != -1)
            { pCounts[w][(*pSegment)[w]] += 1;}
            else 
            { for (int n = 0; n < 4; n++)
                pCounts[w][n] += 0.25;
            }
          }
        }
        pSegment = NULL;
      }
    }
    pInsts = NULL;
  }

  // apply a pseudo to cover for missed counts
  double pseudoscale = 1/sqrt(_nbrGenes);
  double * pseudo = new double[4];
  for (int i = 0; i < 4; i++)
  { pseudo[i] = pEstSNF[i] * pseudoscale;}

  // create PWM
  PWM * pReturn = new PWM(*_pWidth, pCounts, pseudo);
//pReturn->StderrPrintMatrix();

  //cleanup
  delete[] pseudo; pseudo = NULL;
  pEstSNF = NULL;
  pInsts = NULL; pSegment = NULL;
  //cerr << "--debug::NOrthoSamplerRun::BuildReducedRootMotif-end" << endl;
  return pReturn; 
}

/******************************************************************************
  Description:  CountReducedEvolution, applies on one branch parent-daughter
                count the nucleotide pairs in the reduced map (w*[4*4])
  // a check is done if the branch in reduced genes does exist,
  // and if it has instances 
  Author_Date:  MC_2014/01/19
******************************************************************************/
double *** 
NOrthoSamplerRun::CountReducedEvolution(string gene, string * parentid, 
						string * daughterid, bool bSpecies)
{
//cerr << "--debug::NOrthoSamplerRun::CountReducedEvolution-begin" << endl;
 
  // local variable
  Sequence *pSite1(NULL), *pSite2(NULL);
  double weight;
  GeneMapIterator genIter;
  InstanceMap * pInstsParent = NULL; InstanceIterator iParent;
  InstanceMap * pInstsDaughter = NULL; InstanceIterator iDaughter;

  // new variable to sum counts
  double *** pCounts = new double**[*_pWidth]; 
  for (int j = 0; j < (*_pWidth); j++)
  { pCounts[j] = new double*[4]; 
    for (int n1 = 0; n1 < 4; n1++) 
    { pCounts[j][n1] = new double[4];
      for (int n2 = 0; n2 < 4; n2++) 
        pCounts[j][n1][n2] = 0;
    }
  }
  // FIRST get the max-psp present to act as norm-factor in the counts
  double maxpsp = -1;
  if (!bSpecies) 
  { maxpsp = 1;} // no psp-differentiation applicable for nodes
  else
  { // iterate through daughter(bSpecies=1)-instances in reduced genes
    for (genIter =_pMap->begin(); genIter!=_pMap->end(); genIter++)
    { if ( (*genIter).first != gene) // skip given gene
      { 
        pInstsDaughter = (*genIter).second->GetSpeciesInstances();
        for (iDaughter = pInstsDaughter->begin(); 
             iDaughter != pInstsDaughter->end(); iDaughter++)
        { // get the right daughter-instance (there may be none or more than one)
          if ( (*((*iDaughter)->PrintSeqName())) == *daughterid )
          { // get the psp of this instance
            if ((*iDaughter)->Strand() == plus_strand)
              weight = (*((*((*genIter).second->GetSequences()->find(
				(*iDaughter)->ParentSequence()))).second))[0]->
				GetPSPAt((*iDaughter)->Start());
            else // minus strand
              weight = (*((*((*genIter).second->GetSequences()->find(
				(*iDaughter)->ParentSequence()))).second))[1]->
				GetPSPAt((*iDaughter)->Start());
            // track the highest psp in the branch in the reduced dataset
            if (weight > maxpsp) {maxpsp = weight;}
          }
        }
      }
    }
  }
  if (maxpsp == -1)
  { cerr << "--Error--NOrthoSamplerRun::CountReducedEvolutiion: maxpsp cannot be -1 ---debug... Exit." << endl;
    exit(-1);
  }
          
  // now count each of the parent-daughter nt-pairs in the reduced gene-set
  for (genIter =_pMap->begin(); genIter!=_pMap->end(); genIter++)
  { 
    if ( (*genIter).first != gene) // skip given gene
    {
//cerr << (*genIter).first << endl;
      // get the parent instanceMap
      pInstsParent = (*genIter).second->GetNodeInstances();
      // get the daughter instance(s)Map
      if (bSpecies) {pInstsDaughter = (*genIter).second->GetSpeciesInstances();}
      else { pInstsDaughter = (*genIter).second->GetNodeInstances();}

      // get the parent instance in this gene 
      // (!may be none in one of the reduced genes if the branch does not exist)
      for (iParent = pInstsParent->begin(); iParent != pInstsParent->end(); iParent++)
      {
        if ( (*((*iParent)->PrintSeqName())) == *parentid )
        {
//cerr << "site1=" << *(*iParent)->PrintSite() << endl;
          pSite1 = (*iParent)->Site();

          // get the daughter instance in this gene 
          // (!also here " it may be none or more than one)
          for (iDaughter = pInstsDaughter->begin(); iDaughter != pInstsDaughter->end(); iDaughter++)
          {
            if ( (*((*iDaughter)->PrintSeqName())) == *daughterid )
            {
//cerr << "site2=" << *(*iDaughter)->PrintSite() << endl;
              pSite2 = (*iDaughter)->Site();
              // add nucleotide pair count for parent-daughter in the given gene
              // first get the psp
              weight = 1;
              if (bSpecies)
              {
                if ((*iDaughter)->Strand() == plus_strand)
                  weight = (*((*((*genIter).second->GetSequences()->find((*iDaughter)->
							ParentSequence()))).second))[0]->GetPSPAt((*iDaughter)->Start());
                else // minus strand
                  weight = (*((*((*genIter).second->GetSequences()->find((*iDaughter)->
							ParentSequence()))).second))[1]->GetPSPAt((*iDaughter)->Start());
              }
              // add the (norm-weighted) pair-count
              for(int w = 0; w < *_pWidth; w++)
                pCounts[w][(*pSite1)[w]][(*pSite2)[w]] += (weight/maxpsp);
            }
            // there may be other species-daughter-instances to be counted
          }
          break; // node-instances always have maximum one instance per gene
        }
        // do nothing if the branch is 'not existing' in this particular gene
      }
    }
  }

/*
	cerr << "debug-NOrthoSamplerRun::CountReducedEvolution" << endl;
	cerr << "Print counts:" << endl;
	for (int i = 0; i < (*_pWidth); i++)
	{ cerr << "position=" << i << " ::"; 
      for (int n1 = 0; n1 < 4; n1++) 
      { cerr <<"(n1=" << n1 << "):";
        for (int n2 = 0; n2 < 4; n2++) 
          cerr << pCounts[i][n1][n2] << " ";
	  }
	  cerr << endl;
    } 
*/

  pInstsParent = NULL; pInstsDaughter = NULL; pSite1 = NULL; pSite2 = NULL;
//cerr << "--debug::NOrthoSamplerRun::CountReducedEvolution-end" << endl;
  return pCounts;
}
/******************************************************************************
  Description:  PrintNodeInstances : list of individual instances
  Author_Date:  MC_2013/06/18
******************************************************************************/
void 
NOrthoSamplerRun::PrintNodeInstances(GFFWriter * gff)
{
  //cerr << "--debug::NOrthoSamplerRun::PrintNodeInstances-begin" << endl;
  ostringstream text;
  InstanceMap * instMap; InstanceIterator i;
  GeneMapIterator iiiter; 

  // report instances for each GeneObject
  for (iiiter =_pMap->begin(); iiiter != _pMap->end(); iiiter++)
  {
    // get the instances for this gene
    instMap = (*iiiter).second->GetNodeInstances();
    // print each instance
    for(i = instMap->begin(); i != instMap->end(); i++)
    {
      text << (*iiiter).first << "-"; // genid
      text << *((*i)->PrintSeqName()); //node id
      // add instance-code
      text << "\tsite " << *((*i)->PrintSite()) << endl;
    }
  }
  gff->AddComment(text.str());
  text.flush();
  //cerr << "--debug::NOrthoSamplerRun::PrintNodeInstances-end" << endl;
  return;
}
