#include "PhyloGeneObject.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
PhyloGeneObject::PhyloGeneObject(const bool &bMinus):GeneObject(bMinus)
{
  //cerr << "Debug-phyloGeneObject::PhyloGeneObject -begin" << endl;
  _pNtWeights = NULL; 
  _pInfo = NULL;
  _dY = 0; 
  _pACGTsamples = NULL;
  _phylo_2 = 0;
  //cerr << "Debug-phyloGeneObject::PhyloGeneObject- end" << endl;
}
/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2009/01/25
******************************************************************************/
PhyloGeneObject::~PhyloGeneObject()
{
  //cerr << "Debug-GeneObject::~GeneObject- begin" << endl;
  // _pNtWeights
  if (_pNtWeights != NULL)
  {
    for (int w = 0; w < *_pW; w++)
      delete[]_pNtWeights[w];
    delete[] _pNtWeights;
  }
  _pNtWeights = NULL;

  // delete the pInfo container
  if (_pInfo != NULL)
  {
    for (int w = 0; w < *_pW; w++)
    { for (int n = 0; n < 4; n++)
      { delete[] _pInfo[w][n][0];
        delete[] _pInfo[w][n][1];
        delete[] _pInfo[w][n];
      }
      delete[] _pInfo[w];
    }
    delete _pInfo;
  }
  _pInfo = NULL;

  // delete _pACGTsamples
  if (_pACGTsamples != NULL)
  {
    ACGTiter = _pACGTsamples->begin();
    for(; ACGTiter != _pACGTsamples->end(); ACGTiter++)
    {
      if ((*ACGTiter) != NULL)
        delete[] (*ACGTiter); 
      (*ACGTiter) = NULL; 

    }
    delete _pACGTsamples;
  }
  _pACGTsamples = NULL;
  //cerr << "Debug-PhyloGeneObject::~PhyloGeneObject- end" << endl;
}

/******************************************************************************
  Description:  InitializeInstanceTools() : _pInstances  & _pNtWeights 
  Author_Date:  MC_2007/12/20
******************************************************************************/
void 
PhyloGeneObject::InitializeInstanceTools()
{  
  //cerr << "Debug-PhyloGeneObject::::InitializeInstanceTools -begin" << endl;

 // empty InstanceMap
  _pSpeciesInstances = new InstanceMap(_pW); 
  // empty _pNtWeights
  _pNtWeights = new double*[*_pW]; 
  for (int w = 0; w < *_pW; w++)
  {
    _pNtWeights[w] = new double[4]; 
    for (int n = 0; n < 4; n++)
      _pNtWeights[w][n] = 0; 
  }
  //cerr << "Debug-PhyloGeneObject::::InitializeInstanceTools -end" << endl;
  return;
}
/******************************************************************************
  Description:  InitializeComputationTools 
                create phyloSequenceComputation objects
	and _pInfo, and _pACGTsamples and set _dY
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
PhyloGeneObject::InitializeComputationTools(SpeciesMap * spMap,
											const double &y, const double &z,
											const bool &phylo2)
{
  //cerr << "Debug-PhyloGeneObject::::InitializeComputationTools -begin" << endl;
  int M(0);
  _phylo_2 = phylo2;
  map<string, speciesgeg>* pMap = spMap->GetMap();
  // initialize local variables for computations of NtWeights
  _dY = y; 
  // _pInfo special container wxnx2xnbrQ's
  _pInfo = new double***[*_pW];
  for (int w = 0; w < *_pW; w++)
  { _pInfo[w] = new double**[4]; 
    for (int n = 0; n < 4; n++)
    { _pInfo[w][n] = new double*[2]; 
      _pInfo[w][n][0] = NULL; // to store prox's 
      _pInfo[w][n][1] = NULL; // to store psp's
    }
  }
  // generate samples for non-conserved computation
  _pACGTsamples = new list<double*>;
  double freqA, freqC, freqG; 
  for (freqA = z; freqA <= (1.0001-3*z); (freqA=freqA+z))
  {
    for (freqC = z; freqC <= (1.0001-freqA-2*z); (freqC=freqC+z))
    {
      for (freqG = z; freqG <= (1.0001-freqA-freqC-z); (freqG=freqG+z))
      {
        double * pACGT = new double[4];
        pACGT[0] = freqA; 
        pACGT[1] = freqC;
        pACGT[2] = freqG;
        pACGT[3] = 1-freqA-freqC-freqG;
        _pACGTsamples->push_back(pACGT);
      }
    }
  }
  // create computation objects 
  // only do this in case -M > 0 and SeqL > motif width
  // if not, then leave phyloSequenceComputation object to NULL 
// so the objects are not removed as such when sequences are not suitable for motif detection, 
// but by setting NULL, they will not be assessed throughout motif detection
// ! check for NULL on phyloSequenceComputation must be implemented further on !!!!!!!!!!!!!!!
  vector<dualSequenceComputation *> *seqComp = NULL;
  phyloSequenceComputation * seqCompObj = NULL;
  string *species = NULL;
  for(seqIter = _pSequences->begin(); seqIter!=_pSequences->end(); seqIter++)
  {
    species = (*seqIter).first->GetID();
    M = 0;
    // check if -M is not zero for this species
    if ((*pMap)[*species].pNbrInstDistr != NULL)
    {
      if ((*(*pMap)[*species].pNbrInstDistr)[0] != NULL) 
        M = (int)(*(*pMap)[*species].pNbrInstDistr)[0]->Size()-1;
      else M = (int) (*pMap)[*species].pNbrInstDistr->size()-1;
    }
    // and the fasta sequences is long enough
    if( (*seqIter).first->Length() <= *_pW )
    { // too short sequence, falls back to M = 0
      M = 0;
    }
    if (M == 0)
    { // do not create computation object
      continue;
    }
    // create phyloSequenceComputation-vector
    seqComp = new vector<dualSequenceComputation *>;
    // link seqobj, width, proximity, priorDistr
    seqCompObj = new phyloSequenceComputation( plus_strand, 
                       (*seqIter).first, _pW, (*_pProximityMap)[*species], 
                       (*pMap)[*species].pNbrInstDistr, phylo2 );
    // compute segment/sequence bg scores
    seqCompObj->ComputeBgScores((*pMap)[*species].pBgM);
    // add 
    seqComp->push_back(seqCompObj);
    if ((int)_strand.size() == 2)
    { seqCompObj = new phyloSequenceComputation(minus_strand, 
                         (*seqIter).first, _pW, (*_pProximityMap)[*species], 
                         (*pMap)[*species].pNbrInstDistr, phylo2 );
      seqCompObj->ComputeBgScores((*pMap)[*species].pBgM);
      seqComp->push_back(seqCompObj);//(*seqIter).second[1] = seqComp[1];
    }
    (*seqIter).second = seqComp;
    // reset 
    seqComp = NULL;
    seqCompObj = NULL;
    species = NULL;
  }
  //cerr << "Debug-PhyloGeneObject::InitializeComputationTools -end" << endl;
  return;  
}

/******************************************************************************
  Description: UpdateScoresStep(SeqIterator Iter, double** redgen); 
  update motif scores for all segments for a given sequence (seqIter)
  Author_Date:  MC_2014/02/17
******************************************************************************/
void
PhyloGeneObject::UpdateScoresStep(SeqIterator Iter, double** redgen)
{
  //cerr << "Debug-PhyloGeneObject::UpdateScoresStep -begin" << endl;
  double ** pRedSeq = NULL;
  // clear existing instances
  _pSpeciesInstances = _pSpeciesInstances->RemoveInstanceFromSequence((*Iter).first);

  // 2014/02/28
  // also compute the background probability of the reduced instances set
  // logaritmic score
  double redSeqLogBackground = ComputeReducedInstBackground();

  // Compose extern2-variable : gene excluding the current sequence
  pRedSeq = ComputeReducedInstMutProb(redgen); // redgen = extern1
  // compute probability of the red-sequences-set score (logaritmic!)
  // will be applied as denumerator (fixed for all segments of a given sequence,
  // but different denumerator over different sequences)
  double redSeqLogDenum = 0;   double scoreW;
  for (int w = 0; w < *_pW; w++)
  { scoreW = 0;
    // sum over parent nucleotides A,C,G,T (pRedSeq consists of logs)
    for(int n = 0; n < 4; n++) scoreW += exp(pRedSeq[w][n]);
    // multiply (~ add log) over motif width positions
    redSeqLogDenum += log(scoreW);
  }

  dualSequenceComputation * comp = NULL;
  // update the scores _pExpWx in the given sequence
  for(int i = 0; i < (int)_strand.size(); i++)
  {
    comp = (*(*Iter).second)[i];
    // Update the local _mutationProb in phyloSequenceComputation 
    comp->ComputeBasicMutProb(redgen);// redgen = extern1
    // Update the local _redinstBg in phyloSequenceComputation 
    comp->StoreRedInstBg(redSeqLogBackground);// = extern"2" for BG

    // update motif scores
    comp->UpdateMotifScores(pRedSeq); // pRedSeq = extern2
    // apply mask, segment-bg and redSeqLogDenum(fixed) contributions
    comp->UpdatepExpWx(redSeqLogDenum); 
  }
  
  // cleanup
  for (int w = 0; w < *_pW; w++)
    delete[] pRedSeq[w];
  delete[] pRedSeq; pRedSeq = NULL;
  //cerr << "Debug-PhyloGeneObject::UpdateScoresStep -end" << endl;
  return;
}

/******************************************************************************
  Description:  UpdateNtWeights() : based on instances in _pSpeciesInstances
  // default psp and phylo correction always applicable (no user-choise)
     // mark : psp could be uniform and overall have no effect
  Author_Date:  MC_2014/02/15 
******************************************************************************/
void 
PhyloGeneObject::UpdateNtWeights()
{
  //cerr << "Debug-PhyloGeneObject::UpdateNtWeights -begin" << endl;

  // first update the (prox,psp) information on each position w 
  // and nucleotide n in the local 'work-variable'
  UpdateProxPSPinLocalpInfoVariable();

  // compute the "average" psp-scores for each nucleotide on each position
  // store it on _pInfo[w][n][1][0]
  double * avg = new double[4]; double max_avg;
  for (int w = 0; w < *_pW; w++)
  { max_avg = -1;
    for (int n = 0; n < 4; n++)
    { avg[n] = 0;
      for(int i = 1; i <= _pInfo[w][n][1][0]; i++)
      { avg[n] += _pInfo[w][n][1][i];}
      avg[n] /= _pInfo[w][n][1][0];
      if (avg[n] > max_avg){max_avg = avg[n];}
    }
    // normalize with the max_avg 
    // so on each position one avg_psp value will be 1, other three lower
    for (int n = 0; n < 4; n++)
    { avg[n] /= max_avg;
    // store the computed psp_avg now in the first pointer
    // the "nbr_psp's" is overwritten now
      _pInfo[w][n][1][0] = avg[n];
    }
  } 
  delete[] avg; avg = NULL;

  // now compute the weights on each postion the motif
  double pNonCons; 
  for (int w = 0; w < *_pW; w++)
  {
    // reset existing weights  
    for (int n = 0; n < 4; n++) {_pNtWeights[w][n] = 0;}
    // first compute the conserved weigths for acgt based on q's (no psp's)
    // store the results already in _pNtWeights[w][n] 
    for (int n = 0; n < 4; n++)
    { ComputeConservedWeight(w,n);
//cerr << "w=" << w << "-conserved weight[" << n << "]=" << _pNtWeights[w][n] << endl;
	}
    // now compute the non conserved weight
    pNonCons = ComputeNonConservedWeight(w);
//cerr << "non-conserved weight = " << pNonCons << endl;
    // multiply non-conserved with conserved weight
    for (int n = 0; n <4; n++) 
    { if (_pNtWeights[w][n] > 0) // else _pInfo[w][n]=NULL (no [1][0] accessible)
      {
        _pNtWeights[w][n] *= pNonCons;
        // PSP correction : 
        // weigh each _pNtWeight for ACGT with the nucleotide-averaged-psp
        _pNtWeights[w][n] *= _pInfo[w][n][1][0]; // a value < or == 1.
      }
//cerr << "final weight["<<w<<"]["<<n<<"] = " << _pNtWeights[w][n] << endl;
    } 
  }
  //cerr << "Debug-PhyloGeneObject::UpdateNtWeights -end" << endl;
  return; 
}
/******************************************************************************
  Description:  UpdateProxPSPinLocalpInfoVariable
                i.e. update the (prox,psp) information on each position w 
                and nucleotide n in _pInfo
                based on instances in _pSpeciesInstances
  Author_Date:  MC_2014/02/15 
******************************************************************************/
void 
PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable()
{
//cerr << "Debug-PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable() -begin" << endl;
  // update the (prox,psp) information on each position w 
  // and nucleotide n in the local 'work-variable'
  strand_modes strand;
  dualSequenceComputation *pComp = NULL;
  Sequence *pSegment = NULL;
  list<Instance *>::iterator iter;

  // get the list of all 'species' for each nt at each postion (Wxn vectors)
  StringVector*** pCounts = _pSpeciesInstances->GetInstanceCounts();
  int nbr_inst = 0;
  int nbr_species;
  for (int w = 0; w < *_pW; w++)
  {
    for (int n = 0; n < 4; n++)
    { nbr_species = int(pCounts[w][n]->size()); 
      if (w==0){nbr_inst += nbr_species;} // track the number of instances 
      if (_pInfo[w][n][0] != NULL) delete[] _pInfo[w][n][0];
      _pInfo[w][n][0] = new double[nbr_species+1]; // to store prox's 
      if (_pInfo[w][n][1] != NULL) delete[] _pInfo[w][n][1];
      _pInfo[w][n][1] = new double[nbr_species+1]; // to store psp's
      for(int i=1; i <= nbr_species; i++)
      { // store proximity and psp-score
        _pInfo[w][n][0][i] = 0; // initialize proximity
        _pInfo[w][n][1][i] = 1; // initialize psp-score
      }
      //// !!!! store nbr_species in the first field !!!!!!
      _pInfo[w][n][0][0] = (double)nbr_species; // store it twice, it is as indicator
      _pInfo[w][n][1][0] = (double)nbr_species; // of the number of following entries
    }
  }
 
/*
// print the local variable
	cerr << "printing initial _pInfo:" << endl;
	  for (int w = 0; w < *_pW; w++)
	{ for (int n = 0; n < 4; n++)
      { 
		nbr_species = int(pCounts[w][n]->size()); 
//cerr << "nbrSpecies = " << nbr_species << " ";
      //cerr << "_pInfo["<<w<<"]["<<n<<"][0][0]=" <<  _pInfo[w][n][0][0] << " ";
      cerr << "_pInfo["<<w<<"]["<<n<<"][1][0]=" <<  _pInfo[w][n][1][0] << endl;
		for(int i=0; i <= nbr_species; i++)
		{ // print proximity and psp-score
  		 //cerr << "_pInfo["<<w<<"]["<<n<<"][prox]["<<i<<"] = " << _pInfo[w][n][0][i] << endl;
  		 //cerr << "_pInfo["<<w<<"]["<<n<<"][psp]["<<i<<"] = " << _pInfo[w][n][1][i] << endl;
		}
	  }
	}
*/
  // iterate through all species instances and store in pInfo accordingly
  double psp_score; double seq_prox;
  int i; 
  for (iter = _pSpeciesInstances->begin();iter != _pSpeciesInstances->end(); iter++)
  { 
    // find the phyloSequenceCompuation object of appropiate strand
    strand = (*iter)->Strand();
    if (strand == plus_strand)
      pComp = (*FindSequence(*(*iter)->PrintSeqName()))[0];
    else 
      pComp = (*FindSequence(*(*iter)->PrintSeqName()))[1];
    // get the PSP score at the appropiate START position
    psp_score = pComp->GetPSPAt((*iter)->Start());
    // get the proximity of this sequence
    seq_prox = *pComp->GetProximity();
    // store (seq_prox, psp_score) for the observed nt on each position w
    pSegment = (*iter)->Site();
    if (pSegment != NULL)
    { for (int w = 0; w < *_pW; w++)
      { 
        if ((*pSegment)[w] != -1)
        { i = _pInfo[w][(*pSegment)[w]][0][0]; 
          if (i==0)
          { cerr << "--error0-debug-PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable(). inconsistency. Exit" << endl;
            exit(-1);
          }
          _pInfo[w][(*pSegment)[w]][0][i] = seq_prox;
          _pInfo[w][(*pSegment)[w]][1][i] = psp_score;
          // count down, and reset then at the end 
          _pInfo[w][(*pSegment)[w]][0][0] -= 1;
//cerr << "_pInfo["<<w<<"]["<<(*pSegment)[w]<<"][0][0]="<<_pInfo[w][(*pSegment)[w]][0][0]<<endl;
        }
        else
        { cerr << "--Warning--PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable(). nt=N?" << endl;}
      }
    } 
    else 
    { cerr << "--error2-debug-PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable() Inst-Site?" << endl;
      exit(-1);
    }
    i++;
  }
  // reset the 'count of q's on the [w][n][0][0] position
  for (int w = 0; w < *_pW; w++)
  { for (int n = 0; n < 4; n++)
    { if (_pInfo[w][n][0][0] != 0) 
      { cerr << "--error3-debug-PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable(). inconsistency. Exit" << endl;exit(-1);}
      //// !!!! store nbr_species in the first field !!!!!!
      _pInfo[w][n][0][0] = _pInfo[w][n][1][0];
    }
  }
  //cerr << "Debug-PhyloGeneObject::UpdateProxPSPinLocalpInfoVariable() -end" << endl;

  return;
}

/******************************************************************************
  Description:  ComputeConservedWeight(double * all_q)
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
PhyloGeneObject::ComputeConservedWeight(int W, int NT)
{
  //cerr << "Debug-PhyloGeneObject::ComputeConservedWeight -begin" << endl;
	  
	  // MARK: no psp applied here, the psp weighing is done on 'nt' level
	  // i.e. an 'average' nucleotide-psp is applied in UpdateNtWeights above

  double * all_q = _pInfo[W][NT][0];
  int nbr_prox = (int)all_q[0]; // number of q's
  if (nbr_prox < 2){_pNtWeights[W][NT] = nbr_prox; return;} //0 or 1

//cerr << "nbrprox="<< nbr_prox << endl;
  // variable to store iterative computations
  ScoreVector * pWeightForFreq = new ScoreVector;
  double term1, term2, term, numerator, denumerator;
  // first compute term which is fixed for all iterative freq's (=prod(all(1-q)))
  term2 = 0; 
  for (int s = 1; s <= nbr_prox; s++) 
  { term2 += log( (1- all_q[s]) ); }

  // now compute weight for each of iterative frequencies (for a given deltaY)
  for (double freq = _dY; freq <= (1.0000001 -_dY); (freq = freq + _dY))
  { 
    term1 = 0; 
    for (int s = 1; s <= nbr_prox; s++)
    { term1 += log(  all_q[s] + ( freq * (1- all_q[s]))  );}
// log( freq * exp(term1) + (1-freq) * exp(term2 + nbr_prox * log(freq)) ) / log(freq)
    numerator = freq * exp(term1);
    term = exp(term2 + nbr_prox * log(freq));
    numerator += (1-freq)*term;
    numerator = log(numerator);
    denumerator = log(freq);
    term = numerator / denumerator;
    pWeightForFreq->push_back(term);
//cerr << "pWeightStored["<<freq<<"]=" << pWeightForFreq->back() << " ";
  }
  // take average of all these iterative weight(y)
  _pNtWeights[W][NT] = PHYLINCLUSIVE::SumVector(pWeightForFreq)/(int)pWeightForFreq->size(); 
  // cleanup
  delete pWeightForFreq; pWeightForFreq = NULL; 
  all_q = NULL;
  //cerr << "Debug-PhyloGeneObject::ComputeConservedWeight -end" << endl;
  return;
}

/******************************************************************************
  Description:  ComputeNonConservedWeight(double * conserved_weights, 
											double ** info_w)
  Author_Date:  MC_2014/02/15
******************************************************************************/
double 
PhyloGeneObject::ComputeNonConservedWeight(int W)
{
  //cerr << "Debug-PhyloGeneObject::ComputeNonConservedWeight -begin" << endl;
  // for now the conserved weights are in _pNtWeights[w]
  double * conserved_weights = _pNtWeights[W];
  double *** info_w = _pInfo[W];

  double non_conserved_weight = 0;
  int nbr_prox;  
  double * all_q;
  ScoreVector * pWeightForSample = new ScoreVector;
  // first compute term which is fixed for all iterative freq's (=prod(all_other(1-q)))
  // compute this for each given nucleotide
  double *term2 = new double[4];
  for (int n = 0; n < 4; n++)
  { term2[n] = 0;
    for (int nn = 0; nn < 4; nn++)
    { if (nn != n)
      { nbr_prox = (int)info_w[nn][0][0]; // number of q's for this 'other' nucleotide
        all_q = info_w[nn][0];
        for (int s = 1; s <= nbr_prox; s++)
        { term2[n] += log( (1- all_q[s]) );}
      }
    }
  }

  // now compute weight for each of iterative frequencies (samples of ACGT)
  // (for a given deltaZ)
  double term1, term3, freq, numerator, logdenumerator; 
  ACGTiter = _pACGTsamples->begin();
  for (; ACGTiter != _pACGTsamples->end(); ACGTiter++)
  { 
    numerator = 0; 
    logdenumerator = 0;
    for (int n = 0; n < 4; n++)
    { freq = (*ACGTiter)[n];
      term1 = 0; 
      nbr_prox = info_w[n][0][0]; // number of q's for this nucleotide
      all_q = info_w[n][0];
      for (int s = 1; s <= nbr_prox; s++)
      { term1 += log(  all_q[s] + ( freq * (1- all_q[s]))  );}
      term3 = 0; 
      for(int nn = 0; nn < 4; nn++)
      { if (nn != n)
          term3 += (info_w[nn][0][0] * log((*ACGTiter)[nn]));
      }
      numerator += freq * exp(term1+term2[n]+term3);
      logdenumerator += conserved_weights[n] * log(freq);
    }
    // weight
    pWeightForSample->push_back(log(numerator)/ logdenumerator);
  }
  // average
  non_conserved_weight = PHYLINCLUSIVE::SumVector(pWeightForSample)/
                          (int)pWeightForSample->size(); 

  // cleanup
  delete[] term2; term2 = NULL;
  delete pWeightForSample; pWeightForSample = NULL; 
  conserved_weights = NULL;
   info_w = NULL;
  //cerr << "Debug-PhyloGeneObejct::ComputeNonConservedWeight -end" << endl;
  return non_conserved_weight;
}
/******************************************************************************
  Description:  ComputeReducedInstMutProb: computes 'extern2'
                // logaritmic !!
     based on adjusted F81-evolution model for all instances 
     in this geneobject, excluding the sequence of interest (has been removed)
  // redPWM (=extern1) is based on instances in other geneobjects 
  // NO psp correction here (psp has been applied in redPWM)
  Author_Date:  MC_2014/02/15
******************************************************************************/
double**
PhyloGeneObject::ComputeReducedInstMutProb(Matrix redPWM)
{
  //cerr << "Debug-PhyloGeneObject::ComputeReducedInstMutProb -begin" << endl;
  // variable to be returned = extern 2
  double ** redMutProb = new double*[*_pW];
  for(int w = 0; w < *_pW; w++)
  { redMutProb[w] = new double[4];
    for(int n = 0; n < 4; n++) redMutProb[w][n] = 0;
  }

  // applying phylogenetic correction
  // first update the (prox,psp) information on each position w 
  // and nucleotide n in the local 'work-variable'
  // _pSpeciesInstances is now all instances in 'other' sequences of this gene
  UpdateProxPSPinLocalpInfoVariable();

  double freq, terms; 
  double proximity = 0; // for !_phylo_2 (i.e. no phylo correction applied but extern1)
  // now use the q's on each position and each nucleotide 
  // (mark that psp's are not used here)
  for (int w = 0; w < *_pW; w++)
  { for (int n = 0; n < 4; n++)
    {
      freq = redPWM[w][n]; 
      terms = log(freq); // ancestor
      for (int s = 1; s <= (int)_pInfo[w][n][0][0]; s++)
      { // probability in each conserved branch
        if (_phylo_2) proximity = _pInfo[w][n][0][s];
        terms += log(proximity + (freq * (1-proximity)));
      }
      for (int nn = 0; nn < 4; nn++)
      { if (nn != n)
        {
          for (int s = 1; s < (int)_pInfo[w][nn][0][0]; s++)
          { // probability in the other non-conserved branches
            if (_phylo_2) proximity = _pInfo[w][nn][0][s];
            terms += log( redPWM[w][nn] * (1-proximity)); 
          }
        }
      }
      redMutProb[w][n] = terms;
    }
  }
  //cerr << "Debug-PhyloGeneObject::ComputeReducedInstMutProb -end" << endl;
  return redMutProb; // logaritmic 
}


/******************************************************************************
  Description:  ComputeReducedSeqBackground(SeqIterator Iter)
                // logaritmic !!
     compute the background probability of the instances 
     in the "other" sequences (like extern2, but bg instead of motif prob.
  Author_Date:  MC_2014/02/28
******************************************************************************/
double
PhyloGeneObject::ComputeReducedInstBackground()
{
  //cerr << "Debug-PhyloGeneObject::ComputeReducedSeqBackground -begin" << endl;
  // variable to be returned = extern"2"-score FOR BACKGROUND
  double redBgProb = 0;  
  vector<dualSequenceComputation*> * pSeqComps = NULL;
  dualSequenceComputation* comp = NULL;
  int scale = 0;
  // iterate through _pSpeciesInstances 
  // (the "selected-sequence" instance(s) are already removed)
  for (instIter = _pSpeciesInstances->begin(); instIter != _pSpeciesInstances->end(); instIter++)
  {
    pSeqComps = FindSequence(*(*instIter)->PrintSeqName());
    for(int i = 0; i < (int)_strand.size(); i++)
    {
      comp = (*pSeqComps)[i];
      // bg-score of segment is logaritmic, add this instance-bg to redBgProb
      redBgProb += comp->GetBgScoreAt((*instIter)->Start());
      scale += 1;
    }
  }
  if (scale > 1) // experimental set
  // apply a scale else the Pr(N=0) shows to become very low (order e-74) 
  { redBgProb /= pow(scale,1.2); }// scale by the number of contributing instances

  //cerr << "Debug-PhyloGeneObject::ComputeReducedInstBackground -end" << endl;
  return redBgProb; // logaritmic 
}
