#include "PhyloSamplerRun.h"

/******************************************************************************
  Description:  constructor 
  Author_Date:  MC_2014/02/15 - inherit class
******************************************************************************/
PhyloSamplerRun::PhyloSamplerRun():SamplerRun()
{
  //cerr << "--debug::PhyloSamplerRun-constructor - begin" << endl;
  //cerr << "--debug::PhyhoSamplerRun-constructor - end" << endl;
}
/******************************************************************************
  Description:  destructor 
  Author_Date:  MC_2014/02/15 - inherit class
******************************************************************************/
PhyloSamplerRun::~PhyloSamplerRun()
{ 
  //cerr << "destructor PhyloSamplerRun - begin" << endl;
  //cerr << "destructor PhyloSamplerRun - end" << endl;
}

/******************************************************************************
  Description:  AddGene : add gene if not yet available, return corresponding
                GeneObject
  Author_Date:  MC_2014/02/15
******************************************************************************/
GeneObject *
PhyloSamplerRun::AddGene(string * gene)
{
  //cerr << "--debug::PhyloSamplerRun-AddGene - begin" << endl;
  GeneObject * Second = NULL;
  // check if gene is already present
  if (_pMap->find(*gene) == _pMap->end())
  {
    Second = new PhyloGeneObject(_bMinusStrand);
    (*_pMap)[*gene] = Second;
    _nbrGenes++;
  }
  else // there will follow sequences for this gene
    Second = (*_pMap)[*gene];
  //cerr << "--debug::PhyloSamplerRun-AddGene -end" << endl;
  return Second; // is never NULL;
}

/******************************************************************************
  Description:  InitializeLocalVariables2()
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
PhyloSamplerRun::InitializeLocalVariables2(const double &dY, const double &dZ,
										   const bool &bPhylo_2)
{
  //cerr << "--debug::PhyloSamplerRun-InitializeLocalVariables2 - begin" << endl;
  // initialize local variables in each geneobject
  // *************************************************
  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
  { // _pInstances & _pNtWeights
    (*iter).second->InitializeInstanceTools(); 
    // _pComputations
    (*iter).second->InitializeComputationTools(_pSpeciesMap, dY, dZ, bPhylo_2);
  }
  //cerr << "--debug::PhyloSamplerRun-InitializeLocalVariables2 - end" << endl;
  return;
}

/******************************************************************************
  Description:  InitInstanceMap :  for each geneobject
  Author_Date:  MC_2014/01/15
******************************************************************************/
void
PhyloSamplerRun::InitInstanceMap()
{ 
  //cerr << "--debug::PhyloSamplerRun-InitInstanceMap() - begin" << endl;
  for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
  {     // repeat for each geneObject 
    (*iter).second->InitSpeciesInstanceMap();
    // compute nucleotide weights based on instances (for reduced PWM)
    (*iter).second->UpdateNtWeights();
  }

  //cerr << "--debug::PhyloSamplerRun-InitInstanceMap() - end" << endl;
  return; 
}
/******************************************************************************
  Description:  DetectionStep : for nbrIter iterations, 
                optimize x instances in each sequence, 
                using phylogenetic information and random sampling
                repeat for each gene
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
PhyloSamplerRun::DetectionStep(const string &step, const int &nbrIter)
{
//cerr << "PhyloSamplerRun::DetectionStep - begin" << endl;
  int maxShift = 3;
  int shiftTime = 15;
  SequenceMap *pSeqs = NULL;
  SeqIterator seqIter;
  // extern1-variable (reduced_matrix from other genes)
  double** pReducedGen = NULL; // extern 1

//cerr << "step = "  << step << endl;
  // repeat motif detection for # iterations
  for(int n = 0; n < nbrIter; n++)
  { 
   //cerr << "it " << n+1 << " "; // to follow up progress when running
   // periodically shift instanceMap if IC score is better
    if (step == "core" && maxShift > 0 && (n % shiftTime) == 0 ) 
      ShiftFullInstanceMap(maxShift);

    // repeat for each gene
    for (iter =_pMap->begin(); iter != _pMap->end(); iter++)
    {  
      // build extern1-variable based on other genes
      if (step != "conv"){ pReducedGen = BuildReducedGen(iter);}
      else { // build extern1-variable based on all/other genes 
             pReducedGen = BuildConvergenceGen(iter);}

      // get all sequences of this gene
      pSeqs = (*iter).second->GetSequences();
      // repeat for each sequence of this gene
      for (seqIter = pSeqs->begin(); seqIter != pSeqs->end(); seqIter++)
      {
        // skip non-existing sequences for this gene
        if ( (*seqIter).second == NULL) // phyloSequenceComputation does not exist
        { continue;} 

        // use the extern 1 variable to update the local _mutationProb (~"intern3")
        // (_mutationProb is based on all instances in reduced-seq of this gene)
        // and use extern2 and intern3 to score the segments in each sequence of this gene
        (*iter).second->UpdateScoresStep(seqIter, pReducedGen);
        // and now select the new instances in this sequence
        (*iter).second->SamplingStep(seqIter, step);
      }
      // all sequences of this gene finished now

      // cleanup
      if (pReducedGen != NULL)
      { for(int j = 0; j < *_pWidth; j++) delete[] pReducedGen[j];
        delete[] pReducedGen; } pReducedGen = NULL;
      // check on nbr of instances
      if ((*iter).second->GetSpeciesInstances()->NumberOfInstances() < 2)
      { throw AbortSampling("not enough instances to proceed procedure");
      }

      // update Ntweights for this gene based on all new instances
      (*iter).second->UpdateNtWeights();
    }
  }
//cerr << "PhyloSamplerRun::DetectionStep - end" << endl;
  return;
}


/******************************************************************************
  Description:  BuildReducedCounts : extern1-counts based on other genes
  Author_Date:  MC_2014/02/15
******************************************************************************/
double ** 
PhyloSamplerRun::BuildReducedCounts(GeneMapIterator iter)
{
  //cerr << "--debug::PhyloSamplerRun-BuildReducedCounts - begin" << endl;
  // variable to store counts, start with pseudo (same on each position)
  double ** pCounts = new double*[(*_pWidth)]; 
  for (int i = 0; i < (*_pWidth); i++)
  { pCounts[i] = new double[4]; 
    for (int n = 0; n < 4; n++)
      pCounts[i][n] = _pDatasetPseudo[n]; // _pDatasetPseudo !!
  } 

  // add phylo-psp-weighted counts of other genes
  if (_nbrGenes > 1)
  {
    double **pAddCounts = NULL; 
    // iterate through _pMap
    GeneMapIterator Iter; 
    for (Iter=_pMap->begin(); Iter!=_pMap->end(); Iter++)
    {
      if (Iter != iter) // skip given gene
      {
        pAddCounts = (*Iter).second->GetNtWeights(); // update computed before
        for (int w = 0; w < (*_pWidth); w++)
          for (int n = 0; n < 4; n++) pCounts[w][n] += pAddCounts[w][n];
      }
    }
    pAddCounts = NULL;
  }
  //cerr << "--debug::PhyloSamplerRun-BuildReducedCounts - end" << endl;
  return pCounts;
}
/******************************************************************************
  Description:  BuildReducedGen : build extern1-variable based on other genes
  Author_Date:  MC_2014/02/15
******************************************************************************/
double ** 
PhyloSamplerRun::BuildReducedGen(GeneMapIterator iter)
{
  //cerr << "--debug::PhyloSamplerRun-BuildReducedGen - begin" << endl;
  // variable with counts
  double ** pExtern1 = BuildReducedCounts(iter); 

  // normalize to matrix-probabilities for reduced_matrix
  double sum;
  for (int w = 0; w < *_pWidth; w++) // for each position
  { // calculate sum over nucleotides
    sum = 0;
    for (int n = 0; n < 4; n++) sum += pExtern1[w][n];
    // normalize 
    for (int n = 0; n < 4; n++) pExtern1[w][n] /= sum;
  }

  //cerr << "--debug::PhyloSamplerRun-BuildReducedGen - end" << endl;
  return pExtern1;
}
/******************************************************************************
  Description:  BuildConvergenceGen : build extern1-variable 
                // full_matrix 
  Author_Date:  MC_2014/02/15
******************************************************************************/
double ** 
PhyloSamplerRun::BuildConvergenceGen(GeneMapIterator iter)
{
  //cerr << "--debug::PhyloSamplerRun-BuildConvergenceGen - begin" << endl;
//cerr << "--debug::PhyloSamplerRun-BuildConvergenceGen - next-end" << endl;
  // convergence works with full_matrix
  return BuildReducedGen(_pMap->end());
}
