#include "SamplerRun.h"

/******************************************************************************
 Description:  constructor : 
	 // the common one for NOrthoSamplerRun and PhyloSamplerRun
  Author_Date:  MC_2014/02/15
******************************************************************************/
SamplerRun::SamplerRun()
{
  //cerr << "debug-SamplerRun - constructor - begin " << endl;
  // initialize
  _pMap = new GeneMap;
  _nbrGenes = 0;
  _nbrSequences = 0;
  _pSpeciesMap = new SpeciesMap();
  _tree = false;
  _pWidth = NULL;
  _pDatasetSnf = NULL;
  _pDatasetPseudo = NULL;
//cerr << "debug-SamplerRun - constructor - end " << endl;
}

/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2014/02/15
******************************************************************************/
SamplerRun::~SamplerRun()
{
  //cerr << "destructor SamplerRun - begin" << endl;
  // genemap
  if (_pMap != NULL)
  { 
    for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
    { if ((*iter).second != NULL);
        delete (*iter).second;
      (*iter).second = NULL; 
    }
    delete _pMap; 
  }
  _pMap = NULL;
  // speciesMap
  if (_pSpeciesMap != NULL) delete _pSpeciesMap; _pSpeciesMap = NULL;

  if (_pWidth != NULL) delete _pWidth; _pWidth = NULL;
  // motif
  if (_pDatasetSnf != NULL) delete[]_pDatasetSnf; _pDatasetSnf = NULL;
  if (_pDatasetPseudo != NULL) delete[]_pDatasetPseudo; _pDatasetPseudo = NULL;
  //cerr << "destructor SamplerRun - end" << endl;
}
/******************************************************************************
  Description:  LoadSequences into _pGeneMap / GeneObject / SeqObjList
                format '>>genid
                        >speciesid
                        sequencecode
  Author_Date:  MC_2012/11/20
******************************************************************************/
void
SamplerRun::LoadSequences(string *pFasta, const strand_modes &s)
{
  //cerr << "debug-SamplerRun - LoadSequences - begin " << endl;
  _bMinusStrand = (s == both);
  // open file and set cursor on first >>
  phyloFastaIO *fileIO = new phyloFastaIO(pFasta);
  
  // local variables 
  string * pGene = NULL; 
  SequenceObject *pSeqObj = NULL; 
  GeneObject *pGeneObj = NULL;

  // read input from file
  while ( fileIO->HasNext() || fileIO->HasGene() )
  {
    try
    {
      if (fileIO->HasGene())
      {
        // cleanup previous genid 
        if (pGene != NULL) delete pGene; pGene = NULL;
        // set pointer to new genid
        pGene = fileIO->ReadGene();
        // add gene and/or reset pointer of pGeneObj
        pGeneObj = AddGene(pGene);
      }
      else if(fileIO->HasNext() && pGene != NULL)
      {
        // load the speciesid, sequence code and set genid
        pSeqObj = fileIO->ReadSequence(s);
        pSeqObj->SetGenID(pGene);
        // add SeqObj to the list in pGeneObj
        pGeneObj->AddSequence(pSeqObj); 
        _nbrSequences++;
        // add species to _pSpeciesMap
        _pSpeciesMap->AddSpecies(pSeqObj->GetID(), 1);
        pSeqObj = NULL; // reset (do not delete)
      }
    }
    catch(InvalidInput){}
  }
  // close fastafile
  delete fileIO;
  // cleanup
  if (pGene != NULL) delete pGene; pGene = NULL;
  // check dataset size
  CheckDatasetDimension();
  //cerr << "debug-SamplerRun - LoadSequences - end " << endl;
  return; 
}

/******************************************************************************
  Description:  CheckDatasetDimension() : if not ok, exit to mainPHMS
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::CheckDatasetDimension()
{
  //cerr << "debug-SamplerRun - CheckDatasetDimension() - begin " << endl;
  if (_nbrGenes == 0) 
    throw InputDimension("fastafile : at least 1 gene required");
    // catch in main

  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
    if ( (*iter).second->NbrSequences() > 1 ) break;
  if (iter == _pMap->end())
    throw InputDimension(// catch in main
   "fastafile : only input in coregulation space : use MotifSampler program."); 

  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
  {
    if ( (*iter).second->NbrSequences() < 2 ) 
      throw InputDimension(// catch in main
                        "fastafile : at least 2 sequences per gene required."); 
  }
  //cerr << "debug-SamplerRun - CheckDatasetDimension() - end " << endl;
  return;
}

/******************************************************************************
  Description:  LoadBackgroundModels into _pSpeciesMap
                format >speciesid  BgModel-filename
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::LoadBackgroundModels(string *pBg)
{
  //cerr << "Debug-SamplerRun::LoadBackgroundModels - begin" << endl;
  // open file and set cursor on first >
  phyloFastaIO *bgfileIO = new phyloFastaIO(pBg, "dummy");
  
  // some local variables
  string *pInput = NULL;
  int nbrspecies = 0; 
  
  // read input from file
  while (bgfileIO->HasNext())
  {
    try
    {
      pInput = bgfileIO->ReadNext();
      // store backgroundmodel in _pSpeciesMap
      nbrspecies += _pSpeciesMap->StoreBgModel(pInput);
      // cleanup 
      delete[] pInput; pInput = NULL; 
    }
    catch(InvalidInput){} // catch from phyloFastaIO
  }
  // close the Backgroundname-file
  delete bgfileIO;
 
  // check if each species from the dataset has been covered
  if (nbrspecies != _pSpeciesMap->NbrSpecies())
    throw InputDimension("-b file does not cover all species of the dataset");
	
  //cerr << "debug-SamplerRun - LoadBackgroundModels - end" << endl;
  return;
}

/******************************************************************************
  Description:  LoadCorrelations 
  // tree input is given in Newick format
                format >(spec:br,...); // nodes to be build here !
                       #optional parameters for default motif evolution model
                       >CoregBias(spec:w,...) // towards more conserved
                       >GeneBias(gene:r,...) // towards more mutation
  Author_Date:  MC_2014/02/15
******************************************************************************/
string 
SamplerRun::LoadCorrelations(string *pCorr)
{
  //cerr << "debug-SamplerRun - LoadCorrelations - begin" << endl;
  _tree = false;
  // open file and set cursor on first >
  phyloFastaIO *corrfileIO = new phyloFastaIO(pCorr, "dummy");

  // some local variables
  string *pInput = NULL;
  string Feedback = "";
  // read input from file, catch ErrorInputfiles in mainPHMS
  while (corrfileIO->HasNext())
  { 
    try 
    { pInput = corrfileIO->ReadNext();
      // process legend en info
      if (pInput[0] == "Tree" || pInput[0] == "Star") 
        Feedback = StoreTreeBranchLenghts(pInput);
      else if (pInput[0] == "GeneBias") 
        StoreGeneMutationRates(pInput);
      else if (pInput[0] == "CoregBias") 
        StoreCoregulationWeights(pInput);
      // cleanup
      delete[] pInput; pInput = NULL;  
    }
    catch (InvalidInput){}
  }
  // close corrfile
  delete corrfileIO;
  if (!_tree)     // catch in main
    throw ErrorInputfiles("-c file : no valid tree topology processed ");

  // remove non-splitting internal nodes by merging its branchlengths
  // this is done further in InitializeLocalVariables
/*
  // report (test phaze - can be removed from code later)
	cerr << "#Main-Correlations:" << endl;
map<string, speciesgeg>::iterator Iter = _pSpeciesMap->GetMap()->begin();
	for(; Iter != _pSpeciesMap->GetMap()->end(); Iter++)
	{
		cerr << "Id=" << (*Iter).first << ", Relations:";
		for(int i = 0; i < (int)(*Iter).second.pRelations->size(); i++)
		{ cerr << (*(*Iter).second.pRelations)[i] << ",";}
		cerr << "." << endl;
	}
   */
  //cerr << "debug-SamplerRun - LoadCorrelations - end" << endl;
  return Feedback; 
}
/******************************************************************************
  Description:  StoreTreeBranchLenghts as proximities in each GeneObject
                  format (.....);
                  nodes/root need to be created internally
  Author_Date:  MC_2014/02/15
******************************************************************************/
string  
SamplerRun::StoreTreeBranchLenghts(string *input)
{
  //cerr << "debug--SamplerRun::StoreTreeBranchLengths - begin" << endl;
  string::size_type pos(0); 
  string * pNewick = new string(input[1]); pNewick->append(";");
  // first check if correct format of nodes i.e. equal number of '(' and ')'
  int nbrOpen(0), nbrClose(0);
  while ( (pos=pNewick->find_first_of("(", pos)) != string::npos)
  { pos++; nbrOpen++;}
  pos = 0;
  while ( (pos=pNewick->find_first_of(")", pos)) != string::npos)
  { pos++; nbrClose++;}
  if (nbrOpen != nbrClose)
  { // catch in main
    delete pNewick; pNewick = NULL;
    delete[]input; input = NULL;
    throw ErrorInputfiles("-c file: inconsistent '(' and ')' formatting in newick tree.");
  }

  // first check if tree needs preprocessing (eg not all seq in fasta)
  string feedback = "#Your input newick tree = ";
  feedback.append(*pNewick); feedback.append("\n");
  //cerr << "newick-input=" << *pNewick << endl;
  if (RemoveSequences(pNewick))
  { feedback.append("#Tree after removing non-fasta species = ");
    feedback.append(*pNewick); feedback.append("\n");
  }
  //cerr << "newick-RemoveSequences=" << *pNewick << endl;
  if (UpdateRoot(pNewick))
  { feedback.append("#Final tree after root updating = ");
    feedback.append(*pNewick); feedback.append("\n");
  }
  //cerr << "newick-UpdateRoot=" << *pNewick << endl;

  double branch; int nbrSpecies = 0; bool bSpecies = false;
  int nodeNr = 0;  string id = "NULL";
  // to keep track of processed nodes
  list <string> *pOpen = new list<string>;
  list <string> *pClosed = new list<string>; 
  string::size_type pos1, pos2, pos3, pos4,pos5;
  // read through the inputline 
  while ( (pos5 = pNewick->find_first_of(";\n\r")) != 0)
  {
    pos1 = pNewick->find_first_of("(");
    pos2 = pNewick->find_first_of(")");
    pos3 = pNewick->find_first_of(":");
    pos4 = pNewick->find_first_of(",");

    if (pos4 == 0) //',' 
    { 
      pNewick->erase(0,1);
    } 
    else if (pos1 == 0) //'(' 
    { 
      // create a new node and store it in pOpen
      id = "N";
      ostringstream ostr;ostr << nodeNr;string second = ostr.str();
      id.append(second); 
      pOpen->push_back(id); 
      // store the node in SpeciesMap(string) with empty pRelations
      string * ID = new string(id);
      _pSpeciesMap->AddSpecies(ID, 0); // 0 indicates it is a node
      delete ID; ID = NULL;
      nodeNr++; 
      pNewick->erase(0,1); 
    }
    else if (pos2 == 0) //')'
    { 
      if (pos5 == 1){break;} // end
      // transfer last-opened node from pOpen to pClosed
      if ((int)pOpen->size() == 0)
      { delete pNewick; pNewick = NULL;
        delete[]input; input = NULL;
        delete pOpen; pOpen = NULL; delete pClosed; pClosed = NULL;
        throw ErrorInputfiles("-c file : incorrect newick format >();");
      }
      id = pOpen->back(); 
      pClosed->push_back(id); pOpen->pop_back();
      pNewick->erase(0,1);
      // do not use the user-supplied name of a node
      if (pNewick->find_first_of(":,();") == pos3-1 && pos3 != 1)
      { pNewick->erase(0,pos3-1);}
    }
    else if (pNewick->find_first_of(":,();") == pos3)
    { 
      bSpecies = false;
      // all other cases : '[abc]:b' :=pos3, followed by ,=pos4 or )=pos2
      // process the info for species-node or node-node
      if (pos3 != 0) // read the speciesid
      { 
        id = pNewick->substr(0,pos3); bSpecies = true; 
      }
      // get the branchlength
      pos4 = pNewick->find_first_of(",)");
      istringstream istr(pNewick->substr(pos3+1,pos4-pos3-1)); 
      istr >> branch; // convert to double
      if (branch < 0 || branch == DBL_MAX)
      { // catch in main
        delete pNewick; pNewick = NULL;
        delete[]input; input = NULL;
        delete pOpen; pOpen = NULL; delete pClosed; pClosed = NULL;
        throw ErrorInputfiles("-c file >(...); invalid branch value in newick.");
      }
      try
      { 
        // add id only if present in speciesMap
        if ( _pSpeciesMap->GetMap()->find(id) != _pSpeciesMap->GetMap()->end())
        { 
          // add node-species&&species-node relation to SpeciesMap(pRelations)
          _pSpeciesMap->AddRelation(pOpen->back(), id);
          // store proximity in _pProximityMap(id) for each geneobject
          for( iter = _pMap->begin(); iter != _pMap->end(); iter++)
          { // 2014/01/06 : factor (4/3) to convert bg-brlength to bgprox !
            (*iter).second->StoreProximity(id, exp(branch * (-1.33333)), bSpecies);
          }
          // keep track
          if (bSpecies) {nbrSpecies++;}
        }
      }
      catch (InvalidInput) 
      { // catch in main
        delete pNewick; pNewick = NULL;
        delete[]input; input = NULL;
        delete pOpen; pOpen = NULL; delete pClosed; pClosed = NULL;
        throw ErrorInputfiles("-c file: double input in newick tree.");
      }
      // remove id:branchlength
      pNewick->erase(0,pos4);
    }
  } // reached ; end of newick input

  // check if each species from the dataset has been covered
  if (nbrSpecies != _pSpeciesMap->NbrSpecies())
  { // catch in main
    delete pNewick; pNewick = NULL;
    delete[]input; input = NULL;
    delete pOpen; pOpen = NULL; delete pClosed; pClosed = NULL;
    throw InputDimension(
         "-c file: newick tree does not cover all species of the dataset");
  }
  else 
  { _tree = true;}


  //cleanup
  if (pOpen != NULL) delete pOpen; pOpen = NULL;
  if (pClosed != NULL) delete pClosed; pClosed = NULL; 
  delete pNewick; pNewick = NULL;
  //cerr << "debug--SamplerRun::StoreTreeBranchLengths - end" << endl;
  return feedback;
}

/******************************************************************************
  Description:  RemoveSequences(string * pNewick)
  Author_Date:  MC_2014/02/15
******************************************************************************/
bool  
SamplerRun::RemoveSequences(string * pNewick)
{
  //cerr << "debug--SamplerRun::RemoveSequences - begin" << endl;
  // remove all sequences from the tree not in fasta /format= SEQ5:0.01[,]
  // also remove user supplied node-ids if so
  // also check on incorrect tokens format
  string newick2 = *pNewick;

  // "(%s:%f,"  ",%s:%f,"  ",%s:%f)"
  const char *tokens = "(),:;";
  const char *tokens1 = "),";
  const char *tokens2 = "(:;";
  string::size_type pos(0), pos1, pos2, pos3, pos4;
  // scan through all id
  while ( (pos=pNewick->find_first_not_of(tokens, pos)) != string::npos)
  { 
    // skip if word is a branchlength
    if (pNewick->compare(pos-1,1,":")  == 0) 
    {pos = pNewick->find_first_of(tokens, pos+1);continue;}
    // positioned at start of an id-description
    pos1 = pNewick->find_first_of(":", pos+1);
    pos2 = pNewick->find_first_of(tokens, pos+1);
    pos3 = pNewick->find_first_of(tokens1, pos1+1); // , or )
    pos4 = pNewick->find_first_of(tokens2, pos1+1); // ( or : or ; 
    if (pos1 != pos2 || pos4 < pos3)
    { // catch in main
      delete pNewick; pNewick = NULL;
      throw ErrorInputfiles("-c file: incorrect format [(,]species:branchlength[,)] in newick tree.");
    }
    // remove in case of node-id
    if ( pNewick->compare(pos-1,1,")")  == 0)
    { pNewick->erase(pos,pos1-pos);}
    else // remove species:branch if not in _pSpeciesMap
    { 
      if (_pSpeciesMap->GetMap()->find(pNewick->substr(pos,pos1-pos))
           == _pSpeciesMap->GetMap()->end())
      { pNewick->erase(pos,pos3-pos); 
        if ( pNewick->compare(pos-1,1,",")  == 0){ pNewick->erase(pos-1,1); }
      }
      else
      { pos = pos3;}
    }
  }
  // cleanup kommas and empty node [(,]  ():%f  [,)]
  pos = 0;  
  while ( (pos=pNewick->find("():",pos)) != string::npos)
  { pNewick->erase(pos, 3);
    pNewick->erase(pos,pNewick->find_first_of(tokens,pos+1)- pos); 
  }
  pos = 0;
  while ( (pos=pNewick->find(",,",pos)) != string::npos )
  { pNewick->erase(pos, 1);}
  pos = 0;
  while ( (pos=pNewick->find(",)",pos)) != string::npos )
  { pNewick->erase(pos, 1);}
  pos = 0;
  while ( (pos=pNewick->find("(,",pos)) != string::npos)
  { pNewick->erase( pos+1, 1);}
  //cerr << "debug--SamplerRun::RemoveSequences - end" << endl;
  return (newick2 != *pNewick);
}

/******************************************************************************
  Description:  UpdateRoot(string newick, list<int> *pDelete);
  Author_Date:  MC_2014/02/15
******************************************************************************/
bool  
SamplerRun::UpdateRoot(string *pNewick)
{
  //cerr << "debug--SamplerRun::UpdateRoot - begin" << endl;
  string newick2 = *pNewick;
  string::size_type pos1, pos4;
  // REMOVE non-splitting top-tree branches ((FFF):X[nokomma]); => FFF;
  int counter; 
  while ( pNewick->compare(0,2,"((") == 0 )
  { 
    // get the closing pos2 of the first subnode
    counter=1; pos1 = 2;
    while (counter > 0)
    { if ( pNewick->compare(pos1, 1,"(") == 0) {counter++;}
      else {if ( pNewick->compare(pos1, 1,")") == 0) {counter--;}}
      pos1++;
    }
    if ((pos4 = pNewick->find(",", pos1)) != string::npos)
    {break;}
    // else delete non-splitting root-end (FFF:X[nokomma]); => FFF;
    *pNewick = pNewick->substr(1,pos1-1); //(FFF)
    pNewick->append(";"); 
  }
  //cerr << "debug--SamplerRun::UpdateRoot - end" << endl;
  return (newick2 != *pNewick);
}


/******************************************************************************
  Description:  StoreGeneMutationRates : adjust proximities in GeneObjects
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::StoreGeneMutationRates(string *input)
{
  //cerr << "debug--SamplerRun::StoreGeneMutationRates - begin" << endl;
  // check if tree has been processed
  if (!_tree)
  { delete[]input; input = NULL; // catch in mainPHMS
    throw ErrorInputfiles(
          "-c file format : newick tree should be supplied as first entry.");
  }
  
  string::size_type pos1, pos2; 
  double rate;

  // read through gene inputs
  while ( (pos1 = input[1].find_first_of(":")) != string::npos &&
          (pos2 = input[1].find_first_of(",)")) != string::npos)
  {
    // convert rate to double
    istringstream istr(input[1].substr(pos1+1,pos2-pos1-1));
    istr >> rate;
    if (rate <= 0 || rate == DBL_MAX)
    { 
      delete[] input; input = NULL; // catch in mainPHMS
      throw ErrorInputfiles("-c file >GeneBias(...) invalid rate value");
    }
    if (rate != 1) 
    {
      // apply rate on proximities in the applicable geneObject
      // skip if gene is not present in geneMap
      if ( (iter = _pMap->find(input[1].substr(0,pos1))) != _pMap->end()) 
        (*iter).second->ApplyGeneRate(rate);
    }
    // remove genid:rate,/)
    input->erase(0,pos2+1);
  }
  //cerr << "debug--SamplerRun::StoreGeneMutationRates - end" << endl;
  return; 
}
/******************************************************************************
  Description:  StoreCoregulationWeights : adjust proximities in GeneObjects
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::StoreCoregulationWeights(string *input)
{
  //cerr << "debug--SamplerRun::StoreCoregulationWeights - begin" << endl;
  // check if tree has been processed
  if (!_tree)
  { delete[]input; input = NULL; // catch in mainPHMS
    throw ErrorInputfiles(
          "-c file format : newick tree should be supplied as first entry.");
  }
  
  string::size_type pos1, pos2; 
  double weight;

  // read through species inputs
  while ( (pos1 = input[1].find_first_of(":")) != string::npos &&
          (pos2 = input[1].find_first_of(",)")) != string::npos)
  {
    // convert rate to double
    istringstream istr(input[1].substr(pos1+1,pos2-pos1-1));
    istr >> weight;
    if (weight <= 0 || weight == DBL_MAX)
    { 
      delete[] input; input = NULL; // catch in mainPHMS
      throw ErrorInputfiles("-c file >CoregBias(...) invalid weight value");
    }
    if (weight != 1)
    {
      // apply rate on proximities for applicable species in each geneObject
      // skip if species not present speciesMap
      if ( _pSpeciesMap->GetMap()->find(input[1].substr(0,pos1)) != _pSpeciesMap->GetMap()->end())
      { for( iter = _pMap->begin(); iter != _pMap->end(); iter++)
          (*iter).second->ApplySpeciesWeight(input[1].substr(0,pos1), weight);
      }
    }
    // remove genid:rate,/)
    input->erase(0,pos2+1);
  }
  //cerr << "debug--SamplerRun::StoreCoregulationWeights - end" << endl;
  return; 
}
/******************************************************************************
  Description:  SetNbrInstPrior(max, string, default)
                string can be a filename or direct -p input
                store as <max,type,prob0,prob1,...> in _pSpeciesMap
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::SetNbrInstPrior(const int &max, string *prior, string Default)
{
//cerr << "debug::SamplerRun::SetNbrInstPrior -begin" << endl;
//cerr << "max = " << max << ", prior = " << *prior << ", default = " << Default << endl;
  // case of no -p user-input 
  if (prior == NULL) 
  { // apply default max and prior on all species 
    _pSpeciesMap->StoreMaxPriorInput(max, Default);
    return;
  }
  // in case prior is not a filename
  if ( (prior)->find_first_not_of(".0123456789_ubef\n\r\f\t ") == string::npos )
  { // apply max, prior on all species
    _pSpeciesMap->StoreMaxPriorInput(max, *prior);
    return;
  }

  // in other cases it is a file
  cerr << "#Loading prior input on #inst/seq from " << *prior << endl;
  // open file and set cursor on first >
  phyloFastaIO *priorfileIO = new phyloFastaIO(prior , "dummy");
  int nbrspecies = 0;
  // read input from file
  string * pInput = NULL;
  while (priorfileIO->HasNext())
  {
    try
    { pInput = priorfileIO->ReadNext();
      // store >speciesid -p ... -M ... in _pSpeciesMap
      nbrspecies += _pSpeciesMap->StoreMaxPriorInput(max, Default, pInput);
      // cleanup 
      delete[] pInput; pInput = NULL; 
    }
    catch(InvalidInput){} // catch from phyloFastaIO
  }
  // close the prior-file
  delete priorfileIO;
  // set default for species not listed in this file
  if (nbrspecies != _pSpeciesMap->NbrSpecies())
    _pSpeciesMap->StoreMaxPriorInput(max, Default);

  //cerr << "debug::SamplerRun::SetNbrInstPrior -end" << endl;
  return;
}
/******************************************************************************
  Description:  InitializeLocalVariables1()
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
SamplerRun::InitializeLocalVariables1(const int &w)
{
//cerr << "debug::SamplerRun::InitializeLocalVariables1-begin" << endl;
  // cerr << "debug::SamplerRun::InitializeLocalVariables1 -begin" << endl;
  // 1. initialize _pDatasetSnf and _pDatasetPseudo
  // *************************************************
  // create and initialize
  _pDatasetSnf = _pSpeciesMap->ComputeDatasetSnf();
  _pDatasetPseudo = new double[4]; 
  // pseudo : scale snf with scale = 1/sqrt(nbr sequences)
  double scale = 1/sqrt(_nbrSequences);
  for (int n = 0; n < 4; n++) _pDatasetPseudo[n] = _pDatasetSnf[n] * scale;

  // 2. set parameters
  // *************************************************
  _pWidth = new int(w);

  // 3. initialize local variables in each geneobject
  // *************************************************
  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
  { // _pWidth
    (*iter).second->LinkMotifWidth(_pWidth);
  }
//cerr << "debug::SamplerRun::InitializeLocalVariables1- end" << endl;
  return;
}

/******************************************************************************
  Description:  LoadPSP 
                format '>>genid
                        >speciesid
                        probabilities separated by spaces
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::LoadPSP(string *pPSPfile, bool bPSPs)
{
//cerr << "debug::SamplerRun::LoadPSP - begin" << endl;
  // open file and set cursor on first >>
  phyloFastaIO *fileIO = new phyloFastaIO(pPSPfile);
	
  int nbrPSPgenes = 0;
  int nbrSeqs;
  string trackgenes = "";
  string trackseqs = "";
  string * pGene = NULL;
  string *id = NULL;
  vector<dualSequenceComputation *> * pSeqComps = NULL;
  string copy = "";
  ScoreVector * pPSP = NULL;
  GeneObject * pGeneObj = NULL;
  int L,w;

  // read input from file
  while ( fileIO->HasNext() || fileIO->HasGene() )
  {
    try
    {
      if (fileIO->HasGene())
      {
        // cleanup previous genid 
        if (pGene != NULL) 
        { // check if all seqs for this gene were processed
          try
          {
            if (nbrSeqs != 0)
            { copy = "The number of sequences in -q file does not equal the number of sequences in FASTA file (-f) for gene >>";
              copy.append(*pGene);
              copy.append(". PSP was stored for only [");
              copy.append(trackgenes);
              copy.append("].");
              throw InvalidInput(copy);
            }
          }
          catch(InvalidInput){}
          // delete
          delete pGene; pGene = NULL;
        }
        // set pointer to new genid
        pGene = fileIO->ReadGene();
        // find the gene(object) in GeneMap
        if (_pMap->find(*pGene) == _pMap->end())
        { // skip this gene is not available in FASTA
          copy = *pGene;
          delete pGene; pGene = NULL;// delete so skip full >>block
          throw InvalidInput(copy.append(" (>>gene) is not available in FASTA file (-f)."));
        }
        pGeneObj = (*_pMap)[*pGene];
        nbrPSPgenes++;
        // to report in case of inconsistency
        trackgenes.append(">>");trackgenes.append(*pGene);trackgenes.append(" ");
        // number of sequences to be loaded for this gene
        nbrSeqs = pGeneObj->NbrSequences(); 
        trackseqs = ""; //reset
      }
      else if(fileIO->HasNext() && pGene != NULL)
      {
        // load the speciesid and find in GeneObject
        id = fileIO->ReadSeqID();
        if (id == NULL)
        { copy = *pGene;
          delete pGene; pGene = NULL;
          delete fileIO;
          throw ErrorInputfiles(copy.append(" (>>gene) has incorrect >sequence description(s)."));
        }
        // find species and comp-objects
        pSeqComps = pGeneObj->FindSequence(*id);
        if (pSeqComps == NULL)
        { // sequence id not found, skip the PSP information
          copy = ">"; copy.append(*id); 
          copy.append(" (in >>");copy.append(*pGene);
          delete id; id = NULL;
          fileIO->ProceedCursor(); // skip the current probs line
          throw InvalidInput(copy.append(") is not available in FASTA file (-f)."));
        }
        // load the PSP data
        L = (*pSeqComps)[0]->SeqLength(); 
        w = (*pSeqComps)[0]->MotifLength(); 
        try 
        {
          pPSP = fileIO->LoadPspData(L, w); // has throws to ErrorInputfiles
        }
        catch(ErrorInputfiles)
        { 
          copy = ">"; copy.append(*id); 
          copy.append(" (in >>");copy.append(*pGene);
          delete pGene; pGene = NULL;
          delete fileIO;
          delete id; id = NULL;
          throw ErrorInputfiles(copy.append("): check PSP description."));
        }
        // store pPSP data into compObjects
        try 
        {
          for (int i = 0; i < (int)pSeqComps->size(); i++)
          { 
            (*pSeqComps)[i]->LoadPspScores(pPSP, bPSPs);// has throws to ErrorInputfiles
          }
        }
        catch(ErrorInputfiles)
        { 
          copy = ">"; copy.append(*id); 
          copy.append(" (in >>");copy.append(*pGene);
          delete pGene; pGene = NULL;
          delete fileIO;
          delete id; id = NULL;
          delete pPSP; pPSP = NULL;
          throw ErrorInputfiles(copy.append("): check PSP description."));
        }
        nbrSeqs--; // should count down to zero
        // to report in case of inconsistency
        trackseqs.append(">");trackseqs.append(*id); trackseqs.append(" ");
        delete id; id = NULL;
        pSeqComps = NULL;
        delete pPSP; pPSP = NULL;
      }
    }
    catch(InvalidInput){}
  }
  // close fastafile
  delete fileIO;

  // check if all genes have been covered
  try
  {
    if (nbrPSPgenes != _nbrGenes)
    { 
      char pTmp[128];
      copy = "The number of genes in PSP file (";
      sprintf(pTmp, "%d", nbrPSPgenes);
      copy.append(pTmp);
      copy.append(") does not equal the number of genes in FASTA file (");
      sprintf(pTmp, "%d", _nbrGenes);
      copy.append(pTmp);
      copy.append("). PSP was stored for only [");
      copy.append(trackgenes);
      copy.append("].");
      throw InvalidInput(copy);
    }
  }
  catch(InvalidInput){}
//cerr << "debug::SamplerRun::LoadPSP - end" << endl;
  return;
}
/******************************************************************************
  Description:  ResetMasks in all sequences at start of a new run 
                all maskvalues to 1 in SeqComp
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::ResetMasks()
{
//cerr << "debug--SamplerRun::ResetMasks() - begin" << endl;
  SequenceMap * pseq = NULL; 
  SeqIterator i;
  // iterate through all GeneObjects
  for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
  {
    pseq = (*iter).second->GetSequences(); 
    // iterate through all sequences in a GeneObject
    for (i = pseq->begin(); i != pseq->end(); i++)
    {
      (*(*i).second)[0]->GetMaskVector()->ResetMask(); // all 1
      // last w-1 values zero
      (*(*i).second)[0]->GetMaskVector()->UpdateMask((*i).first->Length() - *_pWidth + 1, *_pWidth - 1, 0); 
      if (_bMinusStrand)
      { (*(*i).second)[1]->GetMaskVector()->ResetMask();
        (*(*i).second)[1]->GetMaskVector()->UpdateMask(
            (*i).first->Length() - *_pWidth + 1, *_pWidth - 1, 0); 
      }
    }
  }
//cerr << "debug--SamplerRun::ResetMasks() - end" << endl;
  return;
}


/******************************************************************************
  Description:  ShiftFullInstanceMap() : if shifted species instances 
                map has higher IC score, then shift all instances
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
SamplerRun::ShiftFullInstanceMap(const int &max)
{
  //cerr << "--debug::SamplerRun-ShiftFullInstanceMap - begin" << endl;

  double ic(0), icBest(0); PWM * pPWM = NULL; int bestShift = 0; 

  //cerr << "- IC : ";
  // shift map and compare ic-scores
  for (int s = -max; s <= max; s++)
  { // create the (shifted) motif
    pPWM = BuildMotifFromSpeciesInstanceMap(s, "all"); 
    // compare score
    ic = pPWM->InformationContent(_pDatasetSnf); 
    //cerr << s << "=" << setw(5) << setprecision(3) << ic << "  "; 
    if (ic > icBest) { icBest = ic;  bestShift = s;}
    delete pPWM; pPWM = NULL;
  }
  
  // store shifted map with best ic-score
  if (bestShift != 0)
  { // replace all instances to shifted instancemap
    for (iter=_pMap->begin();iter!=_pMap->end(); iter++)
    { InstanceMap *pShiftMap = (*iter).second->BuildShiftedSpeciesMap(bestShift, "all");
      (*iter).second->SetSpeciesInstances(pShiftMap);
    }
    // report
    //cerr << " Instances were shifted with " << bestShift << " position(s)" ; 
  }
  //cerr << "--debug::SamplerRun-ShiftFullInstanceMap - end" << endl;
  return; 
}

/******************************************************************************
  Description:  EvaluateMotif() : check if enough sequences with instance
                and return nbrseq-1-inst and nbrinst for this motif
	  Species may be "all" or a particular species
  Author_Date:  MC_2014/02/15
******************************************************************************/
int *
SamplerRun::EvaluateMotif(const double &treshold, const string &Species)
{
  //cerr << "--debug::SamplerRun-EvaluateMotif - begin" << endl;
  int n1 = 0; // number of sequences with at least one instance [of Species]
  int n2 = 0; // number of instances [of Species]
  list<string> *pSeqs = new list<string>; // to track seq-ID of each instance
  string * pSeqID = NULL;

  InstanceMap * InstMap = NULL;
  InstanceIterator i;
  for (iter = _pMap->begin();iter != _pMap->end(); iter++)
  {
    InstMap = (*iter).second->GetSpeciesInstances();
    for(i = InstMap->begin(); i != InstMap->end(); i++)
    {
      pSeqID = new string(*((*i)->ParentSequence()->GetID())); // speciesid
      if (Species == "all" || Species == *pSeqID)
      { 
        n2++; //  number of instances [of Species]
        pSeqID->append("_");
        pSeqID->append((*iter).first);
        pSeqs->push_back(*pSeqID); // add seqid as 'speciesid_geneid'
      }
      delete pSeqID; pSeqID = NULL;
    }
  }

  // remove duplicates
  pSeqs->sort();  pSeqs->unique();
  // number of sequences with at least one instance
  n1 = (int)pSeqs->size();
  // cleanup
  delete pSeqs; pSeqs = NULL; 
  InstMap = NULL;

  if (n1 < treshold * _nbrSequences)
    throw AbortSampling("too few sequences have at least one instance.");

  int * n = new int[2]; n[0] = n1; n[1] = n2;
  //cerr << "--debug::SamplerRun-EvaluateMotif - end" << endl;
  return n;
}

/******************************************************************************
  Description:  PrintMotifScores : 
                 compute CS,IC,LL,dLL and print all motifinformation
			  Species may be "all" or a particular species
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
SamplerRun::PrintMotifScores(GFFWriter * gff, string Motif, 
								   PWM * motif, const string &Species)
{
  //cerr << "--debug::SamplerRun-PrintMotifScores - begin" << endl;
  // compute ll 
  double ll(0);
  SequenceMap *pSeq = NULL;
  SeqIterator i;
  // iterate through all genes 
  for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
  { // iterate through all sequences of this gene for ll-score
    pSeq = (*iter).second->GetSequences();
    for (i = pSeq->begin(); i != pSeq->end(); i++)
    {
      if (Species == "all" || Species == *((*i).first->GetID()))
      {
        ll += (*(*i).second)[0]->GetLLscore();
        if (_bMinusStrand) ll += (*(*i).second)[1]->GetLLscore();
      }
    }
  }
  pSeq = NULL;

  // store and report the scores 
  ostringstream text;
  text << Motif; // add information coming from main
  text << "\tcs: " << setprecision(3) << motif->ConsensusScore()
  << "\tic: " << setprecision(3) << motif->InformationContent(_pDatasetSnf);
  motif->SetLL(ll); 
  text << "\tll: " << setprecision(6) << ll;

  gff->AddComment(text.str());
  text.flush();
  return; 
  //cerr << "--debug::SamplerRun-PrintMotifScores - end" << endl;
}

/******************************************************************************
  Description:  UpdateMasks : mask the positions of the  
                retrieved instances zero in each pSeqComp
                note : '+1' in start-argument and '-1' in length-argument
                       '2*W' voor opposite strand (bug motifsampler)
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::UpdateMasks(const double &maxMask, const int &overlap)
{
  //cerr << "--debug::SamplerRun-UpdateMasks - begin" << endl;
  int start(0); // instance start
  InstanceMap *pInstMap = NULL; InstanceIterator ii; 
  SequenceMap * pSeq = NULL; 
  vector<dualSequenceComputation *> * pSeqComp = NULL;

  // iterate through all instances of all genes
  for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
  {
    pInstMap = (*iter).second->GetSpeciesInstances();
    pSeq = (*iter).second->GetSequences();
    for (ii = pInstMap->begin(); ii != pInstMap->end(); ii++)
    { // instance
      start = (*ii)->Start(); 
      pSeqComp = (*pSeq)[(*ii)->ParentSequence()];
      // mask 
      if ((*ii)->Strand() == plus_strand)
      {
        (*pSeqComp)[0]->GetMaskVector()->UpdateMask(start -(*_pWidth) + overlap + 1, 
                           2 * ((*_pWidth) - overlap) - 1, 0);
        if (_bMinusStrand) (*pSeqComp)[1]->GetMaskVector()->UpdateMask((*ii)->
                       ParentSequence()->Length() - start - 2 * (*_pWidth) + 
                       overlap + 1, 2 * ((*_pWidth) - overlap) - 1, 0);
      }
      else // minus_strand
      {
        (*pSeqComp)[0]->GetMaskVector()->UpdateMask((*ii)->ParentSequence()
                           ->Length() - start - 2 * (*_pWidth) + overlap + 1, 
                           2 * ((*_pWidth) - overlap) - 1, 0);
        (*pSeqComp)[1]->GetMaskVector()->UpdateMask( start - (*_pWidth) + overlap 
                                   + 1, 2 * ((*_pWidth) - overlap) - 1, 0);
      }
      // evaluate against maxMask treshold
      if ((*pSeqComp)[0]->GetMaskVector()->MaskedFraction() >= maxMask ||
   (_bMinusStrand  && (*pSeqComp)[1]->GetMaskVector()->MaskedFraction() >= maxMask))
      {
        string id = (*iter).first; id.append("-"); 
               id.append(*((*ii)->ParentSequence()->GetID()));
        throw AbortMasking(id.append("/Too much masking "));
      }
    }
  }
  //cerr << "--debug::SamplerRun-UpdateMasks - end" << endl;
  return;
}
/******************************************************************************
  Description:  PrintToOutput : print full motif to txt- and mtrx- files
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::PrintToOutput(PWM * pmotif, GFFWriter * pText, PWMIO * pMtrx)
{
  //cerr << "--debug::SamplerRun-PrintToOutput - begin" << endl;
  // this was earlier in main, now one level lower
  string * firstline = new string("#id: ");
  firstline->append(*pmotif->GetID()); // box_1_1_ACGT
  firstline->append("\t");
  firstline->append(*pmotif->GetInfo()); // nbrSeq	nbrInst

  // compute scores and print
  PrintMotifScores(pText, *firstline, pmotif, "all");
  delete firstline; firstline = NULL;

  // print instances 
  PrintMotifInstances(pText, "all");

  // print matrices to matrixfile
  if (pMtrx != NULL)
  { pmotif->SetScore(pmotif->GetLL()); // default 'score'
    pMtrx->WriteMatrix2(pmotif); 
  }
  //cerr << "--debug::SamplerRun-PrintToOutput - end" << endl;
  return;
}

/******************************************************************************
  Description:  PrintMotifInstances : list of individual instances
	            Enable to select on specific 'species'
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::PrintMotifInstances(GFFWriter * gff, const string &Species)
{
  //cerr << "--debug::SamplerRun-PrintMotifInstances - begin" << endl;
  // Species may be "all" or a given particular species.

  ostringstream text;
  InstanceMap * instMap; InstanceIterator i;
  GeneMapIterator iiiter; 
  
  // report 'Species' instances for each GeneObject
  for (iiiter =_pMap->begin(); iiiter != _pMap->end(); iiiter++)
  {
    // get the instances for this gene
    instMap = (*iiiter).second->GetSpeciesInstances();
    // print each instance
    for(i = instMap->begin(); i != instMap->end(); i++)
    {
      if (Species == "all" || Species == *((*i)->ParentSequence()->GetID()))
      {
        text << (*iiiter).first << "-"; // genid
        text << *((*i)->ParentSequence()->GetID()); // speciesid
        int length = (*i)->ParentSequence()->Length();
        if ((*i)->Strand() == plus_strand) 
        { // plus- positions/score
          text << "\t" << (*i)->Start() + 1 << "\t";
          text << (*i)->Start() + (*i)->Length() << "\t";
          text << (int)(*i)->Score() << "\t";
          text << "+\t";
        }
        else
        { // minus- positions/score
          text << "\t" << length - (*i)->Start() - (*i)->Length() + 1 << "\t";
          text << length - (*i)->Start() << "\t";
          text << (int)(*i)->Score() << "\t";
          text << "-\t";
        }
        // add instance-code
        text << "site " << *((*i)->PrintSite()) << endl;
      }
    }
  }
  if (text.str() != ""){gff->AddComment(text.str());  text.flush();}
  //cerr << "--debug::SamplerRun-PrintMotifInstances - end" << endl;
  return;
}

/******************************************************************************
  Description:  PrintToSpecOutput : print species motif's to txt- and mtrx- files
  Author_Date:  MC_2014/02/15
******************************************************************************/
void 
SamplerRun::PrintToSpecOutput(string box_r_m, GFFWriter * pText_spec, PWMIO * pMtrx_spec)
{
  //cerr << "--debug::SamplerRun-PrintToSpecOutput - begin" << endl;
  PWM * pSpecMotif = NULL;

  // iterate through all species
  map<string, speciesgeg>::iterator specIter;
  for (specIter = _pSpeciesMap->GetMap()->begin(); 
	   specIter != _pSpeciesMap->GetMap()->end(); specIter++)
  {
    // only for species
    if ((*specIter).second.pBgM == NULL) {continue;} // skip node
    // EvaluateMotif (but not against limiting threshold)
    int * n1n2 = EvaluateMotif(0,(*specIter).first);
    if ( n1n2[1] == 0) { continue;} // no instances for this species

    // compose species-specific PWM
    pSpecMotif = BuildMotifFromSpeciesInstanceMap(0, (*specIter).first);  
    //if (pSpecMotif == NULL) { continue;} // no instances for this species

    // compose id
    string * firstline = new string(box_r_m); 
    firstline->append("_");
    firstline->append(*(pSpecMotif->GetConsensus()));
    firstline->append("_");
    firstline->append((*specIter).first); // box_1_1_ACGT_species1
    pSpecMotif->SetID(*firstline); // store this id
    // store info
    ostringstream info;
    info << "nbrSeq: " << n1n2[0] << "\tnbrInst: " << n1n2[1];
    pSpecMotif->SetInfo(info.str());
    info.flush();
	delete[] n1n2; n1n2 = NULL;
	  
    firstline->insert(0, "#id: "); // #id: box_1_1_ACGT_species1
    // add info on nbrSeq and nbrInst 
    firstline->append("\t");
    firstline->append(info.str()); 
    // compute scores and print
    PrintMotifScores(pText_spec, *firstline, pSpecMotif, (*specIter).first);
    delete firstline; firstline = NULL;

    // print motif instances to text file
    PrintMotifInstances(pText_spec, (*specIter).first);

    // print PWMs to matrix file
    if (pMtrx_spec != NULL)
    { pSpecMotif->SetScore(pSpecMotif->GetLL()); // default 'score'
      pMtrx_spec->WriteMatrix2(pSpecMotif); 
    }
    // cleanup
    delete pSpecMotif; pSpecMotif = NULL;
  }
  //cerr << "--debug::SamplerRun-PrintToSpecOutput - end" << endl;
  return;
}

/******************************************************************************
  Description:  BuildMotifFromSpeciesInstanceMap(int s, string Species) 
                s indicates a shift of all instances for s positions
                Species may be "all" or a specific one
  Author_Date:  MC_2014/02/15
******************************************************************************/
PWM * 
SamplerRun::BuildMotifFromSpeciesInstanceMap(const int &s, string Species)
{
  //cerr << "--debug::SamplerRun-BuildMotifFromSpeciesInstanceMap - begin" << endl;
  GeneMapIterator Iter; // debug 
  // variable to store counts
  int ** pCounts = new int*[(*_pWidth)]; 
  for (int i = 0; i < (*_pWidth); i++)
  { pCounts[i] = new int[4]; 
    for (int n = 0; n < 4; n++) pCounts[i][n] = 0; 
  } 
  int **pAddCounts = NULL; 

  if (s==0 && Species == "all") // no shift, all instances
  { for (Iter=_pMap->begin();Iter!=_pMap->end(); Iter++)
    { pAddCounts = (*Iter).second->GetSpeciesInstances()->GetIntegerCounts();
      if (pAddCounts != NULL)
      {
        for (int w = 0; w < (*_pWidth); w++)
        { for (int n = 0; n < 4; n++) pCounts[w][n] += pAddCounts[w][n];
          delete[] pAddCounts[w];
        }
        delete[] pAddCounts; pAddCounts = NULL;
      }
    }
  }
  else // s>0 or a particular species
  {
    // make up shifted counts of all instances
    InstanceMap *pShiftMap = NULL;
    for (Iter=_pMap->begin();Iter!=_pMap->end(); Iter++)
    { pShiftMap = (*Iter).second->BuildShiftedSpeciesMap(s, Species);
      pAddCounts = pShiftMap->GetIntegerCounts();
      delete pShiftMap; pShiftMap = NULL; 
      if (pAddCounts != NULL)
      {
        for (int w = 0; w < (*_pWidth); w++)
        { for (int n = 0; n < 4; n++)
            pCounts[w][n] += pAddCounts[w][n];
          delete[] pAddCounts[w]; 
        }
        delete[] pAddCounts; pAddCounts = NULL;
      }
    }
  }
  // build the pwm
  PWM * pPWM = new PWM((*_pWidth), pCounts, _pDatasetPseudo);

  // cleanup
  for (int w = 0; w < (*_pWidth); w++) delete[] pCounts[w]; 
  delete[] pCounts; pCounts = NULL;  
  //cerr << "--debug::SamplerRun-BuildMotifFromSpeciesInstanceMap - end" << endl;
  return pPWM; 
}
