#include "SpeciesMap.h"

/******************************************************************************
  Description:  constructor 
  Author_Date:  MC_2012/11/26
******************************************************************************/
SpeciesMap::SpeciesMap()
{
  _pMap = new map<string, speciesgeg>;
  _nbrSpecies = 0;
  _nbrNodes = 0;
}
/******************************************************************************
  Description:  destructor 
  Author_Date:  MC_2012/11/26
******************************************************************************/
SpeciesMap::~SpeciesMap()
{
  if (_pMap != NULL)
  {
    for (iter = _pMap->begin(); iter != _pMap->end(); iter++)
    {
      // only delete snf for nodes (for species, it is linked with pBgModel)
      if ((*iter).second.pBgM == NULL && (*iter).second.pSNF != NULL)
        delete[] (*iter).second.pSNF;
      (*iter).second.pSNF = NULL; 
      if ((*iter).second.pBgM != NULL)
        delete (*iter).second.pBgM;
      (*iter).second.pBgM = NULL; 
      if ((*iter).second.pNbrInstDistr != NULL)
      {
        for(int i = 0; i < (int)(*iter).second.pNbrInstDistr->size(); i++)
        { // delete distribution
          if ( (*(*iter).second.pNbrInstDistr)[i] != NULL)
            delete (*(*iter).second.pNbrInstDistr)[i];
          (*(*iter).second.pNbrInstDistr)[i] = NULL;
        }
        delete (*iter).second.pNbrInstDistr; // delete vector
      }
      (*iter).second.pNbrInstDistr = NULL;
      if ((*iter).second.pRelations != NULL)
        delete (*iter).second.pRelations; // delete vector (not strings)
      (*iter).second.pRelations = NULL; 
    }
    delete _pMap; 
  }
  _pMap = NULL;
}
/******************************************************************************
  Description:  AddSpecies : if not yet present, speciesgeg contents is NULL 
  Author_Date:  MC_2012/11/26
******************************************************************************/
void
SpeciesMap::AddSpecies(string *pSpecies, bool species)
{
//cerr << "--debug--SpeciesMap::AddSpecies - begin" << endl;
  // add species if not yet present
  if (_pMap->find(*pSpecies) == _pMap->end())
  {
    speciesgeg x; 
    x.pBgM = NULL;
    // 2014/01/19
    if (species) x.pSNF = NULL; 
    else 
    { x.pSNF = new double[4]; 
      for (int i = 0; i < 4; i++) 
      {
        x.pSNF[i] = 0;
      }
    }
    x.pNbrInstDistr = NULL;
    x.pRelations = new vector<string>;
    (*_pMap)[*pSpecies] = x;
    if (species) _nbrSpecies++;
    else _nbrNodes++;
  }
  return;
}

/******************************************************************************
  Description:  StoreBgModel : for a species present in dataset 
                input[0] = speciesid, input[1] = backgroundmodelname
  Author_Date:  MC_2012/11/26
******************************************************************************/
int
SpeciesMap::StoreBgModel(string * input)
{
  // skip species that is not in dataset
  if ( _pMap->find(input[0]) == _pMap->end()) return 0; 
  
  // skip double input (bgModel was already stored)
  if ((*_pMap)[input[0]].pBgM != NULL)
  {
    delete[] input; input = NULL; // catch in mainPHMS
    throw ErrorInputfiles(" double input in -b file"); 
  }
  // open BgMfile for reading and store model
  BackgroundIO *bgIO = new BackgroundIO(input[1], READ);
  if (bgIO->IsOpen()) 
  {
    BackgroundModel* pBgModel = bgIO->ReadBackgroundModel();
    if ( pBgModel != NULL )
    {
      (*_pMap)[input[0]].pBgM = pBgModel;
      // link the snf 
      (*_pMap)[input[0]].pSNF = pBgModel->GetSNF();
      return 1;
    }
  }
  return 0;
}
/******************************************************************************
  Description:  AddRelation : parent to daughter and daughter to parent
  parent is always the first string, any other daughter is added at the end
  Author_Date:  MC_2012/11/22
******************************************************************************/
void
SpeciesMap::AddRelation(string parent, string daughter)
{
  //cerr << "debug--SpeciesMap::AddRelation - begin" << endl;
  // daughter=species can only have 1 parent and no daughters
  if ((*_pMap)[daughter].pBgM != NULL)
  { if (int((*_pMap)[daughter].pRelations->size()) != 0)
    { 
      string error = daughter;
      error.append("/"); error.append(parent); 
      error.append(" -incorrect SpeciesMap::AddRelation processing.");
      throw InvalidInput(error);
    }
  }

  // first add the parent to the daughter
  vector<string> * pLink = (*_pMap)[daughter].pRelations;
  if ( (int)pLink->size() == 0 )
  { 
    pLink->push_back(parent);
  }
  else   
  { 
    pLink->push_back("last");   // add one blank at end
    // need to put parent on first position (push_front n/a for vector)
    // first move all daughters one position up
    for(int i = (int)pLink->size()-1; i > 0; i--)
    { 
      (*pLink)[i] = (*pLink)[i-1];
    }
    // set parent on pos0
    (*pLink)[0] = parent;
  }

  // now add the daughter to the parent
  (*_pMap)[parent].pRelations->push_back(daughter);

   //cerr << "debug--SpeciesMap::AddRelation - end" << endl;
  return;
}

/******************************************************************************
  Description:  StoreMax&PriorInput(int,string) string = contents from -p
                copy this for all species where .second.pNbrInstDistr == NULL
	  Mark : Max is never zero in this function
  Author_Date:  MC_2012/11/26
******************************************************************************/
void
SpeciesMap::StoreMaxPriorInput(const int &Max, string Input)
{
//cerr << "debug-SpeciesMap::StoreMaxPriorInput - begin " << endl;

  // process the Input-line (format 0,u,b,e,f)
  ScoreVector * pPrior = PHYLINCLUSIVE::ExtractPriorParameters(Max, Input);


  // iterate through _pMap
  for(iter = _pMap->begin(); iter != _pMap->end(); iter++)
  {
    // store the input where not yet done
    if ((*iter).second.pNbrInstDistr == NULL)
      // created each time again to avoid multiple allocation
      (*iter).second.pNbrInstDistr = 
                    PHYLINCLUSIVE::CreateNbrInstDistribution(Max, pPrior);
  }
  // cleanup
  delete pPrior; pPrior = NULL;
  //cerr << "debug-SpeciesMap::StoreMaxPriorInput - end " << endl;
  return;
}
/******************************************************************************
  Description:  StoreMax&PriorInput : input[0]=species, input[1]=-p...&/-M...
  Author_Date:  MC_2012/11/26
******************************************************************************/
int
SpeciesMap::StoreMaxPriorInput(const int &max, string pdefault, string *input)
{
//cerr << "debug-SpeciesMap::StoreMaxPriorInput(.,.,.) - begin " << endl;
  // skip species that is not in dataset
  if ( _pMap->find(input[0]) == _pMap->end()) return 0;
  
  // skip double input (it was already stored)
  if ((*_pMap)[input[0]].pNbrInstDistr != NULL)
  { delete[] input; input = NULL; // catch in mainPHMS
    throw ErrorInputfiles(" double input in -p file"); 
  }

  // process the -p and or -M input
  string p = pdefault; int M = max;
  try
  {
    while (!input[1].empty()) // -p ... \(-M ...\)
    {
      if( PHYLINCLUSIVE::ReadPriorIndex(input[1])) // -p input
      {
        // substract -p input 
        string::size_type pos = input[1].find_first_not_of("fueb0.123456789_");
        p = input[1].substr(0, pos);
//		  cerr << "p = " << p << endl;
        input[1].erase(0,pos); 
        // and erase leading whites
        while ( input[1].find_first_of("_ \t") == 0) input[1].erase(0,1);
      }
      else M = (int)PHYLINCLUSIVE::ReadValue(input[1]); // -M max input
//		  cerr << "M = " << M << endl;}
    }
    if (M > 0)
    {
      // process the Input-line (format 0,u,b,e,f)
      ScoreVector * pPrior = PHYLINCLUSIVE::ExtractPriorParameters(M, p);
      // create distributions and add to speciesMap
      (*_pMap)[input[0]].pNbrInstDistr = 
                            PHYLINCLUSIVE::CreateNbrInstDistribution(M, pPrior);
      delete pPrior; pPrior = NULL;
      return 1;
    }
    else { return 0;}
  }
  catch(InvalidInput){ delete[]input; input = NULL; // catch from phylutilities
                       throw ErrorInputfiles(
                                       "invalid -p ... (-M ...) in -p file"); 
                     }
}

/******************************************************************************
  Description:  GetSpeciesDaughters(int node) : 
	  create a new vector with only species daughters, NULL if only node-daughters
  Author_Date:  MC_2012/12/06
******************************************************************************/
StringVector *
SpeciesMap::GetSpeciesDaughters(const int nodeNbr)
{
  // create parentid
  string * parentid = new string("N");
  ostringstream ostr; ostr << nodeNbr; parentid->append(ostr.str());

  StringVector * pReturn = new StringVector;
  // get the full (parent)/daughters 
  StringVector * pAll = (*(_pMap->find(*parentid))).second.pRelations;
  // screen through daughters
  for (int i = 0; i < int(pAll->size()); i++)
  {
    // check if this is a species-id (via pBgM != NULL)
    if ( (*(_pMap->find((*pAll)[i]))).second.pBgM != NULL)
    { 
      pReturn->push_back((*pAll)[i]);
    }
  }
  delete parentid; parentid = NULL;
  if (int(pReturn->size()) == 0) { delete pReturn; pReturn = NULL;}
  return pReturn;
}

/******************************************************************************
  Description:  ComputeDatasetSnf
  Author_Date:  MC_2012/11/26
******************************************************************************/
double *
SpeciesMap::ComputeDatasetSnf()
{
  // create and initialize
  double * snf = new double[4]; 
  for (int n = 0; n < 4; n++) snf[n] = 0;

  // update based on SNF from individual species : compute sum
  double * spsnf = NULL; 
  for(iter =_pMap->begin(); iter!=_pMap->end(); iter++)
  {
    if ((*iter).second.pBgM != NULL) // only species attribute
    {
      spsnf = (*iter).second.pBgM->GetSNF();
      for (int n = 0; n < 4; n++) 
        snf[n] += spsnf[n];
    }
  }
  // and normalize
  double sum = 0;
  for (int n = 0; n < 4; n++) sum += snf[n];
  for (int n = 0; n < 4; n++) snf[n] /= sum;

  return snf;
}
