#include "dualSequenceComputation.h"

using namespace INCLUSIVE;

/******************************************************************************
  Description:  constructor 
  Author_Date:  MC_2014/02/15
******************************************************************************/
dualSequenceComputation::dualSequenceComputation(const strand_modes s, 
SequenceObject *pSeqObj, int *w, double &prox, vector<Distribution*> * dists)
{
   //cerr << "Debug-dualSequenceComputation::constructor -begin" << endl;
  // 2014/02/28
  // used for Pr(N=0) computation in UpdateNbrInstDistribution for PHMS
  _LogRedInstBg = 0; // neutral contribution
  // pointer link
  _pSeqObj = pSeqObj;
  if (s == plus_strand) _pSequence = pSeqObj->GetSequenceVector();
  else _pSequence = pSeqObj->GetReverseSequenceVector();
  _pwlength = w; 
  _pProx = &prox;
  // initialize variables
  _strand = s; 
  _seqlength = pSeqObj->Length();
  _LL = 0;

  // segment vectors
  _pSegmentBgScore = new ScoreVector(_seqlength - *_pwlength + 1, 0);
  _dLogP0 = 0;

  _pMask = new MaskVector(_seqlength - *_pwlength+1); // all 1
  _pPspScore = new ScoreVector(_seqlength - *_pwlength + 1, 1); // all 1

  // computation tools
  _pMotifScore = new ScoreVector(_seqlength - *_pwlength + 1, 0);
  _pExpWx = new ScoreVector(_seqlength - *_pwlength + 1, 0);
  
  // #inst/seq : link pointers and set maximum nbr instances
  _pFixedDistr = (*dists)[0];
  _pComputedDistr = NULL;
  if (_pFixedDistr != NULL) // case f
  { _pPriorDistrs = NULL;
    _maxNbrInst = (int)_pFixedDistr->Size() -1;
    if (_maxNbrInst > _seqlength/(*_pwlength)) // reset max for short sequences
      _maxNbrInst = (int)(_seqlength/(*_pwlength));
    _pPrior2NormFactor = NULL;
  }
  else 
  { _pPriorDistrs = dists; // first element will be NULL now
    _maxNbrInst = (int)dists->size()-1;
    if (_maxNbrInst > _seqlength/(*_pwlength)) // reset max for short sequences
      _maxNbrInst = (int)(_seqlength/(*_pwlength));
    // create the vector to store prior2 normalization factors 
    _pPrior2NormFactor = new ScoreVector(_maxNbrInst,0);
  }

  PSP_s = false; // default no PSP impact whilst sampling
  
  //cerr << "Debug-dualSequenceComputation::constructor -end" << endl;
}
/******************************************************************************
  Description:  destructor 
  Author_Date:  MC_2014/02/15
******************************************************************************/
dualSequenceComputation::~dualSequenceComputation()
{
  //cerr << "destructor-dualSequenceComputation -begin" << endl;
  if (_pSegmentBgScore != NULL) delete _pSegmentBgScore; _pSegmentBgScore = NULL;
  if (_pMask != NULL) delete _pMask;  _pMask = NULL; 
  _pProx = NULL;
  if (_pPspScore != NULL) delete _pPspScore; _pPspScore = NULL;
  if (_pMotifScore != NULL) delete _pMotifScore; _pMotifScore = NULL;
  if (_pExpWx != NULL) delete _pExpWx; _pExpWx = NULL;

  if (_pComputedDistr != NULL) delete _pComputedDistr; _pComputedDistr = NULL;
  if (_pPrior2NormFactor != NULL) delete _pPrior2NormFactor; _pPrior2NormFactor = NULL;
  
  _pFixedDistr = NULL;// link
  _pPriorDistrs = NULL; // unlink
  _pSeqObj = NULL;// link
  _pSequence = NULL; // link
  _pwlength = NULL; // link 

  //cerr << "destructor dualSequenceComputation -end" << endl;
}
/******************************************************************************
  Description:  ComputeBgScores(BackgroundModel *bgM) of segments 
  Author_Date:  MC_2014/02/15
******************************************************************************/
void
dualSequenceComputation::ComputeBgScores(BackgroundModel *bgM)
{
  //cerr << "Debug-dualSequenceComputation::UpdateBgScores -begin" << endl;
  // reuse function from Utilities.h
  INCLUSIVE::SegmentLogBackgroundScore( _pSegmentBgScore, _pSeqObj, _strand, bgM, *_pwlength);

  _dLogP0 = INCLUSIVE::SequenceLogBackgroundScore(_pSeqObj, _strand, bgM);


  //cerr << "Debug-dualSequenceComputation::UpdateBgScores -end" << endl;
  return; 
}
/******************************************************************************
  Description:  set the PSP score of all segments
	  last L-w values are already set to zero (in both strands if so)
	  L entries : for + strand, symmetrical if - strand needed
	  2L entries : for + respectively - strand (if needed)
	  convert scores to S(x)/(1-S(x))
	  NO NORMALIZATION HERE
  Date:         2014/02/15
  Author:       Marleen Claeys <mclaeys@qatar.net.qa>
******************************************************************************/
void
dualSequenceComputation::LoadPspScores(ScoreVector * psp, bool psp_s)
{
  //cerr << "Debug-dualSequenceComputation::LoadPspScores -begin" << endl;
  // set local _PSPimpact
  PSP_s = psp_s;

  // define some iterators
  //vector < double >::iterator iter1;
  int L = psp->size();

  // update each PSP prior score
	// Mark : the input psp values are SCORES, not a distribution
	// Mark : the values in psp have been rescaled [0.1-0.9] whilst reading from file
	// Mark : last 'w' psp-positions were reset to zero
  // plus strand
  if (_strand == plus_strand)
  { 
    for (int i = 0; i <= _seqlength - *_pwlength; i++)
    { (*_pPspScore)[i] = (*psp)[i]; }
    if ((*psp)[0] != 1) // else all supplied entries were uniform
    { for (int i = 0; i <= _seqlength - *_pwlength; i++)
      { // convert the S(x) score into Pr(x) probability 
        // last 'w' values are not relevant (skip)
        // Pr(x) ~ S(x)/(1-S(x))
        (*_pPspScore)[i] /= (1-(*psp)[i]);
      }
    }
    // no need to set last L-w (vector is not that long)
  }
  else // minus strand 
  { 
    if (L == 2*_seqlength)
    { for (int i = 0; i <= _seqlength - *_pwlength; i++)
      { (*_pPspScore)[i] = (*psp)[_seqlength+i]; }
      if ((*psp)[_seqlength] != 1)
      { for (int i = 0; i <= _seqlength - *_pwlength; i++)
        { // convert the S(x) score into Pr(x) probability 
          (*_pPspScore)[i] /= (1-(*psp)[_seqlength+i]);
        }
      }
    }
    else // use "shift-symmetrical" values of + strand
    { 
      for (int i = 0; i <= _seqlength - *_pwlength; i++)
      { (*_pPspScore)[i] = (*psp)[_seqlength - *_pwlength - i];}
      if ((*psp)[_seqlength - *_pwlength] != 1)
      {
        for (int i = 0; i <= _seqlength - *_pwlength; i++)
        { // convert the S(x) score into Pr(x) probability 
          (*_pPspScore)[i] /= (1-(*psp)[_seqlength - *_pwlength - i]);
        }
      }
    }
  }
  //cerr << "Debug-dualSequenceComputation::LoadPspScores -end" << endl;
  return;
}

/******************************************************************************
  Description:  UpdateNbrInstDistribution()
                Update _pComputedDistr based on updated motifscores and prior
	  (code based on MotifSamplerRun::UpdateCopyProbability())
  Author_Date:  MC_2015/02/17
******************************************************************************/
Distribution *
dualSequenceComputation::UpdateNbrInstDistribution()
{
  //cerr << "Debug-dualSequenceComputation::UpdateNbrInstDistribution() -begin" << endl;
  // case fixed prior (f)
  if (_pFixedDistr != NULL) return _pFixedDistr; 

  // other cases : compute _pComputedDistr based on scores & prior2 & prior1.
  // (code based on MotifSamplerRun::UpdateCopyProbability())
	
  // first compute prior2 normalization constantes for c=1->M
  // this is only needed once (first time this function is called)
  if ( (*_pPrior2NormFactor)[0] == 0) // first time
  {
    for(int c = 1; c <= _maxNbrInst; c++)
    {
      (*_pPrior2NormFactor)[c-1] = 
          ComputeNInstancesProbability(_pPspScore,c,_seqlength,*_pwlength);
    }
  }
	// some variables
  double nextValue, lastValue = 1;
  uint copyNbr = 1;
  // pPs vector to store log[Pr(Sk|Akc)*Pr(Akc)] (terms without prior1)
  vector < double >*pPs = new vector < double >;
  // assign the prob to have NO instances (all segments are bg)
  // AND add the prob of the "reduced instances" being background too. 
  // only for PHMS because the "motif-score" also works on red-sequence instances
  pPs->push_back((_dLogP0 + _LogRedInstBg));

  ScoreVector * prior1 = NULL; //
  ScoreVector * pSiteScore = new ScoreVector(_seqlength-*_pwlength+1, 0); //

  ScoreVector * pCopyProb = new ScoreVector(_maxNbrInst+1, 0);

    // compute the motif-psp score for each segment
    for(int p = 0; p < _seqlength-*_pwlength+1; p++)
    { (*pSiteScore)[p] = (*_pExpWx)[p] * (*_pPspScore)[p];
    } 

    while (_seqlength > (int) copyNbr * (*_pwlength) 
		   && lastValue >= 0.001 // leave this constraint also for bMax-case...
		   && (int)copyNbr <= _maxNbrInst) // 
    {
//cerr << "copynbr=" << copyNbr << " ";
      // link to the correct distribution from _priorDistrs
      prior1 = (*_pPriorDistrs)[copyNbr]->GetVector(); // first element is NULL case f
      lastValue = 0; // reset

      // compute next value of CopyProbability
      nextValue =  _dLogP0 + 
            ComputeNInstancesProbability(pSiteScore, copyNbr, _seqlength, *_pwlength)
            - (*_pPrior2NormFactor)[copyNbr-1];
//cerr << "nextValue=" << nextValue << endl;
      //if (!isinf(nextValue) && !isnan(nextValue) && nextValue != -HUGE_VAL)
      if (finite(nextValue) && !isnan(nextValue) && nextValue != -HUGE_VAL)
      {
        // add element to pPs
        pPs->push_back(nextValue);

        // intermediate values
        double *pps = new double[copyNbr + 1];
        double stot = 0;
        for (uint j = 0; j < (copyNbr + 1); j++)
        { pps[j] = (*pPs)[j] + log((*prior1)[j]); //
          stot += exp(pps[j] - _dLogP0);
        }
        stot = _dLogP0 + log(stot);

        // update copy probability distribution
        for (uint j = 0; j <= copyNbr; j++)
        { (*pCopyProb)[j] = exp(pps[j] - stot);}

        // define last value in probability distribution
        lastValue = (*pCopyProb)[copyNbr];

        // delete intermediate variables
        delete[] pps; pps = NULL;
      }
      else {(*pCopyProb)[copyNbr] = 0;lastValue = 0;}

      // move to next
      copyNbr++;
    }

  // cleanup
  prior1 = NULL; // do not delete !
  // create distribution
  if (_pComputedDistr != NULL) delete _pComputedDistr; // reset
  _pComputedDistr = new Distribution(pCopyProb);
  // cleanup
  delete pCopyProb; pCopyProb = NULL;
	
/*
vector<double> * vec = _pComputedDistr->GetVector();
	for (int g = 0; g < int(vec->size()); g++) cerr << "vec["<<g<<"]=" << (*vec)[g] << " ";
	cerr <<endl;
*/
  // cerr << "dualSequenceComputation::UpdateNbrInstDistribution()-end" << endl;
  return _pComputedDistr;
}


/******************************************************************************
  Description:  sample one instance at random from uniform (_pMask) distribution 
                (the strand is given in this comp-object)
  Author_Date:  MC_2014/02/15
******************************************************************************/
Instance *
dualSequenceComputation::SampleUniformInstance()
{
  //cerr << "debug - dualSequenceComputation::SampleUniformInstance() - begin" << endl;
  // create uniform distribution

  Distribution * pDistr = new Distribution(_pMask->GetMask());
  int n = pDistr->TakeSample();

  // create instance from sampled start position 
  Instance * inst= new Instance( _pSeqObj, _strand, n, *_pwlength);
  // cleanup
  delete pDistr; pDistr = NULL;
  //cerr << "debug - dualSequenceComputation::SampleUniformInstance() - end" << endl;

  return inst;
}
/******************************************************************************
  Description:  sample instances from  (_pExpWx) distribution 
  Author_Date:  MC_2014/02/15: add impact PSP sampling
******************************************************************************/
InstanceList *
dualSequenceComputation::SampleInstances(int n)
{
  //cerr << "debug - dualSequenceComputation::SampleInstances()- begin" << endl;
  InstanceList * pList = new InstanceList;
  Instance * pInst(NULL);

  ScoreVector * pSiteScore = NULL; // 2013/12/30
  if (PSP_s)
  { // update motif-scores to motif-psp-scores 
    pSiteScore = new ScoreVector(_seqlength- *_pwlength +1, 0);
    for(int p = 0; p < _seqlength- *_pwlength + 1; p++)
    { (*pSiteScore)[p] = exp(log((*_pExpWx)[p])+log((*_pPspScore)[p]));}
  } 
  else { pSiteScore = _pExpWx;}
/*
    for(int p = 0; p < _seqlength- *_pwlength + 1; p++)
    { cerr << "(*pSiteScore)[" << p<<"]=" << (*pSiteScore)[p] << " ";}
	  cerr << endl;
*/
  // create distribution
  Distribution * pDist = new Distribution(pSiteScore);

  // sample n instances
  int p;
  for (int i = 0; i < n; i++)
  {
    p = pDist->TakeSample();
    // create instance from sampled start position 
    pInst= new Instance(_pSeqObj, _strand, p, *_pwlength);
    // store the segment score
    pInst->SetScore((*_pExpWx)[p]);
    pList->push_back(pInst);
    // and mask this position
    pDist->ApplyMask(p - *_pwlength + 1, 2 * (*_pwlength));
  }
  // cleanup
  delete pDist; pDist = NULL;
  if (PSP_s && pSiteScore != NULL) {delete pSiteScore;} 
  pSiteScore = NULL;
 // cerr << "debug - dualSequenceComputation::SampleInstances()- end" << endl;
  return pList;
}
/******************************************************************************
  Description:  select best instances from (_pExpWx) distribution 
  Author_Date:  MC_2012/12/30: add impact PSP sampling
******************************************************************************/
InstanceList *
dualSequenceComputation::SelectBestInstances(int N)
{
  //cerr << "debug - dualSequenceComputation::SelectBestInstances()- begin" << endl;
  InstanceList * pList = new InstanceList;
  Instance * pInst(NULL);

  ScoreVector * pSiteScore = NULL; // 2013/12/30
  if (PSP_s)
  { // update motif-scores to motif-psp-scores (test "extra" impact "sampling")
    pSiteScore = new ScoreVector(_seqlength- *_pwlength +1, 0);
    for(int p = 0; p < _seqlength- *_pwlength + 1; p++)
    { (*pSiteScore)[p] = exp(log((*_pExpWx)[p])+log((*_pPspScore)[p]));}
  } 
  else { pSiteScore = _pExpWx;}
  // create distribution
  Distribution * pDist = new Distribution(pSiteScore);

  // sample n instances
  int p;
  for (int i = 0; i < N; i++)
  {
    p = pDist->SelectMax();
    // create instance from sampled start position 
    pInst= new Instance( _pSeqObj, _strand, p, *_pwlength);
    // store the segment score
    pInst->SetScore((*_pExpWx)[p]);
    pList->push_back(pInst);
    // and mask this position
    pDist->ApplyMask(p - *_pwlength + 1, 2 * (*_pwlength));
  }
  // cleanup
  delete pDist; pDist = NULL;
  if (PSP_s && pSiteScore != NULL) {delete pSiteScore;} 
  pSiteScore = NULL;
  //cerr << "debug - dualSequenceComputation::SelectBestInstances()-end" << endl;

  return pList;
}


/******************************************************************************
  Description:  LogLikelihoodScore
  Author_Date:  MC_2012/12/11
******************************************************************************/
void
dualSequenceComputation::LogLikelihoodScore(InstanceList * pList)
{
  //cerr << "debug--dualSequenceComputation:LogLikelihoodScore - begin" << endl;
  _LL = 0;
  if (pList == NULL) return;

  // select number of instances distribution
  Distribution * pNbrInst = _pFixedDistr;
  if(_pFixedDistr == NULL) pNbrInst = _pComputedDistr;

  if (pNbrInst != NULL)
  {
    // local variables
    InstanceIterator i = pList->begin(); int ii = 1;
    double cp = 0;
    int nbr = (int)pNbrInst->Size();
    // add Pr(0)
    _LL = pNbrInst->GetValue(0);
    // iterate through instances
    for (; i != pList->end(); i++)
    {
      cp = SumArray(pNbrInst->GetVector(), ii, nbr - ii);
      cp *= (*_pExpWx)[(*i)->Start()];
      _LL += cp;
      ii++;
    }
  }
  else
  { cerr << "--Error-dualSequenceComputation::LogLikelihoodScore -- debug" << endl;}

  // take log
  _LL = log(_LL);
  //cerr << "debug--dualSequenceComputation:LogLikelihoodScore - end" << endl;
  return;
}
