#include "phyloFastaIO.h"

/******************************************************************************
  Description:  constructor for fastafile
                open file and set cursor on line starting with >>
  Author_Date:  MC_2009/01/07
******************************************************************************/
phyloFastaIO::phyloFastaIO(string * fastafile)
{
  // open file for reading
  _ifs.open(fastafile->c_str(), ios::in);
  // exit to mainPHMS if file is not open for reading or does not exist
  if (!_ifs) throw ErrorInputfiles((*fastafile).append(" not available"));

  // read lines untill first line starting with a '>>' is found 
  do
    getline(_ifs, _pLine, '\n');
  while ( !_ifs.eof() && _pLine.compare(0,2,">>"));

  // exit to mainPHMS if correct format (>>) is not found
  if (_ifs.eof()) throw ErrorInputfiles((*fastafile).append(" has no >> lines"));
    
  // set line-variables
  _hasGene = true;
  _hasNext = false; 
  _has = false;
}

/******************************************************************************
  Description:  constructor for backgroundfile and correlationsfile
                open file and set cursor on line starting with >
  Author_Date:  MC_2009/01/07
******************************************************************************/
phyloFastaIO::phyloFastaIO(string * borcfile, string dummy)
{
//cerr << "debug::phyloFastaIO" << endl;
  // dummy is not used, only to differentiate between fastafile-constructor
  // open file for reading
  _ifs.open(borcfile->c_str(), ios::in);
  // exit to mainPHMS if file is not open for reading or does not exist
  if (!_ifs) throw ErrorInputfiles((*borcfile).append(" not available"));

  // read lines untill first line starting with a '>' is found 
  do 
    getline(_ifs, _pLine, '\n');
  while ( !_ifs.eof() && !(_pLine[0] == '>' && _pLine[1] != '>'));

  // exit to mainPHMS if correct format (> ) is not found
  if (_ifs.eof()) throw ErrorInputfiles((*borcfile).append(" has no > lines"));
    
  // set line-variables
  _hasGene = false;
  _hasNext = true;
  _has = false;
}
/******************************************************************************
  Description:  create seqobj from speciesid, desription and sequence-code
                // this assumes the file is open and _pLine starts with >
                // must be checked in the function that calls this function
  Author_Date:  MC_2009/04/13
******************************************************************************/
SequenceObject *
phyloFastaIO::ReadSequence(const strand_modes &s)
{
  string::size_type pos = 0;
  string tokens = " \t\n\r\f";
  // create empty string to store sequence
  string *pSeqStr = new string(""); // for code
  string *pDescStr = new string(_pLine); // for >speciesid description
  //reset
  _hasGene = false;
  _hasNext = false;

  while (!_ifs.eof())
  {
    // read next line (should be code)
    getline(_ifs, _pLine, '\n');
    // move to next line if line is empty or starts with #
    if (_pLine.empty() || _pLine[0] == '#')
      continue;
    // break at line with new > or >>
    if (_pLine[0] == '>')
    {
      if (_pLine[1] == '>')
        _hasGene = true; 
      else
        _hasNext = true; 
      break;
    }
    else // line with code
    {
      pos = 0;
      while ((pos = _pLine.find_first_of(tokens)) != string::npos)
        _pLine.erase(pos, 1);
      // append line to seqStr;
      pSeqStr->append(_pLine);
    }
  }
  // create new sequence object
  SequenceObject *pSeq = NULL;
  if (! pSeqStr->empty() && ! pDescStr->empty())
  {
    pSeq = new SequenceObject(pSeqStr, s); 
    pSeq->SetDescription(pDescStr);
  }
  else 
  { delete pSeqStr; delete pDescStr; // catch in PHMSRun::LoadSequences
	  throw InvalidInput("--phyloFastaIO:: >sequence id and/or -code not available");
  }
  delete pSeqStr;
  delete pDescStr;
  return pSeq;
}

/******************************************************************************
  Description:  read genid in line with >>
                // this assumes the file is open and _pLine starts with >>
  Author_Date:  MC_2009/01/25
******************************************************************************/
string * 
phyloFastaIO::ReadGene()
{
  // substract genid from _pLine
  string *pID = new string(_pLine);
  // remove >>
  pID->erase(0, 2);
  // remove everything after white
  string::size_type pos = 0;
  if ((pos = pID->find_first_of(" \t\n\f\r")) != string::npos)
    *pID = pID->substr(0, pos);
  // evaluate id
  if ((int)pID->size() < 1)
  {
    delete pID; pID = NULL; 
    ProceedCursor(); // catch in mainRun::LoadSequences
	  throw InvalidInput("--phyloFastaIO:: >> no gene identification");
  }

  // set cursor ready on the next line
  ProceedCursor();

  if (!_hasNext) // no sequences for this gene
  { 
    delete pID; pID = NULL; // catch in mainRun::LoadSequences
	  throw InvalidInput("--phyloFastaIO:: >>gene without sequences>");
  }
  return pID;
}

/******************************************************************************
  Description:  read line after > 
                it can be '>legend(info)'  or  '>legend info' (corrFile)
				2012/11/21 : it can also be '>(newicktree);'
                // this assumes the file is open and _pLine starts with >
  Author_Date:  MC_2013/05/28
******************************************************************************/
string * 
phyloFastaIO::ReadNext()
{
  string *pID = NULL;
  string::size_type pos1, pos2;

  // evaluate on format
  if ((pos1 = _pLine.find_first_of("(")) != string::npos && 
      (pos2 = _pLine.find_last_of(")")) != string::npos)
  {
    pID = new string[2];
    if (pos1 > 1)
	{
      pID[0] = _pLine.substr(1, pos1-1); 
      pID[1] = _pLine.substr(pos1, pos2);
	} // with '(' and with ')'
    
    // remove all whites
    while ((pos1=pID[0].find_first_of(" \t\n\r\f")) != string::npos)
      pID[0].erase(pos1,1);
    while ((pos1=pID[1].find_first_of(" \t\n\r\f")) != string::npos)
      pID[1].erase(pos1,1);
  }
  else if ( (pos1=_pLine.find_first_of(" \t")) != string::npos)
  { 
    pID = new string[2];
    pID[0] = _pLine.substr(1, pos1-1);
    _pLine.erase(0, pos1+1);
    pID[1] = _pLine;
    // remove all whites
    while ((pos1=pID[1].find_first_of(" \t\n\r\f")) != string::npos)
      pID[1].erase(pos1,1);
  }
  else 
  { ProceedCursor(); 
      throw InvalidInput(
		 "--phyloFastaIO:: format should be >legend(info) or >legend info");
  }
  // set cursor ready on the next line 
  ProceedCursor();

  return pID;
}
/******************************************************************************
  Description:  ProceedCursor() : read next line (not empty, not #)
  Author_Date:  MC_2009/01/25
******************************************************************************/
void
phyloFastaIO::ProceedCursor()
{
  // reset
  _hasGene = false; 
  _hasNext = false; 
  _has = false;
  // read next valid input line
  while (!_ifs.eof())
  {
    getline(_ifs, _pLine, '\n');
    if (_pLine.empty() || _pLine[0] == '#')
      continue;
    if (_pLine[0] == '>') 
    {
      if (_pLine[1] == '>')
        _hasGene = true;
      else
        _hasNext = true;
      break;
    }
    _has = true; break;
  }
  return;
}
/******************************************************************************
  Description:  return id in line of format >id
                and put pointer on the next line
  Author_Date:  MC_2012/09/19
******************************************************************************/
string *
phyloFastaIO::ReadSeqID()
{
  // this line is read, cursor is moved to next line
  string * readid = NULL;

  if (_hasNext)
  {
    // store _pLine into string
    readid = new string(_pLine);
    readid->erase(0,1); // erase >
    // remove everything after white
    string::size_type pos = 0;
    if ((pos = readid->find_first_of(" \t\n\f\r")) != string::npos)
      *readid = readid->substr(0, pos);
    if ( readid->length() < 1) 
    { delete readid; return NULL;} // no id was provided

    ProceedCursor();
  }
  return readid;
}
/******************************************************************************
  Description:  read all the probs (L, or 2L) available in _pLine
                the values are separated by a space
                ! the last L-w values are (if not yet so) overwritten to zero !
                and put pointer on the next line (>)
				Rescale all entries to value range [0.1-0.9]
  Author_Date:  MC_2014/01/13
******************************************************************************/
ScoreVector *
phyloFastaIO::LoadPspData(int seqL, int wL)
{
  ScoreVector * pData = NULL;

  // _pLine format should be e.g. 0.4 0.6 0.8 ... 0.0 0.0 (tabs)
  string::size_type pos;
  pos = _pLine.find_first_not_of(" \t0.123456789\n\r\f");
  if ( pos != string::npos )
  {
	  throw ErrorInputfiles("--Error--phyloFastaIO::PSP entries: only white spaces and decimals characters (with dots) allowed.");
  }
  // load the probs into Scorevector
  int i = 0; // count how many entries loaded
  double Max = -1; double Min = 100000; // track to enable rescaling [0.1-0.9]
  double prob;
  pData = new ScoreVector;
  // now load data
  while (!_pLine.empty())
  { 
    pos = _pLine.find_first_not_of("0.123456789");
    istringstream istr(_pLine.substr(0,pos)); // convert to double
    istr >> prob;
    _pLine.erase(0,pos); // erase prob
    while ( _pLine.find_first_of("\t ") == 0) _pLine.erase(0,1); // erase space
    if ((i > seqL-wL && i < seqL) || (i > 2*seqL-wL && i < 2*seqL))
    { pData->push_back(0);} // overwrite to zero (masked positions at seq-end)
    else 
    { pData->push_back(prob);
      if (prob > Max){Max = prob;}
      if (prob < Min){Min = prob;}
    }
    i++;
  }
  // check if total pData equals L or 2L
  if ( i != seqL && i != 2*seqL)
  { 
    delete pData; pData = NULL;
	  cerr << "--Error--phyloFastaIO::number of PSP entries must equal length, or twice length of FASTA sequence.\n";
	  cerr << "seqL = " << seqL << ", psp loaded = " << i << endl;
	  throw ErrorInputfiles("--Error--phyloFastaIO::number of PSP entries must equal length, or twice length of FASTA sequence.");
  }

  // rescale all the loaded entries [0.1- 0.9]
  if (Max != Min) // else all entries are the same (uniform psp)
  {
    for (int j = 0; j < i; j++)
    { 
      if (!((j > seqL-wL && j < seqL) || (j > 2*seqL-wL && j < 2*seqL)))
      {(*pData)[j] = 0.1 + 0.8*((*pData)[j]-Min)/(Max-Min);}
    }
  }
  else // set all entries to '1' (uniform psp)
  {
    for (int j = 0; j < i; j++)
    { if (!((j > seqL-wL && j < seqL) || (j > 2*seqL-wL && j < 2*seqL)))
      {(*pData)[j] = 1;}
    }
  }

  // set pointer ready on next line
  ProceedCursor();

  return pData;
}
