#include "phyloSequenceComputation.h"


/******************************************************************************
  Description:  constructor 
  Author_Date:  MC_2014/02/15
******************************************************************************/
phyloSequenceComputation::phyloSequenceComputation(const strand_modes s, 
SequenceObject *pSeqObj, int *w, double &prox, vector<Distribution*> * dists, const bool &phylo2):
dualSequenceComputation(s, pSeqObj, w, prox, dists){
  //cerr << "Debug-phyloSequenceComputation::constructor -begin" << endl;
  _phylo_2 = phylo2;
 // computation tools for sampling
  _mutationProb = new double**[*w];
  for (int p = 0; p < *w; p++)
  { _mutationProb[p] = new double*[4];
    for (int n = 0; n < 4; n++)
    { _mutationProb[p][n] = new double[4]; 
      for (int nn = 0; nn < 4; nn++) _mutationProb[p][n][nn] = 0;
    }
  }
   
  //cerr << "Debug-phyloSequenceComputation::constructor -end" << endl;
}
/******************************************************************************
  Description:  destructor 
  Author_Date:  MC_2009/04/13
******************************************************************************/
phyloSequenceComputation::~phyloSequenceComputation()
{
  //cerr << "destructor-phyloSequenceComputation -begin" << endl;
  if (_mutationProb != NULL)
  { for (int w = 0; w < *_pwlength; w++)
    { for (int n = 0; n < 4; n++)
      { delete[] _mutationProb[w][n]; _mutationProb[w][n] = NULL;}
      delete[] _mutationProb[w]; _mutationProb[w] = NULL;
    }
    delete[] _mutationProb;
  }
  _mutationProb = NULL;

  //cerr << "destructor phyloSequenceComputation -end" << endl;
}

/******************************************************************************
  Description:  ComputeBasicMutProb based on reduced(gene)motif  // logaritmic !
	  Author_Date:  MC_2014/02/19 : apply _phylo_2 option
******************************************************************************/
void
phyloSequenceComputation::ComputeBasicMutProb(Matrix reduced)
{
//cerr << "--debug::phyloSequenceComputation::ComputeBasicMutProb-begin" << endl;
  for (int w = 0; w < *_pwlength; w++)
  {
    for (int n = 0; n < 4; n++) // parent=row
    { 
      for (int nn = 0; nn < 4; nn++) // daughter=column
      {
        if (_phylo_2)
        {
          if (n == nn)
            _mutationProb[w][n][nn] = log( *_pProx + (1 - *_pProx)*reduced[w][nn] ); 
          else 
            _mutationProb[w][n][nn] = log((1 - *_pProx)*reduced[w][nn]);
        }
        else {_mutationProb[w][n][nn] = log(reduced[w][nn]);}
      }
    }
  }
//cerr << "--debug::phyloSequenceComputation::ComputeBasicMutProb-end" << endl;

  return; 
}

/******************************************************************************
  Description:  UpdateMotifScores : for each segment
      /// LEAN revision :: do NOT consider combination of multiple instances
      /// so just compute the motif-score for each single segment
	  // the motif scores are logaritmic
  Author_Date:  MC_2014/02/17
******************************************************************************/
void 
phyloSequenceComputation::UpdateMotifScores(double** redseq)
{
  //cerr << "debug -- phyloSequenceComputation::UpdateMotifScores- begin" << endl;
  //cerr <<  "for " << *_pSeqObj->GetID() << "_" << *_pSeqObj->GetGenID() <<  endl;
  int nt; double scoreW;double ** pComp = NULL;
  // update the scores in _pMotifScore(_seqlength - *_pwlength + 1, 0);
  // iterate through all segments
  for (int s = 0; s < _seqlength - *_pwlength +1; s++)
  { 
    // start with probabilities from reduced-sequence-set (extern2-variable)
    pComp = new double*[*_pwlength];
    for (int w = 0; w < *_pwlength; w++)
    { pComp[w] = new double[4];
      for (int i = 0; i < 4; i++) 
      { pComp[w][i] = redseq[w][i];}
    }
    // add the prob of the segment to the reduced-sequence-set of instances
    // iterate through each position of the segment-instance
    for (int w = 0; w < *_pwlength; w++)
    { // read the nucleotide in the sequence
      nt = (*_pSequence)[s + w]; // daughter
      if (nt != -1)
      { // compute the postional-nucleotide-score (logaritmic!!)
        // by selecting the according column from _mutationProb
        for(int n = 0; n < 4; n++) // each possible parent
        { pComp[w][n] += _mutationProb[w][n][nt];}
      }
      else { for(int n = 0; n < 4; n++) pComp[w][n] += log(0.001);}
    }

    double term;
    // now compute the probability of the segment+reduced_instances-set
    (*_pMotifScore)[s] = 0; // reset
    for (int w = 0; w < *_pwlength; w++)
    { 
      scoreW = 0;
      // SUM over candidate parent nucleotides A,C,G,T 
      // pComp consists of logs, thus first revert with exp
      for(int n = 0; n < 4; n++) 
      { term = exp(pComp[w][n]);
        scoreW += exp(pComp[w][n]);
      }
      // multiply (~ add log) over motif width positions
      (*_pMotifScore)[s] += log(scoreW);
    }
    // leave it as a logaritmic score
    //NOT: (*_pMotifScore)[s] = exp((*_pMotifScore)[s]); // non-logaritmic
 
    //cleanup
    for (int w = 0; w < *_pwlength; w++)
    { delete[] pComp[w]; pComp[w] = NULL;}
    delete[] pComp; pComp = NULL;
  }
 //cerr << "debug -- phyloSequenceComputation::UpdateMotifScores- end" << endl;
  return;
}

/******************************************************************************
  Description:  UpdatepExpWx(double fixedLogDenum);
        // this function follows on UpdateMotifScores
        // apply mask, segment-bg and redSeqLogDenum(fixed) contributions
  Author_Date:  MC_2014/02/17
******************************************************************************/
void 
phyloSequenceComputation::UpdatepExpWx(double fixedLogDenum)
{
  //cerr << "debug -- phyloSequenceComputation::UpdatepExpWx-begin" <<  endl;
  vector<int> *mask = _pMask->GetMask();

  // update the scores in _pExpWx(_seqlength - *_pwlength + 1, 0);
  // iterate through all segments
  for (int s = 0; s < _seqlength - *_pwlength +1; s++)
  { 
    // apply mask, segment-bg-score powered by (1/w) (i.e equals log/w), 
    // and fixed red-inst-set score
    // the score in _pExpWx are non-logaritmic
    (*_pExpWx)[s] = (*mask)[s] * 
		  exp((*_pMotifScore)[s] - (*_pSegmentBgScore)[s] 
                 - fixedLogDenum );
  }
  //cerr << "debug -- phyloSequenceComputation::UpdatepExpWx-end" <<  endl;
  return;
}
